unit NeoS0M_CmdList;

interface

uses SysUtils,Classes,Windows,NeoS0M_Types;

const UNKNOWN_PARAM_COUNT=MaxInt-2;

type TPTypes=set of TParamType;

type TParamInfo=record
      PType:TPTypes;
      MayBeDefault,MayBeEnd,MayNotExist:Boolean;
      InputMethod:UnicodeString;  //  
      DefaultValue:UnicodeString; //     '',     ,    
      DescriptionIndex:Integer;   //    
      DataControlsIndex:Integer;  //     ,    
     end;
     TParamInfoList=array of TParamInfo;

     TCmdInfo=record
      CmdCode:Cardinal;
      CmdGroupCode:Cardinal;
      MinParamCount,MaxParamCount:Cardinal;
      CmdName:UnicodeString;
      DescrIdx:Cardinal;
      ParamInfo:TParamInfoList;
      CompileToUnicodeStrings:Boolean;
     end;
     PCmdInfo=^TCmdInfo;

     TCmdInfoList=array of TCmdInfo;

//      
function __LoadCmdList(const fname:UnicodeString; var cl:TCmdInfoList):Integer;

//   
procedure __DeinitializeCmdInfo(var cl:TCmdInfoList);

//     
function GetCmdIndexInList(const cl:TCmdInfoList; const cmdname:UnicodeString):Integer;

//         
function Ansi_UpStrComp(str1:PAnsiChar; str1len:Integer; str2:PAnsiChar; str2len:Integer):Integer;
function Unicode_UpStrComp(str1:PWideChar; str1len:Integer; str2:PWideChar; str2len:Integer):Integer;

implementation

uses
{$IFNDEF COMPILER}
NC_FrameCreator,
{$ENDIF}

NeoS0M_CompilerTypes, StrUtils, Compiler_CaseConversion;

const CMD_LIST_GRANULARITY=32;

function ExtractQuotedLine(var s:String; const start_q,end_q:String):String;
var p1,p2:Integer;
begin
  p1:=Pos(start_q,s);
  if p1=0
    then Result:=''
    else
      begin
        p2:=PosEx(end_q,s,p1+1);
        if p2=0
          then Result:=''
          else
            begin
              Result:=Copy(s,p1+1,p2-p1-1);
              Delete(s,p1,p2-p1+1);
            end;
      end;
end;

function ExtractCmdInfo(s:String; var param:TParamInfo):Boolean;
var i,r2:Integer;
begin
  with param do
  begin
    Result:=True;
    //   
    PType:=[];
    MayBeDefault:=False;
    MayNotExist:=False;
    MayBeEnd:=False;
    InputMethod:='';
    DefaultValue:='';
    i:=1;
    while i<=Length(s) do
    begin
      case s[i] of
        '#':  begin
                Include(PType,ptDWORD);
                InputMethod:=s[i];
              end;
        '!':  begin
                Include(PType,ptBinBlock);
                InputMethod:=s[i];
              end;
        '+':  begin
                Include(PType,ptInteger);
                InputMethod:=s[i];
              end;
        '*':  begin
                Include(PType,ptExtended);
                InputMethod:=s[i];
               end;
        '$':  begin
                Include(PType,ptUnicodeString);
                InputMethod:=s[i];
              end;
        'd':  MayBeDefault:=True;
        'n':  MayNotExist:=True;
        'e':  MayBeEnd:=True;
        '(','[','`','{':
              begin
                Dec(i);
                Break;
              end
        else Result:=False;
      end;//case
      Inc(i);
    end;//while
  //            
  Delete(s,1,i);
  //     (    )
  DefaultValue:=ExtractQuotedLine(s,'(',')');
  if DefaultValue<>''
    then InputMethod:=DefaultValue;
  //     (    )
  Val(ExtractQuotedLine(s,'[',']'),DescriptionIndex,r2);
  Result:=Result and (r2=0);
  //       (    )
  DataControlsIndex:=StrToIntDef(ExtractQuotedLine(s,'{','}'),-1);
  //    ,   
  DefaultValue:=ExtractQuotedLine(s,'`','`');
{$IFNDEF COMPILER}
  //    -      
  if DefaultValue=''
    then DefaultValue:=DefaultByInputMethod(InputMethod);
{$ENDIF}
end;//with
end;

function __InitCmdInfo(const s:UnicodeString; var ci:TCmdInfo):boolean;
var ic,ic2,ic3:Cardinal;
    i:Integer;
    ParamList:TStringList;
begin
 result:=true;
 ParamList:=TStringList.Create;
 ic:=ExtractStrings([','],[' '],PWideChar(s),ParamList);
 if ic<5
 then result:=False //    
 else
  begin
   with ParamList do
    begin
    //        BinaryBlock
    ci.CompileToUnicodeStrings:=False;
    //   
    Val(Strings[0],ci.CmdCode,ic);
    //    
    Val(Strings[1],ci.CmdGroupCode,ic3);
    //  .  
    if (Strings[2]='?') or (Strings[2]='$') then
      begin
        ci.CompileToUnicodeStrings:=Strings[2]='$';
        ci.MaxParamCount:=UNKNOWN_PARAM_COUNT;
        ic2:=0;
      end
     else Val(Strings[2],ci.MaxParamCount,ic2);
    ci.MinParamCount:=0;
    if (ic<>0) or (ic2<>0) or (ic3<>0) //         
     then result:=False
     else
      begin
       ci.CmdName:=Strings[3];
       //    
       Val(Strings[4],ci.DescrIdx,ic);
       if ic<>0
        then Result:=False;

       //    
       i:=ParamList.Count-5; if i<0 then i:=0;
       SetLength(ci.ParamInfo,i);
       for i:=5 to ParamList.Count-1 do
        begin
          Result:=Result and ExtractCmdInfo(Strings[i],ci.ParamInfo[i-5]);
          Inc(ci.MinParamCount);
        end;

      end;//else
    end;//with
  end;//else
 FreeAndNil(ParamList);
end;

procedure QuickSortCmdList(var arr:TCmdInfoList; L, R: Integer);
var
  I, J, P: Integer;
  tmp:TCmdInfo;

function SCompare(i,j:integer):integer;
begin
 if arr[i].CmdName>arr[j].CmdName
  then result:=1
  else begin
        if arr[i].CmdName<arr[j].CmdName
         then result:=-1
         else result:=0;
       end;
end;

begin
if L<R then
begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while SCompare(I, P) < 0 do Inc(I);
      while SCompare(J, P) > 0 do Dec(J);
      if I <= J then
      begin
        tmp:=arr[i];
        arr[i]:=arr[j];
        arr[j]:=tmp;
//        ExchangeItems(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortCmdList(arr, L, J);
    L := I;
  until I >= R;
end;
end;
{ DONE : ,    }
{
procedure Self_Test_Cmd_List(cl:TCmdInfoList);
var i:Integer;
begin
  for i:=Low(cl) to High(cl) do
   begin
     if GetCmdIndexInList(cl,cl[i].CmdName)<>i
      then Windows.MessageBox(0,PCHAR ('Self-test FAILED at '+cl[i].CmdName),'ERROR',0);
   end;
end;
}
function __LoadCmdList(const fname:UnicodeString; var cl:TCmdInfoList):Integer;
var f:TextFile;
    s:UnicodeString;
    ic,str_idx:Integer;
begin
  str_idx:=0;
{$I-}
  ic:=0;
  AssignFile(f,fname);
  Reset(f);
  if IOResult=0 then
  begin
    repeat
      Result:=0; //    -   ,   OK
      Readln(f,s);
      Inc(str_idx);
      if IOResult=0 then
      begin
        if Length(cl)<=ic
          then SetLength(cl,ic+CMD_LIST_GRANULARITY);
        if __InitCmdInfo(s,cl[ic])
          then Inc(ic)
            else //     -     
            begin
              Result:=str_idx;
              Break;
            end;
     end;//if IOResult=0
   until Eof(f);
   CloseFile(f);
  end
 else result:=-1;
{$I+}
 SetLength(cl,ic); //   -       
 if Result<>0
  then __DeinitializeCmdInfo(cl)
  else QuickSortCmdList(cl,0,Length(cl)-1);

// Self_Test_Cmd_List(cl);
end;

function Ansi_UpStrComp(str1:PAnsiChar; str1len:Integer; str2:PAnsiChar; str2len:Integer):Integer;
//         
var i,m:Integer;
    c1,c2:AnsiChar;
begin
  Result:=0;
  if str1len<str2len
    then m:=str1len
    else m:=str2len;
  //    ,    ,  
  for i:=1 to m do
  begin
    c1:=AnsiUpperCaseTable[str1^];
    c2:=AnsiUpperCaseTable[str2^];
    if c1=c2 then
      begin
        Inc(str1);
        Inc(str2);
      end
    else
      begin
        if c1>c2
          then Result:=1
          else Result:=-1;
        Exit;
      end;//else
  end;//for

  //       
  if Result=0 then
  begin
    //         
    if str1len>str2len
      then Result:=1
      else if str1len<str2len
            then Result:=-1;
  end;
end;

function Unicode_UpStrComp(str1:PWideChar; str1len:Integer; str2:PWideChar; str2len:Integer):Integer;
//         
var i,m:Integer;
    c1,c2:WideChar;
begin
  Result:=0;
  if str1len<str2len
    then m:=str1len
    else m:=str2len;
  //    ,    ,  
  for i:=1 to m do
  begin
    c1:=UpperCaseTable[str1^];
    c2:=UpperCaseTable[str2^];
    if c1=c2 then
      begin
        Inc(str1);
        Inc(str2);
      end
    else
      begin
        if c1>c2
          then Result:=1
          else Result:=-1;
        Exit;
      end;//else
  end;//for

  //       
  if Result=0 then
  begin
    //         
    if str1len>str2len
      then Result:=1
      else if str1len<str2len
            then Result:=-1;
  end;
end;

function GetCmdIndexInList(const cl:TCmdInfoList; const cmdname:UnicodeString):Integer;
//        ,   
var l,r,m,c_r:Integer;
begin
  l:=0;
  r:=Length(cl)-1;
  while l<=r do
  begin
    m:=(r+l) shr 1;
    c_r:=Unicode_UpStrComp(PWideChar(cmdname),Length(cmdname),PWideChar(cl[m].CmdName),Length(cl[m].CmdName));
    if c_r>0
      then l:=Succ(m)
      else begin
            if c_r<0
              then r:=Pred(m)
              else begin Result:=m; Exit; end;
           end;
  end;//while
  Result:=-1;
end;

procedure __DeinitializeCmdInfo(var cl:TCmdInfoList);
var i:Integer;
begin
  for i:=0 to Length(cl)-1 do
    begin
      SetLength(cl[i].ParamInfo,0);
      Finalize(cl[i]);
    end;//for
  Finalize(cl);
end;

end.
 