//{$DEFINE DEBUG}
unit NeoS0M_CompilerMain;

interface

uses NeoS0M_CompilerTypes;

function InitializeCompiler(CmdList:PWideChar):Integer; exports InitializeCompiler;
procedure FinalizeCompiler; exports FinalizeCompiler;

procedure BeforeCompiling(ScriptPtr:PRawScript); exports BeforeCompiling;
procedure AfterCompiling(ScriptPtr:PRawScript); exports AfterCompiling;
procedure CompileScript(scr:PRawScript); exports CompileScript;
function SaveCompiledScript(scr:PRawScript; fname:PWideChar; JoinWithExisting:Boolean):Boolean; exports SaveCompiledScript;
function CompileNeoS0MFile(InFName,OutFName:PWideChar;RawScript:PRawScript; var ErrLine:Cardinal; ErrModuleBuffer:PChar):Cardinal;exports CompileNeoS0MFile;
//          -.
// RawScript:  nil,      RawScript,   nil -  
//       
function CompileNeoS0MEXE(InFName,OutFName:PWideChar;RawScript:PRawScript;var ErrLine:Cardinal; ErrModuleBuffer:PChar):Cardinal;exports CompileNeoS0MEXE;
//   ,   CompileNeoS0MFile,      ,     

implementation

uses NeoS0M_CmdList,
     NeoS0M_Types, NeoS0M_Constants,
     NeoS0M_UnifiedArray,
     NeoS0M_ReplaceIcon,
     CRC64,
     Windows,SysUtils,StrUtils,AnsiStrings,Classes
//{$IFDEF DEBUG},UNI_DbgOutput {$ENDIF}
     , NeoS0M_DescrList, NeoS0M_UNIFIED;

const NOT_A_HEX_DIGIT=$FFFFFFFF;
      SCRIPT_MAIN_MODULE=-1;
      LOAD_MAIN_MODULE=-1;
      NO_ERROR_LINE=$FFFFFFFF;

      CHAR_IS_ANSI=1;

      PREFIX_DIRECTIVE='&';

var BuiltInCommands:TCmdInfoList=nil;
    Keywords:TDescrList;
    CharInfoTable:array[WideChar] of Byte;

    _KEYWORD_SECTION:UnicodeString='SECTION';
    _KEYWORD_CODE:UnicodeString='CODE';
    _KEYWORD_DATA:UnicodeString='DATA';

    _FILE_DIRECTIVE:UnicodeString='&FILE';
    _FILESIZE_DIRECTIVE:UnicodeString='&FILESIZE';
    _DB_DIRECTIVE:UnicodeString='&DB';
    _DW_DIRECTIVE:UnicodeString='&DW';
    _DD_DIRECTIVE:UnicodeString='&DD';
    _DI_DIRECTIVE:UnicodeString='&DI';
    _DF_DIRECTIVE:UnicodeString='&DF';
    _INCLUDE_DIRECTIVE:UnicodeString='&INCLUDE';
    _USES_DIRECTIVE:UnicodeString='&USES';
    _ICON_DIRECTIVE:UnicodeString='&ICON';

const LABEL_ARRAY_GRANULARITY=16;
      LABEL_REF_GRANULARITY=16;
type TParamSet=set of TParamType;
const AnyParam=[ptDefault,ptInteger,ptExtended,ptIntArray,ptExtArray,ptBinBlock,ptBinBlockList,ptDWORD,ptDWORDArray,ptVariable,ptEndOfParam,ptUnicodeString,ptUnicodeStringList];
{      //   
      CompatibleParameters:array [TParamType] of TParamSet=(
        [], // ptDefault
        [ptInteger,ptDWORD,ptVariable,ptByte,ptWord], // ptInteger
        [ptExtended,ptInteger,ptDWORD,ptVariable,ptByte,ptWord], // ptExtended
        [ptIntArray,ptVariable], // ptIntArray
        [ptExtArray,ptVariable], // ptExtArray
        [ptBinBlock,ptVariable], // ptBinBlock
        [ptBinBlockList,ptVariable], // ptBinBlockList
        [ptDWORD,ptInteger,ptVariable,ptByte,ptWord], // ptDWORD
        [ptDWORDArray,ptVariable], // ptDWORDArray
        [ptVariable], // ptVariable
        [], // ptEndOfParam
        [ptInteger,ptDWORD,ptVariable,ptByte,ptWord], // ptByte
        [ptInteger,ptDWORD,ptVariable,ptByte,ptWord], // ptWord
        [ptUnicodeString,ptBinBlock,ptVariable], // ptUnicodeString
        [ptUnicodeStringList,ptVariable] // ptUnicodeStringList
      );}

function DequotedString(const s:String):String;
var l:Integer;
begin
 l:=Length(s);
 //       -  
 if l>=2 then
  begin
    if (s[1]='"') and (s[l]='"')
      then Result:=Copy(s,2,l-2)
      else Result:=s
  end
 else Result:=s; //if l>2
end;

function ReparsePath(const RawScript:TRawScript; const s:String):String;
begin
with RawScript do
begin
  //          -   
  if (ReparseFileNames) and (ScriptPath<>'') and (s<>'') then
    begin
      //         ,    
      if (Pos(':\',s)<>0) or (Pos(':/',s)<>0) or (Pos('\\',s)<>0)
        then Result:=DequotedString(s)
        else
          begin
            //  ,   
            Result:=DequotedString(s);
            //      
            if Result<>'' then
              begin
                if CharInSet(Result[1],['/','\'])
                  then Delete(Result,1,1);
                Result:=ScriptPath+Result;
              end;
          end;
    end
  else Result:=DequotedString(s);
end;//with
end;

function RoundUp(value,round_base:cardinal):cardinal;
var rm:cardinal;
begin
 rm:=value mod round_base;
 if rm<>0 then result:=value+(round_base-rm)
  else result:=value;
end;

function IfThen(Value:Boolean; a,b:Integer):Integer;overload;
begin
  if Value then Result:=a else Result:=b;
end;

function HexDigitToDec(c:AnsiChar):Cardinal;
begin
  case c of
   '0'..'9':  Result:=Ord(c)-48;
   'a'..'f':  Result:=Ord(c)-87;
   'A'..'F':  Result:=Ord(c)-55;
  else Result:=NOT_A_HEX_DIGIT;
  end;//case
end;

function WHexDigitToDec(c:WideChar):Cardinal;
begin
  case c of
   '0'..'9':  Result:=Ord(c)-Ord(WideChar('0'));
   'a'..'f':  Result:=Ord(c)-Ord(WideChar('a'))+10;
   'A'..'F':  Result:=Ord(c)-Ord(WideChar('A'))+10;
  else Result:=NOT_A_HEX_DIGIT;
  end;//case
end;

function AnsiUpStrCmp(const s1,s2:AnsiString):Integer;
begin
 Result:=ANSI_UpStrComp(PAnsiChar(s1),Length(s1),PAnsiChar(s2),Length(s2))
end;

function UnicodeUpStrCmp(const s1,s2:UnicodeString):Integer;
begin
 Result:=Unicode_UpStrComp(PWideChar(s1),Length(s1),PWideChar(s2),Length(s2))
end;

procedure ErrLineByLineNo(var scr:TRawScript; LineNo:Integer);
//   ErrLine  ErrModule    
//        
begin
  with scr,scr.ScriptLines[LineNo] do
    begin
      ErrLine:=LineNumber;
      ErrModule:=ModuleIndex;
    end;
end;

function GetCmdIndex(const s:UnicodeString):Pointer;
//  
var i:Integer;
begin
  Result:=nil;
  i:=GetCmdIndexInList(BuiltInCommands,s);
  if i>=0
    then Result:=Addr(BuiltInCommands[i]);
// else
end;

function FindSectionByName(ScriptPtr:PRawScript; const sname:AnsiString):Integer;
var i:Integer;
begin
  for i:=Low(ScriptPtr^.Sections) to High(ScriptPtr^.Sections)
    do if sname=ScriptPtr^.Sections[i].SectionName then
        begin
          Result:=i;
          Exit;
        end;
  Result:=-1;
end;

function AddSection(ScriptPtr:PRawScript; const sname:AnsiString; section_flags:Cardinal):boolean;
var l:Integer;
begin
  with ScriptPtr^ do
  begin
    l:=Length(Sections);
    Result:=False;
    // ,       
    if FindSectionByName(ScriptPtr,sname)<0 then
    begin
      SetLength(Sections,l+1);
      CurrentSection:=l; //      
      with Sections[l] do
      begin
        Ptr:=nil;
        SectionName:=sname;
        Size:=0;
        BufSize:=0;
        Flags:=section_flags;
        Labels:=TStringList.Create;
      end;//with
      Result:=True;
    end;//if FindSectionByName(...
  end;//with
end;

function DoSection(ScriptPtr:PRawScript; const sname:AnsiString; section_flags:UnicodeString):Integer;
//   "SECTION"  . 
var s_idx:Integer;
    flags:Cardinal;
begin
  //     - 
  if (AnsiUpStrCmp(sname,LDS_SECTION_NAME)=0) or (AnsiUpStrCmp(sname,JUMP_SECTION_NAME)=0) then
  begin
    Result:=LS_MODIFICATION_OF_SYS_SECTION;
    Exit;
  end;
  flags:=$FFFFFFFF;
  //       
  if UnicodeUpStrCmp(section_flags,_KEYWORD_CODE)=0 then flags:=CODE_SECTION
    else if UnicodeUpStrCmp(section_flags,_KEYWORD_DATA)=0
          then flags:=DATA_SECTION;
  //           
  if flags=$FFFFFFFF then
    begin
      Result:=LC_INVALID_SECTION_DESCRIPTION;
      Exit;
    end;

  if AddSection(ScriptPtr,sname,flags) then Result:=LC_OK
  else
    //     - ,      
    begin
      with ScriptPtr^ do
      begin
        s_idx:=FindSectionByName(ScriptPtr,sname);
        if Sections[s_idx].Flags<>flags
          then Result:=LS_SECTION_PARAMETERS_DIFFERS
        else
          begin
            CurrentSection:=s_idx;
            Result:=LC_OK;
          end;
      end;//with
    end;
end;

procedure BeforeCompiling(ScriptPtr:PRawScript);
begin
 with ScriptPtr^ do
  begin
   SetLength(Sections,0);
   ScriptText:=nil;
   ScriptModules:=nil;
   SetLength(ScriptLines,0);
   CRC64_Needed:=True;
   NewIcon:='';
   ReparseFileNames:=True;
   StopErrorHandling:=False;
   TrueLabelRefCount:=0;
   AddSection(ScriptPtr,LDS_SECTION_NAME,DATA_SECTION); //   -  LDS
   AddSection(ScriptPtr,JUMP_SECTION_NAME,DATA_SECTION); //   -   
   AddSection(ScriptPtr,GENERAL_SECTION_NAME,CODE_SECTION);
  end;
end;

procedure CleanupScript(RawScript:PRawScript);
//     ,     
begin
  if Assigned(RawScript^.ScriptText)
    then FreeAndNil(RawScript^.ScriptText);
  if Assigned(RawScript^.ScriptModules)
    then FreeAndNil(RawScript^.ScriptModules);
  SetLength(RawScript^.ScriptLines,0);
end;

procedure AfterCompiling(ScriptPtr:PRawScript);
var i:Integer;
begin
 with ScriptPtr^ do
  begin
    CleanupScript(ScriptPtr);
   for i:=Low(Sections) to High(Sections) do
    begin
     with Sections[i] do
     begin
      FreeMem(Ptr);
      FreeAndNil(Labels);
     end;//with
    end;
   Sections:=nil;
  end;
end;

function FindQuotedWordEnd(const s:UnicodeString; _pos:Cardinal):Cardinal;
var l:Cardinal;
    IsExit:Boolean;
    c:WideChar;
begin
  result:=_pos;
  IsExit:=False;
  l:=Length(s);
  c:=s[_pos];
  if _pos=l
    //   -    , ,     
    then Result:=0
    else
      begin
        repeat
          result:=PosEx(c,s,result+1);
          if Result=l // []  -   
            then IsExit:=True
            else
              begin
                //      - ,   
                if s[result+1]=' '
                  then IsExit:=True
                else
                  begin
                    //   - ,    -   ,    
                    if s[result+1]=c then
                      begin
                        Inc(Result,1);
                        if Result=l
                          then Result:=0;
                      end
                    //         -    
                    else Result:=0;
                  end;//else
              end;//else
        until IsExit or (result=0);
      end;
end;

function PosNonSpaceEx(const s:UnicodeString;p:Cardinal):Integer;
//    ;  -   
var i:Integer;
begin
  result:=Length(s);
  for i:=p to Length(s) do
    if s[i]<=' '
      then begin
            result:=i-1;
            Break;
           end;
end;

function GetFirstWord(const s:UnicodeString; var _pos:Cardinal; var Failure:Boolean):UnicodeString;
var l:Cardinal;
    ep:Integer;
    fnsp:Cardinal;
begin
  Failure:=False;
  l:=Length(s);
  //   -
  while (_pos<=l) and (s[_pos]<=' ')
    do Inc(_pos);
  //     -      
  if _pos>l then
    begin
      _pos:=0;
      result:='';
      Exit;
    end
  else
    begin
      fnsp:=_pos;
      //     - 
      case s[_pos] of
        '''','"','!':
          begin
            ep:=FindQuotedWordEnd(s,_pos);
            Failure:=ep=0;
          end
        else
          begin
            ep:=PosNonSpaceEx(s,_pos);
//      if ep=Length(s) then Dec(ep);
          end;
      end;//case
      Result:=IfThen(ep>0,Copy(s,fnsp,ep-Integer(fnsp)+1),'');
      _pos:=ep;
    end;//else
end;

procedure AddSomeBytesToBlock(var block:Pointer; var datasize,blocksize:Cardinal; Bytes:Pointer; BytesCount:Cardinal);
var BytesNeeded,nbs:Cardinal;
    tmp_ptr:Pointer;
begin
 BytesNeeded:=datasize+BytesCount;
 if BytesNeeded>blocksize then
  begin
   nbs:=RoundUp(BytesNeeded,COMPILING_BUFFER_GRANULARITY);
   GetMem(tmp_ptr,nbs);
   if block<>nil then
    begin
     Move(block^,tmp_ptr^,datasize);
     FreeMem(block);
    end;//if block<>nil
    block:=tmp_ptr;
    blocksize:=nbs;
  end;//if BytesNeeded>blocksize
  tmp_ptr:=Pointer(Cardinal(block)+datasize);
  Move(Bytes^,tmp_ptr^,BytesCount);
  Inc(datasize,BytesCount);
end;

function CleanAnsiString(const s:UnicodeString; var res:AnsiString):Cardinal;
var l,h_p:Integer;
    h:Cardinal;
begin
  //  Unicode-  ANSI
  res:=AnsiString(s);
  if Length(res)-2>=0 then
    begin
      //   
      if (res[1]<>'''') or (res[Length(res)]<>'''') then
        begin
          Result:=LC_DIRECTIVE_PARAMETER_IS_INVALID;
          Exit;
        end;
      //     
      Delete(res,1,1);
      Delete(res,Length(res),1);
      //      
      res:=AnsiStrings.StringReplace(res,'''''','''',[rfReplaceAll]);
      h_p:=0;
      repeat
        //   #
        h_p:=AnsiStrings.PosEx('#',res,h_p+1);
        if h_p<>0 then //     '#'
          begin
            l:=Length(res);
            if h_p<=l-1 then //      ?   -     .
              begin
                if res[h_p+1]='#' //     #,    
                  then Delete(res,h_p,1)
                else
                  begin
                    if h_p<=l-2 then // C '#'      ,   2 
                    begin
                      //          ANSI-...
                      h:=(HexDigitToDec(res[h_p+1]) shl 4) or HexDigitToDec(res[h_p+2]);
                      // ...     #-
                      if h<=$FF then
                        begin
                          res[h_p]:=AnsiChar(h);
                          Delete(res,h_p+1,2);
                        end;
                    end;//if h_p<l-2
                  end;//else
              end;//if h_p<=l-1
          end;//h_p<>0
      until h_p=0;
      Result:=LC_OK;
    end
  else Result:=LC_DIRECTIVE_PARAMETER_IS_INVALID;
end;

function CleanUnicodeString(var s:UnicodeString):Cardinal;
var l,h_p:Integer;
    h:Cardinal;
begin
  l:=Length(s)-2;
  if l>=0 then
   begin
      //   
      if (s[1]<>'"') or (s[Length(s)]<>'"') then
        begin
          Result:=LC_DIRECTIVE_PARAMETER_IS_INVALID;
          Exit;
        end;
     //     
     Delete(s,1,1);
     Delete(s,Length(s),1);
     //      
     s:=StringReplace(s,'""','"',[rfReplaceAll]);

    h_p:=0;
    repeat
    h_p:=PosEx('#',s,h_p+1);
    if h_p<>0 then //     '#'
      begin
        l:=Length(s);
        if h_p<=l-1 then
        begin
          if s[h_p+1]='#' //     "#"
            then Delete(s,h_p,1) //   "#" 
            else
              begin
                if h_p<=l-4 then // C '#'      ,   4 
                begin
                  h:=(WHexDigitToDec(s[h_p+1]) shl 12) or (WHexDigitToDec(s[h_p+2]) shl 8) or (WHexDigitToDec(s[h_p+3]) shl 4) or WHexDigitToDec(s[h_p+4]);
                  //   "#"  Uncode-  .    4   "#"
                  if h<=$FFFF then
                    begin
                      s[h_p]:=WideChar(h);
                      Delete(s,h_p+1,4);
                    end;//if h<=$FFFF
              end;//else
            end;//else
        end;//if h_p<=l-1
     end;
    until h_p=0;
    Result:=LC_OK;
   end
  else Result:=LC_DIRECTIVE_PARAMETER_IS_INVALID;
end;

function CompileHardcodedFile(var script:TRawScript; fname:UnicodeString):Cardinal;
//        
var hfile:Cardinal;
    buffer:Pointer;
    size,readed:Cardinal;
begin
 if script.Sections[script.CurrentSection].Flags<>DATA_SECTION then
  begin
    Result:=LC_DATA_COMPILING_NOT_ALLOWED;
    Exit;
  end;

 if not FileExists(fname) then
 begin
   Result:=LC_FILE_NOT_EXISTS;
   Exit;
 end;

 hfile:=CreateFileW(PWideChar(fname),GENERIC_READ,FILE_SHARE_READ,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
 if hFile<>INVALID_HANDLE_VALUE then
 //   
  begin
   SetFilePointer(hfile,0,nil,FILE_BEGIN);
   //        
   size:=GetFileSize(hfile,nil);
   GetMem(buffer,size);
   //     
   if ReadFile(hfile,buffer^,size,readed,nil) then
      //   
      begin
        with script.Sections[script.CurrentSection]
          do AddSomeBytesToBlock(Ptr,Size,BufSize,buffer,readed);
        Result:=LC_OK;
      end
    else Result:=LC_FILE_CANT_BE_LOADED;
   //    -  
   if buffer<>nil then FreeMem(buffer);
   CloseHandle(hfile);
  end
 else //   
  Result:=LC_FILE_CANT_BE_LOADED;
end;

function CompileFileSize(var script:TRawScript; fname:UnicodeString):Cardinal;
var my_size:Cardinal;
    true_size:Int64;
    hfile:CARDINAL;
begin
 if script.Sections[script.CurrentSection].Flags<>DATA_SECTION then
  begin
    Result:=LC_DATA_COMPILING_NOT_ALLOWED;
    Exit;
  end;
  
 if not FileExists(fname) then
 begin
   Result:=LC_FILE_NOT_EXISTS;
   Exit;
 end;

 hfile:=CreateFileW(PWideChar(fname),GENERIC_READ,FILE_SHARE_READ,nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
 if hFile<>INVALID_HANDLE_VALUE then
 //   
  begin
   //   
   my_size:=GetFileSize(hfile,@true_size);
   true_size:=(true_size shl 32)+my_size;
   //  8     
   with script.Sections[script.CurrentSection]
    do AddSomeBytesToBlock(Ptr,Size,BufSize,@true_size,SizeOf(true_size));
   //  ,    
   CloseHandle(hfile);
   Result:=LC_OK;
  end
 else //   
  Result:=LC_FILE_CANT_BE_LOADED;
end;

function CompileByte(var script:TRawScript; b:UnicodeString):Cardinal;
var i64:Int64;
    err:Integer;
    bt:Byte;
begin
  Val(b,i64,err);
  if err<>0
    then Result:=LC_DIRECTIVE_PARAMETER_IS_NOT_NUMBER
    else
      begin
        if (i64>255) or (i64<-128)
          then Result:=LC_DIRECTIVE_PARAMETER_VIOLATES_BOUNDS
          else
            begin
              bt:=i64;
              with script.Sections[script.CurrentSection] do
              begin
                if Flags=DATA_SECTION then
                  begin
                    AddSomeBytesToBlock(Ptr,Size,BufSize,@bt,SizeOf(bt));
                    Result:=LC_OK;
                  end
                else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
              end;//with
          end;
      end;
end;

function CompileDBAnsiString(var script:TRawScript; const s:AnsiString):Cardinal;
begin
  with script.Sections[script.CurrentSection] do
  begin
    if Flags=DATA_SECTION then
      begin
        AddSomeBytesToBlock(Ptr,Size,BufSize,@s[1],Length(s));
        Result:=LC_OK;
      end
    else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
  end;//with
end;

function CompileDBUnicodeString(var script:TRawScript; const s:UnicodeString):Cardinal;
begin
  with script.Sections[script.CurrentSection] do
  begin
    if Flags=DATA_SECTION then
      begin
        AddSomeBytesToBlock(Ptr,Size,BufSize,@s[1],Length(s)*2);
        Result:=LC_OK;
      end
    else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
  end;//with
end;

function CompileWord(var script:TRawScript; b:UnicodeString):Cardinal;
var i64:Int64;
    err:Integer;
    wd:WORD;
begin
  Val(b,i64,err);
  if err<>0
    then Result:=LC_DIRECTIVE_PARAMETER_IS_NOT_NUMBER
    else
      begin
        if (i64>65535) or (i64<-32768)
          then Result:=LC_DIRECTIVE_PARAMETER_VIOLATES_BOUNDS
          else
          begin
            wd:=i64;
            with script.Sections[script.CurrentSection] do
              begin
                if Flags=DATA_SECTION then
                  begin
                    AddSomeBytesToBlock(Ptr,Size,BufSize,@wd,SizeOf(wd));
                    Result:=LC_OK;
                  end
                else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
              end;//with
          end;//else
      end;//else
end;

function CompileDWord(var script:TRawScript; b:UnicodeString):Cardinal;
var i64:Int64;
    err:Integer;
    dw:Cardinal;
begin
  Val(b,i64,err);
  if err<>0
    then Result:=LC_DIRECTIVE_PARAMETER_IS_NOT_NUMBER
    else
      begin
        if (i64>4294967295) or (i64<-2147483648)
          then Result:=LC_DIRECTIVE_PARAMETER_VIOLATES_BOUNDS
          else
          begin
            dw:=i64;
            with script.Sections[script.CurrentSection] do
              begin
                if Flags=DATA_SECTION then
                  begin
                    AddSomeBytesToBlock(Ptr,Size,BufSize,@dw,SizeOf(dw));
                    Result:=LC_OK;
                  end
                else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
              end;//with
          end;
      end;
end;

function CompileInt(var script:TRawScript; b:UnicodeString):Cardinal;
var i64:Int64;
    err:Integer;
begin
  val(b,i64,err);
  if err<>0
    then Result:=LC_DIRECTIVE_PARAMETER_IS_NOT_NUMBER
    else
      begin
      with script.Sections[script.CurrentSection] do
        begin
          if Flags=DATA_SECTION then
            begin
              AddSomeBytesToBlock(Ptr,Size,BufSize,@i64,SizeOf(i64));
              Result:=LC_OK;
            end
          else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
        end;//with
      end;
end;

function CompileFloat(var script:TRawScript; b:UnicodeString):Cardinal;
var f:Extended;
    err:Integer;
begin
  Val(b,f,err);
  if err<>0
    then Result:=LC_DIRECTIVE_PARAMETER_IS_NOT_NUMBER
    else
      begin
      with script.Sections[script.CurrentSection] do
        begin
          if Flags=DATA_SECTION then
            begin
              AddSomeBytesToBlock(Ptr,Size,BufSize,@f,SizeOf(f));
              Result:=LC_OK;
            end
          else Result:=LC_DATA_COMPILING_NOT_ALLOWED;
        end;//with
      end;//else
end;


function CompileCommandName(var script:TRawScript; const cmdname:UnicodeString; var cmdptr:PCmdInfo):Cardinal;
//      4- ;
begin
  cmdptr:=GetCmdIndex(cmdname);
  with script.Sections[script.CurrentSection] do
  begin
  if cmdptr=nil
    then Result:=LC_INVALID_COMMAND_NAME
    else
      begin
        AddSomeBytesToBlock(Ptr,Size,BufSize,cmdptr,SizeOf(Cardinal));
        result:=LC_OK;
      end;//else
  end;//with
end;

function ParseLabelReference(const s:String; const CurrentSectionName:AnsiString; var Name:String; var SectionName:AnsiString; var ModuleGlobal:Boolean):Integer;
var dot_pos,l:Integer;
begin
  l:=Length(s);
  dot_pos:=Pos('.',s);
  // ,      
  if l>2
    then ModuleGlobal:=s[2]='@'
    else ModuleGlobal:=False;
  //   ,   
  if dot_pos>0 then
    begin
      if ModuleGlobal
        then SectionName:=Copy(s,3,dot_pos-3)
        else SectionName:=Copy(s,2,dot_pos-2);
      //   
      Name:=Copy(s,dot_pos+1,l-dot_pos);
    end
  else
    begin
      SectionName:=CurrentSectionName;
      //   
      if ModuleGlobal
        then Name:=Copy(s,3,l-2)
        else Name:=Copy(s,2,l-1);
    end;
  if Name=''
    then Result:=LC_INVALID_LABEL_NAME
    else Result:=LC_OK;
end;

function AddReferenceToLabel(var Script:TRawScript; const LabelName:String; ofs:Cardinal):Integer;
var l:Integer;
    n:String;
    s_n:AnsiString;
    ModuleGlobal:Boolean;
begin
  with Script do
  begin
    Result:=ParseLabelReference(LabelName,Sections[CurrentSection].SectionName,n,s_n,ModuleGlobal);
    if Result=LC_OK then
      begin
          l:=Length(LabelReferences);
          if TrueLabelRefCount=l
            then SetLength(LabelReferences,l+LABEL_REF_GRANULARITY);
          with LabelReferences[TrueLabelRefCount] do
            begin
              // ,     (..  )
              SectionWhere:=Script.CurrentSection;
              Name:=n;
              Section:=s_n;
              if ModuleGlobal
                then Module:=MODULE_ANY
                else Module:=ScriptLines[NowCompiling].ModuleIndex;
              LabelOffsetInFile:=ofs;
              LineNumber:=NowCompiling;
            end;//with
          //     
          Inc(TrueLabelRefCount);
    end;//if Result=LC_OK
  end;//with
end;

function ParseLabel(const s:String; var Name:String; var SectionGlobal,ModuleGlobal:Boolean):Integer;
var l,s_p:Integer;
begin
  l:=Length(s);
  s_p:=2;
  //        ?
  if l>s_p then
    begin
      ModuleGlobal:=s[2]='@';
      if ModuleGlobal
        then Inc(s_p);
    end
  else ModuleGlobal:=False;
  //       ?
  if l>s_p then
    begin
      SectionGlobal:=s[s_p]='.';
      if SectionGlobal
        then Inc(s_p);
    end
  else SectionGlobal:=False;
  //   
  Name:=Copy(s,s_p,l-s_p+1);
  if (Name='') or (Name='.') or (Name='@')
    then Result:=LC_INVALID_LABEL_NAME
    else Result:=LC_OK;
end;

function AddLabel(var scr:TRawScript; const s:UnicodeString; ofs:Cardinal):Integer;
//          
var n:String;
    s_g,m_g:Boolean;
    idx:Integer;
begin
  Result:=ParseLabel(s,n,s_g,m_g);
  if Result=LC_OK then
    begin
      //   
      with scr.Sections[scr.CurrentSection] do
        begin
          idx:=Labels.Add(n);
          if Length(LabelsInfo)<=idx
            then SetLength(LabelsInfo,idx+LABEL_ARRAY_GRANULARITY);
          //    
          with LabelsInfo[idx] do
            begin
              OffsetInSection:=ofs;
              SectionGlobal:=s_g;
              if m_g
                then ModuleIndex:=MODULE_ANY
                else ModuleIndex:=scr.ScriptLines[scr.NowCompiling].ModuleIndex;
            end;
        end;
    end;
end;

function CompileParameter(var script:TRawScript; const paramname:UnicodeString; param_no:Cardinal; var specials:Cardinal; ParamWaiting:TParamSet):Cardinal;
//    ;
// specials:   -        UnicodeString'
//     -    ,        
var ParamType:TParamType;
    ForceCompilingToUnicodeStr:Boolean;
    Buf:array[0..31] of Byte;


  procedure subCompileAsDWORD(p:Cardinal; pt:TParamType);
  begin
    Buf[0]:=Byte(pt);
    PCardinal(@Buf[1])^:=p;
    with script.Sections[script.CurrentSection]
      do AddSomeBytesToBlock(Ptr,Size,BufSize,@Buf[0],SizeOf(p)+1);
    Result:=LC_OK;
  end;//function

  function CompileAsLocalVarIndex(const s:UnicodeString):Cardinal;
  var idx:Int64;
      res:Integer;
  begin
    Val(s,idx,Res);
    if res=0 then
      begin
        if idx>=LOCAL_VARIABLE_INDEX_FLAG
          then Result:=LC_INVALID_PARAM_VARIDX
          else
            begin
              subCompileAsDWORD(idx or LOCAL_VARIABLE_INDEX_FLAG,ptDWORD);
              Result:=LC_OK;
            end;//else
      end
   else Result:=LC_INVALID_PARAM_VARIDX;
  end;

  procedure subCompileAsFloat(p:Extended);
  begin
    Buf[0]:=Byte(ptExtended);
    PExtended(@Buf[1])^:=p;
    with script.Sections[script.CurrentSection]
      do AddSomeBytesToBlock(Ptr,Size,BufSize,@Buf[0],SizeOf(p)+1);
    Result:=LC_OK;
  end;

  procedure subCompileAsAnyInteger(ptype:TParamType; start:Pointer; Len:Cardinal);
  begin
    Buf[0]:=Byte(ptype);
    Move(start^,Buf[1],Len);
    with script.Sections[script.CurrentSection]
      do AddSomeBytesToBlock(Ptr,Size,BufSize,@Buf[0],Len+1);
  end;

  function CompileAsAnyInteger(const s:UnicodeString):Cardinal;
  //    
  var i64:Int64;
      c:Integer;
  begin
   Val(s,i64,c);
   if c=0 then
    begin
     Result:=LC_OK;
     if (i64<=127) and (i64>=-128) //     
      then subCompileAsAnyInteger(ptByte,@i64,SizeOf(ShortInt))
      else
       begin
        if (i64<=32767) and (i64>=-32768) //     
         then subCompileAsAnyInteger(ptWord,@i64,SizeOf(SmallInt))
         else
          begin
           if (i64<=4294967295) and (i64>=0) //     DWORD
            then subCompileAsAnyInteger(ptDWord,@i64,SizeOf(Cardinal))
            //   -   Int64
            else subCompileAsAnyInteger(ptInteger,@i64,SizeOf(Int64))
          end;
       end;//if (i64<=127) and (i64>=-128)
    end
   else Result:=LC_INVALID_PARAM_INT;
  end;

  function CompileAsCharCode(const s:UnicodeString):Cardinal;
  //     
  var _s:AnsiString;
  begin
    Result:=LC_OK;
    case Length(s) of
      2:  begin
            _s:=AnsiString(s);
            if CharInfoTable[s[2]] and CHAR_IS_ANSI=0
              then Result:=LC_INVALID_CHAR_CODE
              else subCompileAsAnyInteger(ptByte,@_s[2],SizeOf(AnsiChar));
          end;
      3:  begin
            if s[2]='~'
              then subCompileAsAnyInteger(ptWord,@s[2],SizeOf(WideChar))
              else Result:=LC_INVALID_CHAR_CODE;
          end
      else Result:=LC_INVALID_CHAR_CODE;
    end;//case
  end;

  function CompileAsVarIndex(const s:UnicodeString):Cardinal;
  //     
  var s1:UnicodeString;
      val32:Cardinal;
  begin
   case WhichVariable(s,val32,s1) of
    wwrGlobalVariable,wwrLocalVariable:
      begin
        subCompileAsDWORD(val32,ptVariable);
        Result:=LC_OK;
      end;
    else Result:=LC_INVALID_PARAM_VARIDX;
   end;
  end;//function

  function CompileAsFloat(const s:UnicodeString):Cardinal;
  var e:Extended;
      err_c:Integer;
  begin
    Val(s,e,err_c);
    if err_c=0 then
      begin
        subCompileAsFloat(e);
        Result:=LC_OK;
      end
    else Result:=LC_INVALID_PARAM_FLOAT;
  end;

  function CompileAsInt64(const s:UnicodeString):Cardinal;
  var i64:Int64;
      c:Integer;
  begin
   Val(s,i64,c);
   if c=0 then
    begin
     Result:=LC_OK;
     subCompileAsAnyInteger(ptInteger,@i64,SizeOf(Int64))
    end
   else Result:=LC_INVALID_PARAM_INT;
  end;

  function CompileAsLabelIndex(const s:UnicodeString):Cardinal;
  const arr:array[1..5] of Byte=(Byte(ptDWORD),0,0,0,0);
  begin
    with script.Sections[script.CurrentSection] do
     begin
        //     
        Result:=AddReferenceToLabel(Script,s,Size+1);
        AddSomeBytesToBlock(Ptr,Size,BufSize,@arr,SizeOf(arr));
     end;
  end;//function CompileAsLabelIndex

  function TryToCompileAsUnknown(const s:UnicodeString):Cardinal;
  var ex:Extended;
      err:Integer;
  begin
   result:=CompileAsAnyInteger(s);
   if Result<>LC_OK then //  - ,     ?
    begin
     Val(s,ex,err);
     if err<>0
      then result:=LC_UNKNOWN_PARAM_TYPE
      else begin subCompileAsFloat(ex); Result:=LC_OK; end;
    end;
  end;

  procedure subCompileBinaryBlockOrUnicodeString(var Block:AnsiString; ItemSize:Integer);
  var pre_block:PArrayHeader;
      old_l,lb:Integer;
      ofs,i:Integer;
  begin
    { DONE :           LDS,    ,     LDS  ,     }
    old_l:=Length(block); //   
    block:=StringOfChar(AnsiChar(#0),SizeOf(TArrayHeader))+block+#0;
    pre_block:=@block[1];
    //   
    with pre_block^ do
    begin
      ElemSize:=ItemSize;
      //        ,  , ..   0
      ElemCount:=(old_l+2) div ItemSize;
      TrueElemCount:=ElemCount;
      IsDynamic:=False;
    end;//with

    with script.Sections[0] {LDS} do
    begin
      lb:=Length(Block)+1;
      //     LDS    LDS
      ofs:=Size;
      //   LDS   
      for i:=0 to Integer(Size)-lb do
      begin
        if CompareMem(Pointer(Integer(Ptr)+i),PAnsiChar(block),lb) then
        //   -     LDS
        begin
          ofs:=i;
          Break;
        end;
      end;//for

      //     -   LDS,      LDS
      if ofs=Integer(Size)
      //  ANSI- Delphi   WinAPI,  0   
        then AddSomeBytesToBlock(Ptr,Size,BufSize,@block[1],lb);

      //       LDS      LDS
      if ItemSize=1
        then Buf[0]:=Byte(ptBinBlock)
        else Buf[0]:=Byte(ptUnicodeString);
      PCardinal(@Buf[1])^:=ofs;
    end;//with

    with script.Sections[script.CurrentSection]
      do AddSomeBytesToBlock(Ptr,Size,BufSize,@Buf[0],SizeOf(ofs)+1);
  end;

  function CompileAsANSIString(const s:UnicodeString):Cardinal;
  var l,h_p:Integer;
      s1:AnsiString;
      h:Cardinal;
  begin
    { TODO :   ClearAnsiString }
    //  Unicode-  ANSI
    s1:=AnsiString(s);
    if Length(s1)-2>=0 then
    begin
      //     
      Delete(s1,1,1);
      Delete(s1,Length(s1),1);
      //      
      s1:=AnsiStrings.StringReplace(s1,'''''','''',[rfReplaceAll]);
//     StringReplace(s1,'""','"',[rfReplaceAll]);
      h_p:=0;
      repeat
        //   #
        h_p:=AnsiStrings.PosEx('#',s1,h_p+1);
        if h_p<>0 then //     '#'
          begin
            l:=Length(s1);
            if h_p<=l-1 then //      ?   -     .
              begin
                if s1[h_p+1]='#' //     #,    
                  then Delete(s1,h_p,1)
                else
                  begin
                    if h_p<=l-2 then // C '#'      ,   2 
                    begin
                      //          ANSI-...
                      h:=(HexDigitToDec(s1[h_p+1]) shl 4) or HexDigitToDec(s1[h_p+2]);
                      // ...     #-
                      if h<=$FF then
                        begin
                          s1[h_p]:=AnsiChar(h);
                          Delete(s1,h_p+1,2);
                        end;
                    end;//if h_p<l-2
                  end;//else
              end;//if h_p<=l-1
          end;//h_p<>0
      until h_p=0;
      subCompileBinaryBlockOrUnicodeString(s1,1);
      result:=LC_OK;
    end
   else result:=LC_INVALID_PARAM_STRING;
  end;

  function CompileAsUNICODEString(s:UnicodeString; IsBinaryBlock:Boolean):Cardinal;
  var l,h_p:Integer;
      h:Cardinal;
      buf:AnsiString;
  begin
   { TODO :   ClearUnicodeString }
   l:=Length(s)-2;
   if l>=0 then
   begin
     //     
     Delete(s,1,1);
     Delete(s,Length(s),1);
     //      
     s:=StringReplace(s,'""','"',[rfReplaceAll]);

    h_p:=0;
    repeat
    h_p:=PosEx('#',s,h_p+1);
    if h_p<>0 then //     '#'
      begin
        l:=Length(s);
        if h_p<=l-1 then
        begin
          if s[h_p+1]='#' //     "#"
            then Delete(s,h_p,1) //   "#" 
            else
              begin
                if h_p<=l-4 then // C '#'      ,   4 
                begin
                  h:=(WHexDigitToDec(s[h_p+1]) shl 12) or (WHexDigitToDec(s[h_p+2]) shl 8) or (WHexDigitToDec(s[h_p+3]) shl 4) or WHexDigitToDec(s[h_p+4]);
                  //   "#"  Uncode-  .    4   "#"
                  if h<=$FFFF then
                    begin
                      s[h_p]:=WideChar(h);
                      Delete(s,h_p+1,4);
                    end;//if h<=$FFFF
              end;//else
            end;//else
        end;//if h_p<=l-1
     end;
    until h_p=0;

     l:=Length(s);
     SetLength(buf,l+l);//   ANSI-    UNICODE-
     Move(s[1],buf[1],l+l); //  UNICODE-  ANSI-
     //  Unicode-  LDS  UNICODE_STRING  BINARY_BLOCK     
     subCompileBinaryBlockOrUnicodeString(buf,IfThen(IsBinaryBlock,1,2));
     Result:=LC_OK;
    end//if l<>0
   else result:=LC_INVALID_PARAM_STRING;
  end;


  function CompileAsHexDump(const s:UnicodeString):Cardinal;
  var i,cc:Integer;
      s2:AnsiString;
      b,c:Byte;
      _Shift:Boolean;
      HexDigit:Cardinal;
  begin
    cc:=0;
    //       
    for i:=1 to Length(s)
      do if WHexDigitToDec(s[i])<>NOT_A_HEX_DIGIT
          then Inc(cc);
    if Odd(cc)
      then Result:=LC_INVALID_HEX_DIGITS_COUNT
      else
        begin
          //       
          cc:=cc shr 1;
          SetLength(s2,cc);
          _Shift:=False;
          c:=0;
          //       " "
          for i:=Length(s) downto 1 do
            begin
              HexDigit:=WHexDigitToDec(s[i]);
              if HexDigit<>NOT_A_HEX_DIGIT then
              begin
                b:=HexDigit;
                if _Shift then
                  begin
                    s2[cc]:=AnsiChar((b shl 4) or c);
                    Dec(cc);
                    _Shift:=False;
                  end//if _Shift
                else
                  begin
                    c:=b;
                    _Shift:=True;
                  end;//else
              end//if HexDigit<>NOT_A_HEX_DIGIT
            end;//for
          subCompileBinaryBlockOrUnicodeString(s2,1);
          Result:=LC_OK;
    end//if Odd(cc)
  end;

  function ExtractSubstrFromStr(const s,delimiter:UnicodeString; var StartPos:Integer; var IsLast:Boolean):UnicodeString;
  var delim_found:Integer;
  begin
   delim_found:=PosEx(delimiter,s,StartPos);
   if delim_found=0 then //   
     begin
      IsLast:=True;
      Result:=Copy(s,StartPos,Length(s)-StartPos+1);
     end
    else
     begin //  
      Result:=Copy(s,StartPos,delim_found-StartPos);
      IsLast:=delim_found=Length(s);
      StartPos:=delim_found+1;
     end;
  end;

  function Str2Int(const s:UnicodeString; var value:Integer):Boolean;
  //   True -     
  var err_code:Integer;
  begin
    Val(s,value,err_code);
    Result:=err_code=0;
  end;

  function CompileAsDateTime(const s:UnicodeString):Cardinal;
  label DoIt;
  var pos_in_str:Integer;
      IsLast:Boolean;
      tmp_int:Integer;
      st:SYSTEMTIME;
      ft:FILETIME;
  begin
    Result:=LC_INVALID_DATE_TIME;
    with st do //    - 0:00:00.000
    begin
      wHour:=0;
      wMinute:=0;
      wSecond:=0;
      wMilliseconds:=0;

      //    -  
      pos_in_str:=1;
      //   
      if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
        then wDay:=tmp_int
        else Exit;
      if IsLast
        then Exit; //     

      //  
      if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
        then wMonth:=tmp_int
        else Exit;
      if IsLast
        then Exit; //     

      //  
      if Str2Int(ExtractSubstrFromStr(s,',',pos_in_str,IsLast),tmp_int)
        then wYear:=tmp_int
        else Exit;
      if IsLast
        //      -    
        then goto DoIt;

      //  
      if Str2Int(ExtractSubstrFromStr(s,':',pos_in_str,IsLast),tmp_int)
        then wHour:=tmp_int
        else Exit;
      if IsLast
        then goto DoIt;

      //  
      if Str2Int(ExtractSubstrFromStr(s,':',pos_in_str,IsLast),tmp_int)
        then wMinute:=tmp_int
        else Exit;
      if IsLast
        then goto DoIt;

      //  
      if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
        then wSecond:=tmp_int
        else Exit;
      if IsLast
        then goto DoIt;

      //  
      if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
        then wMilliseconds:=tmp_int
        else Exit;
      //      ,     -  (..   )
      if not IsLast
        then Exit;
     end;//with

  DoIt:
    if SystemTimeToFileTime(st,ft) then
      begin
        subCompileAsAnyInteger(ptInteger,@ft,SizeOf(ft));
        Result:=LC_OK;
      end
  end;

begin
  ForceCompilingToUnicodeStr:=specials=FORCE_UNICODE_STRINGS;
  //  specials -        
  specials:=0;
with script.Sections[script.CurrentSection] do
begin
 if paramname='$' then
  begin
    if ptDefault in ParamWaiting then
      begin
        ParamType:=ptDefault;
        AddSomeBytesToBlock(Ptr,Size,BufSize,@ParamType,1);
        Result:=LC_OK;
      end
      else Result:=LC_DEFAULT_NOT_ALLOWED;
  end
  else if paramname='.' then
        begin
          if ptEndOfParam in ParamWaiting then
            begin
              ParamType:=ptEndOfParam;
              AddSomeBytesToBlock(Ptr,Size,BufSize,@ParamType,1);
              Result:=LC_OK;
              Specials:=COMPILED_END_OF_PARAM;
            end
          else Result:=LC_END_OF_PARAM_NOT_ALLOWED;
        end//if paramname='.'
   else
    begin
     case paramname[1] of
      '_':     Result:=CompileAsLocalVarIndex(Copy(paramname,2,Length(paramname)-1));
      '+','-': Result:=CompileAsAnyInteger(paramname);
      '~':     Result:=CompileAsCharCode(paramname);
      '%':     if ptInteger in ParamWaiting
                then Result:=CompileAsInt64(Copy(paramname,2,Length(paramname)-1))
                else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      '@':     Result:=CompileAsVarIndex(paramname);
      '*':     if ptExtended in ParamWaiting 
                then Result:=CompileAsFloat(Copy(paramname,2,Length(paramname)-1))
                else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      '''':    if (ptBinBlock in ParamWaiting) or (ptUnicodeString in ParamWaiting)
                then Result:=CompileAsANSIString(paramname)
                else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      '"':     if ptBinBlock in ParamWaiting
                //  BinBlock ,      
                //  UnicodeString -   UnicodeString (      ,      )
                then Result:=CompileAsUNICODEString(paramname,not ForceCompilingToUnicodeStr)
                else if ptUnicodeString in ParamWaiting
                      then Result:=CompileAsUNICODEString(paramname,False)
                      else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      '^':     Result:=CompileAsLabelIndex(paramname);
      '#':     if ptInteger in ParamWaiting
                then Result:=CompileAsDateTime(Copy(paramname,2,Length(paramname)-1))
                else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      '!':     if ptBinBlock in ParamWaiting
                then Result:=CompileAsHexDump(paramname)
                else Result:=LC_INCOMPATIBLE_PARAM_TYPE;
      else     Result:=TryToCompileAsUnknown(paramname);
     end;//case
    end;
end;//with
end;

function GetLineCompilingMode(const s:UnicodeString; var lc:TLineCompilingMode):Cardinal;
begin
  Result:=LC_INVALID_COMMAND_NAME;
  if Length(s)>1
    then if s[1]='^' then
          begin
            lc:=lcmLabels;
            Result:=LC_OK;
          end;
end;

function SimpleCompileLine(const ScriptLine:UnicodeString; var script:TRawScript):Cardinal;
// line -   , llen -   
// script:  ;  -  
var CurrentWord:UnicodeString;
    CurrentAnsiWord:AnsiString;
    WordEnd:Cardinal; // ,       
    WordCounter:Integer;
    CommandInfo:PCmdInfo;
    MaxWordCount:Integer;
    Specials:Cardinal;
    Failure:Boolean;

    LineCompilingMode:TLineCompilingMode;
    SectionAttr:UnicodeString;
    DirectiveMode:TDirectiveMode;
    WaitingParam:TParamSet; //   
begin
 result:=LC_OK;
 WordEnd:=1;
 WordCounter:=0;
 MaxWordCount:=0;
 repeat
  CurrentWord:=GetFirstWord(ScriptLine,WordEnd,Failure);
  //    -   ,  ,    ,
  // ..        
//  if CurrentWord=';' then CurrentWord:='';
  Inc(WordEnd);
  if (CurrentWord='') or (CurrentWord=';')
    then WordEnd:=0
    else
      begin
       Inc(WordCounter);
       //    -   
       if WordCounter=1 then
        begin
          //      "&" -  
          if CurrentWord[1]=PREFIX_DIRECTIVE
            then LineCompilingMode:=lcmDirective
            else LineCompilingMode:=lcmCommand;
          //    ""
          if UnicodeUpStrCmp(CurrentWord,_KEYWORD_SECTION)=0 then
            begin
              LineCompilingMode:=lcmSectionName;
              MaxWordCount:=3;
            end;
        end;

     case LineCompilingMode of
      lcmCommand: begin
                   Result:=CompileCommandName(script,CurrentWord,CommandInfo);
                   if Result=LC_OK then
                    begin
                     //    -    ,   .    .  +1
                     LineCompilingMode:=lcmParameters;
                     MaxWordCount:=CommandInfo^.MaxParamCount+1;
                    end
                    else
                     begin
                      //    ,     
                      Result:=GetLineCompilingMode(CurrentWord,LineCompilingMode);
                      if Result=LC_OK then
                      begin
                      case LineCompilingMode of
                       lcmLabels: begin
                                    with script.Sections[script.CurrentSection]
                                      do AddLabel(script,CurrentWord,Size);
                                    MaxWordCount:=1;
                                  end;
                      end;//case
                      end;//if Result=LC_OK
                     end//else
                  end;
      lcmSectionName:
                  begin
                     case WordCounter of
                      1:  ;
                      2:  begin //      -  
                            SectionAttr:=CurrentWord;
                          end;
                      3:  begin //      -  
                            Result:=DoSection(@script,AnsiString(CurrentWord),SectionAttr);
                          end;
                      else Result:=LC_INVALID_SECTION_DESCRIPTION;
                     end;//case
                  end;
      lcmParameters: begin
                      if WordCounter<=MaxWordCount then
                        begin
                          //   -       ,   = -2
                          if WordCounter-2<Length(CommandInfo.ParamInfo)
                            then begin
                                    with CommandInfo.ParamInfo[WordCounter-2] do
                                      begin
                                        WaitingParam:=PType;
                                        if MayBeDefault then Include(WaitingParam,ptDefault);
                                        if MayBeEnd then Include(WaitingParam,ptEndOfParam);
                                      end;//with
                                 end
                            else WaitingParam:=AnyParam;
                         Specials:=IfThen(CommandInfo.CompileToUnicodeStrings,FORCE_UNICODE_STRINGS,0);
                         result:=CompileParameter(script,CurrentWord,WordCounter+1,Specials,WaitingParam);
                         //       , ,       .  +1
                         // (..     ,  )
                         if Specials=COMPILED_END_OF_PARAM
                          then WordCounter:=MaxWordCount;
                        end//WordCounter<=...
                       else result:=LC_TOO_MANY_PARAMETERS;
                     end;
      lcmDirective:  begin
                       if WordCounter=1 then
                       //      ,    
                       begin
                          DirectiveMode:=dmNone;
                          //       ,    
                          if UnicodeUpStrCmp(CurrentWord,_FILE_DIRECTIVE)=0
                            then DirectiveMode:=dmFile;
                          if UnicodeUpStrCmp(CurrentWord,_FILESIZE_DIRECTIVE)=0
                            then DirectiveMode:=dmFileSize;
                          if UnicodeUpStrCmp(CurrentWord,_DB_DIRECTIVE)=0
                            then DirectiveMode:=dmDB;
                          if UnicodeUpStrCmp(CurrentWord,_DW_DIRECTIVE)=0
                            then DirectiveMode:=dmDW;
                          if UnicodeUpStrCmp(CurrentWord,_DD_DIRECTIVE)=0
                            then DirectiveMode:=dmDD;
                          if UnicodeUpStrCmp(CurrentWord,_DI_DIRECTIVE)=0
                            then DirectiveMode:=dmDI;
                          if UnicodeUpStrCmp(CurrentWord,_DF_DIRECTIVE)=0
                            then DirectiveMode:=dmDF;
                          if UnicodeUpStrCmp(CurrentWord,_INCLUDE_DIRECTIVE)=0
                            then DirectiveMode:=dmInclude;
                          if UnicodeUpStrCmp(CurrentWord,_USES_DIRECTIVE)=0
                            then DirectiveMode:=dmUses;
                          if UnicodeUpStrCmp(CurrentWord,_ICON_DIRECTIVE)=0
                            then DirectiveMode:=dmIcon;

                          if DirectiveMode=dmNone
                            then Result:=LC_UNKNOWN_DIRECTIVE;
                        end
                       else
                       //       ,     
                        begin
                           case DirectiveMode of
                             dmFile:     Result:=CompileHardcodedFile(script,ReparsePath(script,CurrentWord));
                             dmFileSize: Result:=CompileFileSize(script,ReparsePath(script,CurrentWord));
                             dmDB:       begin
                                          case CurrentWord[1] of
                                            '''':  begin
                                                    Result:=CleanAnsiString(CurrentWord,CurrentAnsiWord);
                                                    if Result=LC_OK
                                                      then Result:=CompileDBAnsiString(script,CurrentAnsiWord)
                                                   end;
                                            '"':   begin
                                                    Result:=CleanUnicodeString(CurrentWord);
                                                    if Result=LC_OK
                                                      then Result:=CompileDBUnicodeString(script,CurrentWord)
                                                   end;
                                            else   Result:=CompileByte(script,CurrentWord);
                                          end;
                                         end;
                             dmDW:       begin
                                          if CurrentWord[1]='"' then
                                            begin
                                              Result:=CleanUnicodeString(CurrentWord);
                                              if Result=LC_OK
                                                then Result:=CompileDBUnicodeString(script,CurrentWord)
                                            end
                                            else Result:=CompileWord(script,CurrentWord);
                                         end;
                             dmDD:       Result:=CompileDWord(script,CurrentWord);
                             dmDI:       Result:=CompileInt(script,CurrentWord);
                             dmDF:       Result:=CompileFloat(script,CurrentWord);
                             dmInclude,
                             dmUses:     Result:=LC_OK;
                             dmIcon:     begin
                                          script.NewIcon:=ReparsePath(script,CurrentWord);
                                          Result:=LC_OK;
                                         end;
                           end;//case
                        end;//else
                     end;//lcmDirective
     end;//case
    end;
 until (WordEnd=0) or (result<>0);
 if Failure
  then Result:=LC_SYNTAX_ERROR;
 if Result=CMD_OK then
  begin
   //         -   " "
   if (LineCompilingMode in [lcmCommand,lcmParameters]) and (WordCounter<>MaxWordCount)
    then Result:=LC_OUT_OF_PARAMETERS;
   //     ,   -   "  "
   if (LineCompilingMode=lcmSectionName) and (WordCounter<>MaxWordCount)
    then Result:=LC_INVALID_SECTION_DESCRIPTION;
  end;
end;

procedure SecondPass(var scr:TRawScript);
var idx:Cardinal;
begin
  with scr do
  begin
    ErrLine:=NO_ERROR_LINE;
    ErrModule:=SCRIPT_MAIN_MODULE;
    if ScriptText.Count=0 then
      begin
        CompilingResult:=LC_OK;
        ErrLine:=NO_ERROR_LINE;
        Exit;
      end;//if ScriptText.Count=0
    for idx:=0 to ScriptText.Count-1 do
      begin
        NowCompiling:=idx;
        CompilingResult:=SimpleCompileLine(ScriptText.Strings[idx],scr);
        if CompilingResult<>LC_OK then
          begin
            ErrLineByLineNo(scr,idx);
            Break;
          end//if CompilingResult<>LC_OK
      end;//for
  end;//with
end;

function FindLabelInSection(const scr:TRawScript; SectionIndex:Integer; const LabelName:String; ModuleIndex:Integer; IsSectionGlobal:Boolean):Integer;
//             
var i:Integer;
begin
  Result:=-1;
  with scr.Sections[SectionIndex] do
    begin
      for i:=0 to Labels.Count-1 do
        begin
          //        
          if (LabelsInfo[i].ModuleIndex=ModuleIndex) and (IsSectionGlobal=LabelsInfo[i].SectionGlobal) then
            begin
              //       
              if StrIComp(PChar(Labels.Strings[i]),PChar(LabelName))=0 then
                begin
                  Result:=i;
                  Break;
                end;
            end;
        end;//for
    end;//with
end;

function FindLabelInAllSections(const scr:TRawScript; const LabelName:String; ModuleIndex:Integer; var SectionIndex:Integer):Integer;
//           .      
var i:Integer;
begin
  SectionIndex:=-1;
  Result:=-1;
  for i:=Low(scr.Sections) to High(scr.Sections) do
    begin
      Result:=FindLabelInSection(scr,i,LabelName,ModuleIndex,True);
      if Result>=0 then
        begin
          SectionIndex:=i;
          Break;
        end;//if Result>=0
    end;//for
end;

procedure CreateJTS(var scr:TRawScript);
type TCardinalArray=array[1..$FFFFFFF] of Cardinal;

  procedure BuildSectionOffsets;
  var i:Integer;
      HeaderOfs:Cardinal;
  begin
  with scr do
    begin
      //   
      HeaderOfs:=8+4;
      for i:=Low(Sections) to High(Sections) do
        begin
          if Sections[i].Size>0
            then Inc(HeaderOfs,Length(Sections[i].SectionName)+1+4+4);
        end;//for
      //    
      for i:=Low(Sections) to High(Sections) do
        begin
          Sections[i].OffsetInFile:=HeaderOfs;
          Inc(HeaderOfs,Sections[i].Size);
        end;
    end;//with
  end;//procedure

  function CompileLabels(var pos_in_jts:PCardinal):Integer;
  var i:Integer;
      InSection,PosInSection:Integer;
      true_lbl_offset,where_label:Cardinal;
      write_offset_to:PByte;
  begin
    Result:=LC_OK;
    for i:=0 to scr.TrueLabelRefCount-1 do
      begin
        with scr.LabelReferences[i] do
          begin
            //           ,  ( )   
            if Section=SECTION_ANY
              then PosInSection:=FindLabelInAllSections(scr,Name,Module,InSection)
              else
                begin
                  //      
                  InSection:=FindSectionByName(@scr,Section);
                  if InSection=-1
                    then Result:=LC_SECTION_NOT_EXISTS //      
                    else PosInSection:=FindLabelInSection(scr,InSection,Name,Module,False); //     
                end;
            //          
            if Result=LC_OK then
              begin
                if PosInSection=-1
                  then Result:=LC_LABEL_NOT_FOUND //   
                  else //         
                    begin
                      //  ,   ,   
                      where_label:=scr.Sections[SectionWhere].OffsetInFile;
                      //     
                      with scr.Sections[InSection] do
                        begin
                          //     = +   
                          true_lbl_offset:=OffsetInFile+LabelsInfo[PosInSection].OffsetInSection;

                          //   ,     
                          write_offset_to:=scr.Sections[SectionWhere].Ptr;
                          Inc(write_offset_to,LabelOffsetInFile);
                        end;
                      //    _ _  JTS
                      pos_in_jts^:=where_label+LabelOffsetInFile;
                      //          ,   
                      PCardinal(write_offset_to)^:=true_lbl_offset;
                      Inc(pos_in_jts);
                    end;
              end;//if Result=LC_OK
            if Result<>LC_OK then
              begin
                //                
                ErrLineByLineNo(scr,LineNumber);
                //    
                Break;
              end;//if Result<>LC_OK
          end;//with
      end;//for
  end;//function CompileLabelsInSection

var i:Integer;
    jts_sect_num:Integer;
    pos_in_jts:PCardinal;
begin
with scr do
begin
  // ,     
  if TrueLabelRefCount>0 then
  begin
    jts_sect_num:=FindSectionByName(@scr,JUMP_SECTION_NAME);
    //      
    with Sections[jts_sect_num] do
    begin
      AddSomeBytesToBlock(Ptr,Size,BufSize,@i,SizeOf(Cardinal)*TrueLabelRefCount);
      //       
      pos_in_jts:=Ptr;
    end;

    //   JTS
    BuildSectionOffsets;

    //    
    scr.CompilingResult:=CompileLabels(pos_in_jts);
  end;//if ref_count>0
end;//with
end;

procedure CompileScript(scr:PRawScript);
begin
 SecondPass(scr^);
 if scr^.CompilingResult=LC_OK then
    begin
      scr^.ErrLine:=NO_ERROR_LINE;
      CreateJTS(scr^);
    end;
end;


function SaveCompiledScript(scr:PRawScript; fname:PWideChar; JoinWithExisting:Boolean):Boolean;
const MagicWord:Int64=6867198141719203423; // '_NeoS0M_';
var hfile,TotalLen:Cardinal;
    crc64:Int64;

  function SaveBlock(Data:Pointer; len:Cardinal):Boolean;
  //     CRC64,  
  var res:Cardinal;
  begin
    if len>0 then
    begin
      Result:=WriteFile(hfile,Data^,len,res,nil);
      Inc(TotalLen,res);
      if Result and scr^.CRC64_Needed
        then CRC64_Block(Data,len,crc64);
    end
    else result:=True;
  end;

  function SaveSectionList:boolean;
  var i,TrueSectionCount:Integer;
      sn:AnsiString;
      p:PCardinal;
  begin
  with scr^ do
  begin
    //   ,     
    sn:=#0#0#0#0;
    TrueSectionCount:=0;
    Result:=True;
    for i:=Low(Sections) to High(Sections) do
    begin
      with Sections[i] do
      begin
        { DONE :      -      }
        if result and (Size>0) then
        begin
          Inc(TrueSectionCount);
          sn:=sn+SectionName+#0#0#0#0#0#0#0#0#0;
          //    :  ASCIIZ-    4      
          p:=@sn[Length(sn)-7];
          p^:=Flags;
          Inc(p);
          p^:=Size;
        end;//if result and (Size>0)
      end;//with
    end;//for

    //        
    p:=@sn[1];
    p^:=TrueSectionCount;
    //         ;  CRC64,  
    result:=SaveBlock(@sn[1],Length(sn));
  end;//with
  end;

  function SaveSections:Boolean;
  var i:Integer;
  begin
  Result:=True;
  with scr^ do
  begin
   for i:=Low(Sections) to High(Sections) do
    begin
     // .. SaveBlock     ,     
     result:=SaveBlock(Sections[i].Ptr,Sections[i].Size);
     if not Result then Break;
    end;
  end;//with
  end;

begin
 if JoinWithExisting then
  begin
   hfile:=CreateFile(fname,GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ or FILE_SHARE_WRITE,
                     nil,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
   SetFilePointer(hfile,0,nil,FILE_END);
  end
 else
  begin
   hfile:=CreateFile(fname,GENERIC_READ or GENERIC_WRITE,FILE_SHARE_READ or FILE_SHARE_WRITE,
                     nil,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
   SetEndOfFile(hfile);
  end;
 if hfile=0 then Result:=False
 else
  begin
   crc64:=not 0; //     CRC64
   TotalLen:=0;
   //      
   result:=SaveBlock(@MagicWord,SizeOf(MagicWord));
   if Result then Result:=SaveSectionList;
   if Result then Result:=SaveSections;
   //  CRC64
   if Result then Result:=SaveBlock(@crc64,SizeOf(crc64));
   //   
   if Result then Result:=SaveBlock(@TotalLen,SizeOf(TotalLen));
   CloseHandle(hfile);
  end;
end;

function ParseIncludeLine(const s:String; var sl:TStringList; var IsInclude:Boolean):Integer;
//       . 
var i:Integer;
    CurrentWord:Cardinal;
    MayBeDirective:Boolean;
    w:String;
    flag,IsUses:Boolean;
begin
  Result:=LC_OK;
  sl:=nil;
  MayBeDirective:=False;
  for i:=1 to Length(s) do
  begin
    //    
    if s[i]<>' ' then
      begin
        //    ,     &
        MayBeDirective:=s[i]=PREFIX_DIRECTIVE;
        Break;
      end;//if s[i]<>' '
    end;//for i:=1 to Length(s)

  //      -    
  if MayBeDirective then
    begin
      CurrentWord:=1;
      //  flag -      
      w:=GetFirstWord(s,CurrentWord,flag);
      Inc(CurrentWord);

      //    -        
      if not flag then
        begin
          //    -    ,...
          IsInclude:=UnicodeUpStrCmp(w,_INCLUDE_DIRECTIVE)=0;
          if not IsInclude
            then IsUses:=UnicodeUpStrCmp(w,_USES_DIRECTIVE)=0
            else IsUses:=False;
          //    &INCLUDE  &USES -    
          if IsInclude or IsUses then
            begin
              sl:=TStringList.Create;
              repeat
                //   
                w:=GetFirstWord(s,CurrentWord,flag);
                Inc(CurrentWord);
                //       -  
                if flag then
                  begin
                    //   -    , ..    
                    FreeAndNil(sl);
                    Result:=LC_SYNTAX_ERROR;
                  end;
                //   ,     -   
                if (w='') or (w=';') or flag then
                  begin
                    CurrentWord:=0;
                    Continue;
                  end;
                sl.Add(w);
              until (CurrentWord=0) or (Result<>LC_OK);
            end//if IsInclude or IsUses
        end//if not flag
    end;//if MayBeDirective
end;

function LoadScript(const FName:UnicodeString; var RawScript:PRawScript; InsertIntoLine:Integer; var LoadedLines:Integer):Integer;
label _Exit;
var i,j,loaded_lines,old_len,lines_in_subscript,LastLine:Integer;
    ModuleIndex:Integer;
    tmp_list:TStringList;
    w:String;
    IsInclude,UseModule:Boolean;
begin
  loaded_lines:=0;
  with RawScript^ do
  begin
    //     
    if not Assigned(RawScript.ScriptText) then
      begin
        ErrLine:=NO_ERROR_LINE;
        ErrModule:=SCRIPT_MAIN_MODULE;
        ScriptText:=TStringList.Create;
        ScriptModules:=TStringList.Create;
        ScriptModules.CaseSensitive:=False;
      end;//if not Assigned(...

      Result:=LC_OK;
      //  ()        
      if InsertIntoLine=LOAD_MAIN_MODULE then
        begin
          try
            ScriptText.LoadFromFile(FName);
            loaded_lines:=ScriptText.Count;
            SetLength(ScriptLines,loaded_lines);
            for i:=0 to loaded_lines-1 do
              begin
                ScriptLines[i].ModuleIndex:=SCRIPT_MAIN_MODULE;
                ScriptLines[i].LineNumber:=i+1;
              end;
          except
            Result:=LC_CANT_READ_SCRIPT_FILE;
            RawScript.ErrLine:=NO_ERROR_LINE;
            RawScript.ErrModule:=SCRIPT_MAIN_MODULE;
          end;
        end
      else
        begin
          //          
          ModuleIndex:=ScriptModules.Add(FName);
          tmp_list:=TStringList.Create;
          try
            tmp_list.LoadFromFile(FName);
            loaded_lines:=tmp_list.Count;
            //       
            for i:=loaded_lines-1 downto 0
              do ScriptText.Insert(InsertIntoLine,tmp_list.Strings[i]);
            //     
            old_len:=Length(ScriptLines);
            SetLength(ScriptLines,old_len+loaded_lines);
            //          
            Move(ScriptLines[InsertIntoLine],ScriptLines[InsertIntoLine+loaded_lines],SizeOf(TOneScriptLine)*(old_len-InsertIntoLine));
            //         
            for i:=InsertIntoLine to InsertIntoLine+loaded_lines-1 do
              begin
                ScriptLines[i].ModuleIndex:=ModuleIndex;
                ScriptLines[i].LineNumber:=i-InsertIntoLine+1;
              end;
          except
            Result:=LC_CANT_READ_SCRIPT_FILE;
          end;
          //        
          tmp_list.Free;
        end;
      LoadedLines:=loaded_lines;

      //        ,
      //   ,   .
      if Result=LC_OK then
        begin
          //         
          if InsertIntoLine=LOAD_MAIN_MODULE
            then i:=0
            else i:=InsertIntoLine;
          LastLine:=i+loaded_lines-1;
          while i<=LastLine do
            begin
              if ParseIncludeLine(ScriptText.Strings[i],tmp_list,IsInclude)<>LC_OK
                then Break;
              if tmp_list<>nil then
                begin
                  //      ,       
                  ScriptText.Strings[i]:='';
                  // ,       
                  UseModule:=True;
                  for j:=0 to tmp_list.Count-1 do
                    begin
                      w:=ReparsePath(RawScript^,tmp_list.Strings[j]);
                      if not IsInclude
                        then UseModule:=ScriptModules.IndexOf(w)<0;
                      if UseModule then
                        begin
                          //        
                          lines_in_subscript:=0;
                          Result:=LoadScript(w,RawScript,i,lines_in_subscript);
                          if Result=LC_OK then
                            begin
                              //     
                              Inc(LoadedLines,lines_in_subscript);
                              //      (      )
                              Inc(i,lines_in_subscript);
                              //      
                              Inc(LastLine,lines_in_subscript);
                            end
                          else
                            begin
                              //   ,   
                              FreeAndNil(tmp_list);
                              //               
                              if not RawScript^.StopErrorHandling then
                                begin
                                  StopErrorHandling:=True;
                                  ErrLineByLineNo(RawScript^,i);
                                end;
                              goto _Exit;
                            end;
                        end;//if UseModule
                    end;//for j
                  //   ,   
                  FreeAndNil(tmp_list);
                end
            else Inc(i);
//              MessageBox(0,PChar(ScriptText.Strings[i]),'-',0);
            end;//while
_Exit:
        end;
  end;//with
end;

function CompileNeoS0MToAny(InFName,OutFName:PWideChar;RawScript:PRawScript;JoinWithExistsng:Boolean;var ErrLine:Cardinal; ErrModuleBuffer:PChar; IconAllowed:Boolean):Cardinal;
var InternalRawScript:TRawScript;
    ErrModuleName:String;
    LinesLoaded:Integer;
begin
  if RawScript=nil
    then RawScript:=@InternalRawScript;
  //     (      )
  RawScript^.ScriptPath:=ExtractFilePath(INFname);
  if (Pos(':\',RawScript^.ScriptPath)=0) and (Pos(':/',RawScript^.ScriptPath)=0) and (Pos('\\',RawScript^.ScriptPath)=0)
    then RawScript^.ScriptPath:='';
  BeforeCompiling(RawScript);
  //    
  Result:=LoadScript(String(InFName),RawScript,LOAD_MAIN_MODULE,LinesLoaded);

  //     
  if Result=LC_OK then
    begin
      //  
      CompileScript(RawScript);
      Result:=RawScript^.CompilingResult;

      if Result=LC_OK then
        begin
          //     .BIN     
          if (RawScript.NewIcon='') or (not IconAllowed)
            then Result:=LC_OK
            else if ReplaceIcon(String(OutFName),RawScript.NewIcon)
              then Result:=LC_OK
              else Result:=LC_CANT_REPLACE_ICON;
          if Result=LC_OK
            then Result:=IfThen(SaveCompiledScript(RawScript,OutFName,JoinWithExistsng),LC_OK,LC_CANT_WRITE_COMPILED_FILE);
        end;
    end;//if Result=LC_OK

    //    
    ErrLine:=RawScript^.ErrLine;
    //   MAX_PATH   ,      NIL
    if Assigned(ErrModuleBuffer) then
      begin
        //       -     
        if RawScript.ErrModule=SCRIPT_MAIN_MODULE
          then ErrModuleBuffer^:=#0
          else
            begin
              ErrModuleName:=RawScript.ScriptModules.Strings[RawScript.ErrModule];
              //     MAX_PATH,  
              if Length(ErrModuleName)>MAX_PATH
                then ErrModuleName:=''+Copy(ErrModuleName,Length(ErrModuleName)-MAX_PATH+2,MAX_PATH-1);
              Move(ErrModuleName[1],ErrModuleBuffer^,(Length(ErrModuleName)+1)*SizeOf(Char))
            end;
      end;//if Assigned(ErrModuleBuffer)

  //    ,         
  AfterCompiling(RawScript);
end;

function CompileNeoS0MFile(InFName,OutFName:PWideChar;RawScript:PRawScript;var ErrLine:Cardinal; ErrModuleBuffer:PChar):Cardinal;
begin
  Result:=CompileNeoS0MToAny(InFName,OutFName,RawScript,False,ErrLine,ErrModuleBuffer,False);
end;

function CompileNeoS0MEXE(InFName,OutFName:PWideChar;RawScript:PRawScript;var ErrLine:Cardinal; ErrModuleBuffer:PChar):Cardinal;
begin
  Result:=CompileNeoS0MToAny(InFName,OutFName,RawScript,True,ErrLine,ErrModuleBuffer,True);
end;

function InitializeCompiler(CmdList:PWideChar):Integer;
var cl_dir:UnicodeString;
    l:Integer;
begin
 Result:=__LoadCmdList(CmdList,BuiltInCommands);
 if Result=0 then
  begin
    cl_dir:=ExtractFilePath(CmdList);
    if cl_dir[Length(cl_dir)]<>'\'
      then cl_dir:='\';
    if LoadDescrList(cl_dir+'CompilerKeywords.txt',Keywords) then
    begin
      l:=Length(Keywords);
      if l>=1
        then _KEYWORD_SECTION:=Keywords[0];
      if l>=2
        then _KEYWORD_CODE:=Keywords[1];
      if l>=3
        then _KEYWORD_DATA:=Keywords[2];
      if l>=4
        then _FILE_DIRECTIVE:=Keywords[3];
      if l>=5
        then _FILESIZE_DIRECTIVE:=Keywords[4];
      if l>=6
        then _DB_DIRECTIVE:=Keywords[5];
      if l>=7
        then _DW_DIRECTIVE:=Keywords[6];
      if l>=8
        then _DD_DIRECTIVE:=Keywords[7];
      if l>=9
        then _DI_DIRECTIVE:=Keywords[8];
      if l>=10
        then _DF_DIRECTIVE:=Keywords[9];
      if l>=11
        then _INCLUDE_DIRECTIVE:=Keywords[10];
      if l>=12
        then _USES_DIRECTIVE:=Keywords[11];
      if l>=13
        then _ICON_DIRECTIVE:=Keywords[12];
   end
   else Result:=-1;
  end;
end;

procedure FinalizeCompiler;
begin
  __DeinitializeCmdInfo(BuiltInCommands);
end;

var AnsiStr:AnsiString;
    UnicodeStr:UnicodeString;
    c:AnsiChar;
    i:Integer;
begin
  // Unicode-,     ANSI
  SetLength(AnsiStr,255);
  for c:=#1 to #255
    do AnsiStr[Ord(c)]:=c;
  UnicodeStr:=String(AnsiStr);
  for i:=1 to 255
    do CharInfoTable[UnicodeStr[i]]:=CharInfoTable[UnicodeStr[i]] or CHAR_IS_ANSI;

end.
