unit Hashes;

interface

type MD5_CTX=packed record
      state:  packed array [0..3] of Cardinal;
      count:  packed array[0..1] of Cardinal;
      buffer: packed array [0..63] of Byte;
      //    
      Digest: array [0..15] of Byte;
     end;

procedure CRC32_Init(var res:Cardinal);
procedure CRC32_Calc(data:Pointer; len:Cardinal; var res:Cardinal);
procedure MD5_Init(var Context:MD5_CTX);
procedure MD5_Update(var Context:MD5_CTX;Input:Pointer;InputLen:Cardinal);
procedure MD5_Final(var Context:MD5_CTX);

implementation

var CRC32TableReady:Boolean=False;
    CRC32Table:packed array[0..255] of Cardinal;

{$IFDEF Win32}
  //      WiseByte
procedure CRC32_Init(var res:Cardinal);
begin
  if not CRC32TableReady then
    begin
      asm
         push ecx
         push edx
         xor eax,eax
         dec al // EAX=255
       @@10:
         mov edx,eax
         mov cl,8
       @@11:
         shr edx,1
         jnc @@1
         xor edx,0EDB88320h
       @@1:
         dec cl
         jnz @@11
         mov [OFFSET CRC32Table+eax*4],edx
         dec eax
         jns @@10
         pop edx
         pop ecx
      end;//asm
      CRC32TableReady:=True;
    end;
  res:=$FFFFFFFF;
end;
{$ENDIF}

{$IFDEF Win64}
procedure CRC32_Init(var res:Cardinal);
var i,j,c:Cardinal;
begin
  if not CRC32TableReady then
    begin
      for i:=0 to 255 do
      begin
        c:=i;
        for j:=1 to 8 do
          if odd(c)
            then c:=(c shr 1) xor $EDB88320
            else c:=(c shr 1);
        CRC32Table[i]:=c;
      end;//for
    end;
  res:=$FFFFFFFF
end;
{$ENDIF}

procedure CRC32_Calc(data:Pointer; len:Cardinal; var res:Cardinal);
  type _BA=packed array[1..$7FFFFFF0] of Byte;
       _PBA=^_BA;
  var i:Cardinal;
begin
  for i:=1 to len
   do res:=(res shr 8) xor CRC32Table[_PBA(data)^[i] xor (res and $FF)];
end;

//  MD5 -    RFC    Delphi

const S11=7;
      S12=12;
      S13=17;
      S14=22;
      S21=5;
      S22=9;
      S23=14;
      S24=20;
      S31=4;
      S32=11;
      S33=16;
      S34=23;
      S41=6;
      S42=10;
      S43=15;
      S44=21;

//     TByteArray16=array [0..15] of Byte;
type TByteArray=packed array[0..$7FFFFF00] of Byte;
     PByteArray=^TByteArray;
     TCardinalArray4=packed array [0..3] of Cardinal;
     PCardinalArray4=^TCardinalArray4;


var Padding:packed array [0..63] of Byte;
    PaddingReady:Boolean=False;

function F(x,y,z:Cardinal):Cardinal;inline;
begin
  F:=(x and y) or ((not x) and z);
end;

function G(x,y,z:Cardinal):Cardinal;inline;
begin
  G:=(x and z) or (y and (not z));
end;

function H(x,y,z:Cardinal): Cardinal;inline;
begin
  H:=x xor y xor z;
end;

function I(x,y,z:Cardinal):Cardinal;inline;
begin
  I:=y xor (x or (not z));
end;

function ROTATE_LEFT(x,n:Cardinal):Cardinal;inline;
begin
  ROTATE_LEFT:=(x shl n) or (x shr (32-n));
end;

procedure FF(var a:Cardinal;b,c,d,x,s,ac:Cardinal);
begin
  Inc(a,F(b, c, d)+x+ac);
  a:=ROTATE_LEFT(a,s);
  Inc(a,b);
end;

procedure GG(var a:Cardinal; b,c,d,x,s,ac:Cardinal);
begin
  Inc(a,G(b,c,d)+x+ac);
  a:=ROTATE_LEFT(a,s);
  Inc(a,b);
end;

procedure HH(var a:Cardinal;b,c,d,x,s,ac:Cardinal);
begin
  Inc(a,H(b,c,d)+x+ac);
  a:=ROTATE_LEFT(a,s);
  Inc(a,b);
end;

procedure II(var a:Cardinal;b,c,d,x,s,ac:Cardinal);
begin
  Inc(a,I(b,c,d)+x+ac);
  a:=ROTATE_LEFT(a,s);
  Inc(a,b);
end;

procedure MD5Decode(Output:PCardinalArray4;input:PByteArray;Len:Cardinal);
var
  i,j:Cardinal;
begin
  i:=0;j:=0;
  while j < Len do
  begin
    Output[i] := Cardinal(input[j]) or (Cardinal(input[j+1]) shl 8) or
      (Cardinal(input[j+2]) shl 16) or (Cardinal(input[j+3]) shl 24);
    Inc(i);
    Inc(j,4);
  end;
end;

procedure MD5_Init(var Context:MD5_CTX);
begin
 if not PaddingReady then
  begin
    //   Padding
    FillChar(Padding[1],63,0);
    Padding[0]:=$80;
    PaddingReady:=True;
  end;
  with Context do
   begin
    count[0]:=0;
    count[1]:=0;
    state[0]:=$67452301;
    state[1]:=$EFCDAB89;
    state[2]:=$98BADCFE;
    state[3]:=$10325476;
   end;
end;

procedure MD5Transform(State:PCardinalArray4; Buffer: PByteArray);
var
  a,b,c,d:Cardinal;
  x:packed array[0..15] of Cardinal;
begin
  a:=State[0];
  b:=State[1];
  c:=State[2];
  d:=State[3];

  MD5Decode(@x,Buffer,64);

  FF(a,b,c,d,x[0],S11,$D76AA478);
  FF(d,a,b,c,x[1],S12,$E8C7B756);
  FF(c,d,a,b,x[2],S13,$242070DB);
  FF(b,c,d,a,x[3],S14,$C1BDCEEE);
  FF(a,b,c,d,x[4],S11,$F57C0FAF);
  FF(d,a,b,c,x[5],S12,$4787C62A);
  FF(c,d,a,b,x[6],S13,$A8304613);
  FF(b,c,d,a,x[7],S14,$FD469501);
  FF(a,b,c,d,x[8],S11,$698098D8);
  FF(d,a,b,c,x[9],S12,$8B44F7AF);
  FF(c,d,a,b,x[10],S13,$FFFF5BB1);
  FF(b,c,d,a,x[11],S14,$895CD7BE);
  FF(a,b,c,d,x[12],S11,$6B901122);
  FF(d,a,b,c,x[13],S12,$FD987193);
  FF(c,d,a,b,x[14],S13,$A679438E);
  FF(b,c,d,a,x[15],S14,$49B40821);

  GG(a,b,c,d,x[1],S21,$F61E2562);
  GG(d,a,b,c,x[6],S22,$C040B340);
  GG(c,d,a,b,x[11],S23,$265E5A51);
  GG(b,c,d,a,x[0],S24,$E9B6C7AA);
  GG(a,b,c,d,x[5],S21,$D62F105D);
  GG(d,a,b,c,x[10],S22,$2441453);
  GG(c,d,a,b,x[15],S23,$D8A1E681);
  GG(b,c,d,a,x[4],S24,$E7D3FBC8);
  GG(a,b,c,d,x[9],S21,$21E1CDE6);
  GG(d,a,b,c,x[14],S22,$C33707D6);
  GG(c,d,a,b,x[3],S23,$F4D50D87);

  GG(b,c,d,a,x[8],S24,$455A14ED);
  GG(a,b,c,d,x[13],S21,$A9E3E905);
  GG(d,a,b,c,x[2],S22,$FCEFA3F8);
  GG(c,d,a,b,x[7],S23,$676F02D9);
  GG(b,c,d,a,x[12],S24,$8D2A4C8A);

  HH(a,b,c,d,x[5],S31,$FFFA3942);
  HH(d,a,b,c,x[8],S32,$8771F681);
  HH(c,d,a,b,x[11],S33,$6D9D6122);
  HH(b,c,d,a,x[14],S34,$FDE5380C);
  HH(a,b,c,d,x[1],S31,$A4BEEA44);
  HH(d,a,b,c,x[4],S32,$4BDECFA9);
  HH(c,d,a,b,x[7],S33,$F6BB4B60);
  HH(b,c,d,a,x[10],S34,$BEBFBC70);
  HH(a,b,c,d,x[13],S31,$289B7EC6);
  HH(d,a,b,c,x[0],S32,$EAA127FA);
  HH(c,d,a,b,x[3],S33,$D4EF3085);
  HH(b,c,d,a,x[6],S34,$4881D05);
  HH(a,b,c,d,x[9],S31,$D9D4D039);
  HH(d,a,b,c,x[12],S32,$E6DB99E5);
  HH(c,d,a,b,x[15],S33,$1FA27CF8);
  HH(b,c,d,a,x[2],S34,$C4AC5665);

  II(a,b,c,d,x[0],S41,$F4292244);
  II(d,a,b,c,x[7],S42,$432AFF97);
  II(c,d,a,b,x[14],S43,$AB9423A7);
  II(b,c,d,a,x[5],S44,$FC93A039);
  II(a,b,c,d,x[12],S41,$655B59C3);
  II(d,a,b,c,x[3],S42,$8F0CCC92);
  II(c,d,a,b,x[10],S43,$FFEFF47D);
  II(b,c,d,a,x[1],S44,$85845DD1);
  II(a,b,c,d,x[8],S41,$6FA87E4F);
  II(d,a,b,c,x[15],S42,$FE2CE6E0);
  II(c,d,a,b,x[6],S43,$A3014314);
  II(b,c,d,a,x[13],S44,$4E0811A1);
  II(a,b,c,d,x[4],S41,$F7537E82);
  II(d,a,b,c,x[11],S42,$BD3AF235);
  II(c,d,a,b,x[2],S43,$2AD7D2BB);
  II(b,c,d,a,x[9],S44,$EB86D391);

  Inc(State[0],a);
  Inc(State[1],b);
  Inc(State[2],c);
  Inc(State[3],d);

  FillChar(x,SizeOf(x),0);
end;

procedure MD5_Update(var Context:MD5_CTX;Input:Pointer;InputLen:Cardinal);
var
  i,index,partLen:Cardinal;
begin
with Context do
begin
  index:=(count[0] shr 3) and $3F;

  Inc(count[0],InputLen shl 3);
  if count[0]<(InputLen shl 3) then Inc(count[1]);
  Inc(count[1],InputLen shr 29);

  partLen:=64-index;

  if inputLen>=partLen then
  begin
    Move(Input^,buffer[index],PartLen);
    MD5Transform(@Context.state, @Context.buffer);
    i:=partLen;
    while i+63<inputLen do
    begin
      MD5Transform(@state,@(PByteArray(Input)[i]));
      Inc(i,64);
    end;
    index:=0;
  end
  else i:=0;
  Move(PByteArray(Input)[i],buffer[index],inputLen-i);
end;//with
end;

procedure MD5Encode(Output:PByteArray;Input:PCardinalArray4;Len:Cardinal);
var
  i,j:Cardinal;
begin
  i:=0;j:=0;
  while j<Len do
  begin
    output[j]:=Byte(input[i] and $FF);
    output[j+1]:=Byte((input[i] shr 8) and $FF);
    output[j+2]:=Byte((input[i] shr 16) and $FF);
    output[j+3]:=Byte((input[i] shr 24) and $FF);
    Inc(j,4);
    Inc(i);
  end;
end;

procedure MD5_Final(var Context:MD5_CTX);
var
  bits: packed array[0..7] of Byte;
  index,padLen:Cardinal;
begin
  MD5Encode(@bits, PCardinalArray4(@Context.count),8);
  index:=Cardinal((Context.count[0] shr 3) and $3F);
  if index<56
    then padLen:=56-index
    else padLen:=120-index;
  MD5_Update(Context,@PADDING,padLen);
  MD5_Update(Context,@Bits,8);
  MD5Encode(@Context.Digest,@Context.state,16);
//  FillChar(Context,SizeOf(Context),0);
end;


end.
