unit Highlighting;

{$I SynEdit.inc}

interface

uses Windows,
  Graphics,
  SynEditTypes,
  SynEditHighlighter,
  SynUnicode,
  SysUtils,
  Classes;

const SYNS_NeoS0M='Neo Sign 0f Misery internal language';
      SYNS_NeoS0M_Friendly='Neo Sign 0f Misery internal language friendly';
      SYNS_FilterNeoS0M='Neo Sign 0f Misery scripts (*.nsm)|*.nsm';

type

  TtkTokenKind = (tkNull, // C.  -   
      tkNone,             //   - 
      tkSpace,            // 
      tkKeyword,          //  
      tkCommandName,      //  
      tkLabel,            // 
      tkLabelRef,         //   
      tkComment,          // 
      tkHexDump,          //      
      tkAnsiString,       // ANSI-
      tkUnicodeString,    // Unicode-
      tkGlobalVariable,   //  
      tkLocalVariable,    //  
      tkIntNumber,        //  
      tkHexNumber,        //  16- 
      tkFloatNumber,      //    
      tkDate,             // 
      tkDefault,          // "  " ($)
      tkParamEnd,         //    (.)
      tkDirective,        // 
      tkAnsiChar,         // ANSI-
      tkUnicodeChar,      // Unicode-
      tkError);           // 

const
  TokenSYNS:array [TtkTokenKind] of String=
  ( 'NULL token',
    'No highlighting',
    'Space',
    'Keyword',
    'Command',
    'Label',
    'Label reference',
    'Comment',
    'Binary Block as HEX dump',
    'Binary Block as ANSI string',
    'Unicode string',
    'Reference to global variable',
    'Reference to local variable',
    'Integer number',
    'Integer number as HEX',
    'Float-point number',
    'Date-time',
    '"Use default value" parameter',
    'End of parameter list',
    'Directive',
    'ANSI Char',
    'Unicode Char',
    'Syntax error'
  );


type
  TSyntaxHighlightAttributes=record
    Background:TColor;
    Foreground:TColor;
    Style:TFontStyles;
  end;

  TRangeState = (rsAnyText, rsBraceComment, rsCStyleComment, rsString);

  TProcTableProc = procedure of object;

  PIdentFuncTableFunc = ^TIdentFuncTableFunc;
  TIdentFuncTableFunc = function (Index: Integer): TtkTokenKind of object;

type
  TSynNeoS0MSyn = class(TSynCustomHighlighter)
  private
    FirstWord:Boolean;
    fProcTable: array[#0..#255] of TProcTableProc;

    fRange: TRangeState;
    fTokenID: TtkTokenKind;

    //  
    Attributes:array [TtkTokenKind] of TSynHighlighterAttributes;

    function GetKeywordTokenID(start,len:Integer;crc32:Cardinal):TtkTokenKind;

    procedure UnknownProc;
    procedure CommandNameProc;
    procedure NullProc;
    procedure SpaceProc;
    procedure SemiColonProc;
    procedure CRProc;
    procedure LFProc;
    procedure PowerProc;
    procedure NumberProc;
    procedure BucksProc;
    procedure DotProc;
    procedure SignProc;
    procedure PercentProc;
    procedure AsteriskProc;
    procedure ATProc;
    procedure AmpersandProc;
    procedure UnderlineProc;
    procedure QuoteProc;
    procedure DoubleQuoteProc;
    procedure WOWProc;
    procedure HashProc;
    procedure TildaProc;

    function AnyQuoteProc(c:WideChar):Boolean;
    function WhichNumber:TtkTokenKind;
    function IntNumber:TtkTokenKind;
    function VarRef:TtkTokenKind;
    function DateProc:TtkTokenKind;
  protected
    function GetSampleSource: UnicodeString; override;
    function IsFilterStored: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    class function GetFriendlyLanguageName: UnicodeString; override;
    class function GetLanguageName: string; override;
    function GetRange: Pointer; override;
    procedure ResetRange; override;
    procedure SetRange(Value: Pointer); override;
    function GetDefaultAttribute(Index: Integer): TSynHighlighterAttributes; override;
    function GetEol: Boolean; override;
    function GetKeyWords(TokenKind: Integer): UnicodeString; override;
    function GetTokenID: TtkTokenKind;
    function GetTokenAttribute: TSynHighlighterAttributes; override;
    function GetTokenKind: Integer; override;
    function IsIdentChar(AChar: WideChar): Boolean; override;
    function IsWordBreakChar(AChar: WideChar): Boolean; override;
    procedure Next; override;
    procedure DoSetLine(const Value: UnicodeString; LineNumber: Integer);override;

    procedure MakeMethodTables;
    procedure SetStyleByToken(tk:TtkTokenKind; st:TSyntaxHighlightAttributes);
  published


{    property CommentAttri: TSynHighlighterAttributes read fCommentAttri write fCommentAttri;
    property KeyWordAttri: TSynHighlighterAttributes read fKeywordAttri write fKeywordAttri;
    property SpaceAttri: TSynHighlighterAttributes read fSpaceAttri write fSpaceAttri;

    property NoneAttri:TSynHighlighterAttributes
              read fNoneAttri write fNoneAttri;
    property LabelAttri:TSynHighlighterAttributes
              read fLabelAttri write fLabelAttri;
    property LabelRefAttri:TSynHighlighterAttributes
              read fLabelRefAttri write fLabelRefAttri; }

  end;

//     
procedure AddHighlightedWord(const w:UnicodeString; hl:TtkTokenKind);
//     
procedure PrepareHighlightedWords;

implementation

uses NC_Chars, SynEditStrConst;

//    
type TDoubleCharID=WORD;
     TOneHighlightedWord=record
      // ,        
      //     
      SpecCRC32:Cardinal;
      WordItself:UnicodeString;
      HighlightType:TtkTokenKind;
     end;
     //      
     TSameLenHW=array of TOneHighlightedWord;
     //     .
     //        
     THighlightedWords=array of TSameLenHW;



resourcestring
  SYNS_FilterTest = 'All files (*.*)|*.*';
  SYNS_LangTest = 'Test';
  SYNS_FriendlyLangTest = 'Test';
  SYNS_AttrTest = 'Test';
  SYNS_FriendlyAttrTest = 'Test';

const
  // as this language is case-insensitive keywords *must* be in lowercase
{  KeyWords: array[0..2] of UnicodeString = (
    'hello', 'synedit', 'world'
  );}

  KeyIndices: array[0..2] of Integer = (
    0, 2, 1
  );

var HighlightedWords:THighlightedWords;
    CRC32Table:packed array [0..255] of Cardinal;


function GetSpecCRC32(const w:WideString):Cardinal;
var i:Integer;
    c:WideChar;
    b1,b2:PByte;
begin
  Result:=$FFFFFFFF;
  b1:=@c;
  b2:=b1+1;
  for i:=1 to Length(w) do
    begin
      c:=UpperCaseTable[w[i]];
      Result:=(Result shr 8) xor CRC32Table[b1^ xor (Result and $FF)];
      Result:=(Result shr 8) xor CRC32Table[b2^ xor (Result and $FF)];
    end;
end;

procedure QuickSortList(var arr:TSameLenHW; L,R:Integer);
var I, J, P: Integer;
    tmp:TOneHighlightedWord;

    function SCompare(i,j:integer):integer;
    begin
     if arr[i].SpecCRC32>arr[j].SpecCRC32
      then result:=1
      else begin
            if arr[i].SpecCRC32<arr[j].SpecCRC32
             then result:=-1
             else result:=0;
           end;
    end;

    begin
    if L<R then
    begin
      repeat
        I := L;
        J := R;
        P := (L + R) shr 1;
        repeat
          while SCompare(I, P) < 0 do Inc(I);
          while SCompare(J, P) > 0 do Dec(J);
          if I <= J then
          begin
            tmp:=arr[i];
            arr[i]:=arr[j];
            arr[j]:=tmp;
    //        ExchangeItems(I, J);
            if P = I then
              P := J
            else if P = J then
              P := I;
            Inc(I);
            Dec(J);
          end;
        until I > J;
        if L < J then QuickSortList(arr, L, J);
        L := I;
      until I >= R;
    end;
    end;

procedure PrepareHighlightedWords;
var i:Integer;
begin
  for i:=0 to Length(HighlightedWords)-1
    do QuickSortList(HighlightedWords[i],0,Length(HighlightedWords[i])-1);
end;

//       
procedure TSynNeoS0MSyn.MakeMethodTables;
var
  i: WideChar;
begin
  for i:=#1 to #32
    do fProcTable[i]:=SpaceProc;
  for i:=#33 to #255
    do fProcTable[i]:=UnknownProc;
  for i:='0' to '9'
    do fProcTable[i]:=NumberProc;
  fProcTable[#0]:=NullProc;
  fProcTable[#10]:=LFProc;
  fProcTable[#13]:=CRProc;
  fProcTable['$']:=BucksProc;
  fProcTable[';']:=SemiColonProc;
  fProcTable['^']:=PowerProc;
  fProcTable['.']:=DotProc;
  fProcTable['+']:=SignProc;
  fProcTable['-']:=SignProc;
  fProcTable['%']:=PercentProc;
  fProcTable['*']:=AsteriskProc;
  fProcTable['@']:=ATProc;
  fProcTable['&']:=AmpersandProc;
  fProcTable['_']:=UnderlineProc;
  fProcTable['''']:=QuoteProc;
  fProcTable['"']:=DoubleQuoteProc;
  fProcTable['!']:=WOWProc;
  fProcTable['#']:=HashProc;
  fProcTable['~']:=TildaProc;
end;

procedure AddHighlightedWord(const w:UnicodeString; hl:TtkTokenKind);
var len,l_2:Integer;
begin
  len:=Length(w);
  //         -  
  if Length(HighlightedWords)<len
    then SetLength(HighlightedWords,len);
  //         len
  l_2:=Length(HighlightedWords[len-1]);
  SetLength(HighlightedWords[len-1],l_2+1);
  //      ,    
  with HighlightedWords[len-1][l_2] do
    begin
      SpecCRC32:=GetSpecCRC32(w);
      HighlightType:=hl;
      // ,    
      WordItself:=w;
      CharUpperBuffW(PWideChar(WordItself),len);
    end;
end;

procedure TSynNeoS0MSyn.SpaceProc;
begin
  Inc(Run);
  fTokenID:=tkSpace;
  while ((CharInfoTable[FLine[Run]] and CHAR_IS_TRUE_SPACE)<>0) and (not IsLineEnd(Run))
    do Inc(Run);
end;

procedure TSynNeoS0MSyn.TildaProc;
var cc:Integer;
  function _CharCount:Integer;
  var t_run:Integer;
  begin
    t_run:=Run;
    Result:=0;
    while (CharInfoTable[FLine[t_run]] and CHAR_IS_TRUE_SPACE=0) and (not IsLineEnd(t_run)) do
      begin
        Inc(t_run);
        Inc(Result);
      end;
  end;//function
begin
  if FirstWord
    then CommandNameProc
    else
      begin
        fTokenID:=tkError;
        Inc(Run);
        cc:=_CharCount;
        //   ,      / 
        case cc of
          1: if CharInfoTable[FLine[Run]] and CHAR_IS_ANSI<>0
              then fTokenID:=tkAnsiChar;
          2: if FLine[Run]='~'
              then fTokenID:=tkUnicodeChar;
          //         -  
        end;//case
        Inc(Run,cc);
      end;//else
end;

procedure TSynNeoS0MSyn.NullProc;
begin
  fTokenID := tkNull;
  inc(Run);
end;

procedure TSynNeoS0MSyn.SemiColonProc;
begin
  Inc(Run);
  if (CharInfoTable[FLine[Run]] and CHAR_IS_TRUE_SPACE)<>0
    then begin
          // ,     - 
          fTokenID:=tkComment;
          while CharInfoTable[FLine[Run]] and CHAR_IS_END_OF_LINE=0
            do Inc(Run);
         end
    else begin
          // ,     - 
          fTokenID:=tkError;
          while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0
            do Inc(Run);
         end;
end;

procedure TSynNeoS0MSyn.PowerProc;
var HasChars:Boolean;
begin
  Inc(Run);
  HasChars:=False;
  while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0 do
    begin
      HasChars:=True;
      Inc(Run);
    end;//while
  if HasChars
    then if FirstWord
          then FTokenID:=tkLabel
          else FTokenID:=tkLabelRef
    else FTokenID:=tkError;
  if FirstWord
    then FirstWord:=False;
end;

procedure TSynNeoS0MSyn.CRProc;
begin
  fTokenID:=tkSpace;
  FirstWord:=True;
  if FLine[Run+1]=#10
    then Inc(Run,2)
    else Inc(Run);

  fTokenID := tkSpace;
  inc(Run);
  if fLine[Run] = #10 then
    inc(Run);
end;

procedure TSynNeoS0MSyn.LFProc;
begin
  fTokenID:=tkSpace;
  Inc(Run);
  FirstWord:=True;
end;

procedure TSynNeoS0MSyn.DoSetLine(const Value: UnicodeString;
  LineNumber: Integer);
begin
  Run:=0;
  FirstWord:=True;
  inherited;
end;

constructor TSynNeoS0MSyn.Create(AOwner: TComponent);
var t:TtkTokenKind;
begin
  inherited Create(AOwner);
  fCaseSensitive := False;

  for t:=Low(TtkTokenKind) to High(TtkTokenKind) do
    begin
      Attributes[t]:=TSynHighlighterAttributes.Create(TokenSYNS[t],TokenSYNS[t]);
      AddAttribute(Attributes[t]);
    end;

  MakeMethodTables;

  SetAttributesOnChange(DefHighlightChange);
  FirstWord:=True;
  fDefaultFilter := SYNS_FilterTest;
  fRange := rsAnyText;
end;

function TSynNeoS0MSyn.GetKeywordTokenID(start,len:Integer;crc32:Cardinal):TtkTokenKind;
  //           
  function CompareWord(wordidx:Integer):Boolean;
    var i:Integer;
    begin
      Result:=True;
      with HighlightedWords[len-1][wordidx] do
      begin
      for i:=1 to len do
        begin
          if WordItself[i]<>UpperCaseTable[FLine[start+i-1]]
            then begin Result:=False; Break; end;
        end;//for
      end;//with
    end;
  //        . 
  function FindWordInArray:TtkTokenKind;
  var l,r,m:Integer;
      tmp:Cardinal;
  begin
    Result:=tkNone;
    l:=0; r:=Length(HighlightedWords[len-1])-1;
    while l<=r do
      begin
        //  
        m:=(l+r) shr 1;
        tmp:=HighlightedWords[len-1][m].SpecCRC32;
        // crc32  () 
        if crc32>tmp then l:=m+1
          // crc32  () 
          else if crc32<tmp then r:=m-1
            // crc32 c!
            else Break;
      end;//while
    if l<=r //    - crc32 
      then
        begin
          // crc32 ,    . 
          if CompareWord(m)
            //     
            then Result:=HighlightedWords[len-1][m].HighlightType
            else
              begin
                // crc32     
                //     
                for l:=m+1 to Length(HighlightedWords[len-1])-1 do
                  begin
                    if crc32=HighlightedWords[len-1][l].SpecCRC32
                      then if CompareWord(l)
                            then begin
                                  //   crc32,   
                                  Result:=HighlightedWords[len-1][l].HighlightType;
                                  Break;
                                 end
                      //  crc32  
                      else Break;
                  end;
                //     
                for l:=0 to m-1 do
                  begin
                    if crc32=HighlightedWords[len-1][l].SpecCRC32
                      then if CompareWord(l)
                            then begin
                                  //   crc32,   
                                  Result:=HighlightedWords[len-1][l].HighlightType;
                                  Break;
                                 end
                      //  crc32  
                      else Break;
                  end;
              end;
        end;//if l<=r
  end;

begin
  //       
  if (len=0) or (len>Length(HighlightedWords))
    then Result:=tkNone
    else begin
           //         
           if Length(HighlightedWords[len-1])>0
            then Result:=FindWordInArray
            else Result:=tkNone;
         end;
end;

procedure TSynNeoS0MSyn.CommandNameProc;
var SpecCRC32:Cardinal;
    _Start,_Len:Integer;
    b1,b2:PByte;
    c:WideChar;
begin
  _Start:=Run;
  SpecCRC32:=$FFFFFFFF;
  b1:=@c;
  b2:=b1+1;
  while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0 do
    begin
      //    CRC32
      c:=UpperCaseTable[FLine[Run]];
      SpecCRC32:=(SpecCRC32 shr 8) xor CRC32Table[b1^ xor (SpecCRC32 and $FF)];
      SpecCRC32:=(SpecCRC32 shr 8) xor CRC32Table[b2^ xor (SpecCRC32 and $FF)];
      Inc(Run);
    end;
  _Len:=Run-_Start;
  FTokenID:=GetKeywordTokenID(_Start,_Len,SpecCRC32);
  if FirstWord
    then FirstWord:=False;
end;

function TSynNeoS0MSyn.DateProc:TtkTokenKind;
var HasComma,HasDotAfterComma:Boolean;
    DotsBeforeComma,ColonsAfterComma,NumbersBefore:Integer;
begin
  Result:=tkDate;
  HasComma:=False;
  HasDotAfterComma:=False;
  DotsBeforeComma:=0;
  ColonsAfterComma:=0;
  NumbersBefore:=0;
  while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0 do
    begin
     if Result<>tkError then
     begin
      case FLine[Run] of
        '.': begin
              if NumbersBefore=0 //     ? ( )
                then Result:=tkError
                else NumbersBefore:=0;
              //      
              if HasComma
                then begin
                      //         
                      //       .  - 
                      if (ColonsAfterComma=2) and (not HasDotAfterComma)
                        then HasDotAfterComma:=True
                        else Result:=tkError;
                     end
                else
                  begin
                    Inc(DotsBeforeComma);
                    //        
                    if DotsBeforeComma>2 then result:=tkError;
                  end;
             end;
        ',': begin
              if NumbersBefore=0 //     ? ( )
                then Result:=tkError
                else NumbersBefore:=0;
              if HasComma
                then Result:=tkError //    
                else HasComma:=True;
             end;
        ':': begin
              if NumbersBefore=0 //     ? ( )
                then Result:=tkError
                else NumbersBefore:=0;
              if HasComma //    
                then begin
                      Inc(ColonsAfterComma);
                      //        
                      if ColonsAfterComma>2 then Result:=tkError;
                     end
                else Result:=tkError;
             end;
        '0'..'9': Inc(NumbersBefore);
        else Result:=tkError;
      end;//case
     end;
     Inc(Run);
    end;
  if Result=tkDate // .  -    
    then if NumbersBefore=0
          then Result:=tkError;
end;

function TSynNeoS0MSyn.WhichNumber:TtkTokenKind;
var HasDecimalDot, //   
    HasExponent, //  
    HasNumbersInMantissa, //    
//    HasNumbersInMantissaAfterDot, //       
    HasNumbersInExponent, //    
    HasSignInMantissa, //    
    HasSignInExponent:Boolean; //    
begin
  Result:=tkIntNumber;
  HasDecimalDot:=False;
  HasExponent:=False;
  HasNumbersInMantissa:=False;
  HasNumbersInExponent:=False;
//  HasNumbersInMantissaAfterDot:=False;
  HasSignInMantissa:=False;
  HasSignInExponent:=False;
  while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0 do //      
    begin
      if Result<>tkError then
      begin
      case FLine[Run] of
        '.': begin
               if HasDecimalDot or HasExponent
               //         
               //    
                then Result:=tkError
                else Result:=tkFloatNumber;
               HasDecimalDot:=True;
             end;
        'e','E':
             begin
               if HasExponent or (not HasNumbersInMantissa)
               //           
                then Result:=tkError
                else Result:=tkFloatNumber;
               HasExponent:=True;
             end;
        '0'..'9':
             begin
               if HasExponent then //   
                begin
                  HasNumbersInExponent:=True;
                end
               else //   
                begin
                  HasNumbersInMantissa:=True;
{                  if HasDecimalDot //       ?
                    then HasNumbersInMantissaAfterDot:=True;}
                end;
             end;
        '+','-':
             begin
               if HasExponent then //   
                begin
                 //           
                  if HasNumbersInExponent or HasSignInExponent
                     then Result:=tkError;
                  HasSignInExponent:=True;
                end
               else //   
                begin
                 //        ,     
                  if HasNumbersInMantissa or HasDecimalDot or HasSignInMantissa
                     then Result:=tkError;
                  HasSignInMantissa:=True;
                end;
             end;
        //      
        else Result:=tkError;
      end;//case
      end;// if Result<>tkError
      Inc(Run);
    end;
  if (Result<>tkError) and HasExponent then
  //       
    begin
      if not HasNumbersInExponent then Result:=tkError;
    end;
end;

function TSynNeoS0MSyn.IntNumber:TtkTokenKind;
var HasSign,HasDigits,HasDigitsAfterSign,HexNumber:Boolean;
begin
  Result:=tkIntNumber;
  HasSign:=False;
  HasDigits:=False;
  HasDigitsAfterSign:=false;
  HexNumber:=False;
  while (CharInfoTable[FLine[Run]] and CHAR_IS_SPACE)=0 do // ,       
    begin
      if Result<>tkError then
      begin
      case FLine[Run] of
        '$': begin
              //         
               if HexNumber or HasDigits
                then Result:=tkError
                else begin Result:=tkHexNumber; HexNumber:=True; end;
             end;
        '0'..'9':
             begin
              HasDigits:=True;
              if HasSign then HasDigitsAfterSign:=True;
             end;
        'a'..'f','A'..'F':
             begin
               //    16- 
               if HexNumber then
                begin
                  HasDigits:=True;
                  if HasSign then HasDigitsAfterSign:=True;
                end
               else Result:=tkError;
             end;
        '+','-':
             begin
              //      ,    
              if HexNumber or HasDigits or HasSign then Result:=tkError;
              HasSign:=True;
             end;
        //      
        else Result:=tkError;
      end;//case
      end;// if Result<>tkError
      Inc(Run);
    end;
  //  ,     
  if HasSign then
    begin
      if not HasDigitsAfterSign
        then Result:=tkError;
    end;
end;

procedure TSynNeoS0MSyn.BucksProc;
begin
  if FirstWord then
  // $ -    .
  //      
   begin
      CommandNameProc;
   end
  else
    begin
      //  $      - ,   "  "
      if (CharInfoTable[FLine[Run+1]] and CHAR_IS_SPACE<>0) or (IsLineEnd(Run))
        then
          begin
            FTokenID:=tkDefault;
            Inc(Run);
          end
        else
          //  $ -    ,     HEX-
          begin
            FTokenID:=IntNumber;
          end;//else
    end;//else
end;

procedure TSynNeoS0MSyn.NumberProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      FTokenID:=WhichNumber;
    end;
end;

procedure TSynNeoS0MSyn.SignProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      //   +  -     
      FTokenID:=IntNumber;
    end;
end;

procedure TSynNeoS0MSyn.PercentProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      Inc(Run);
      if CharInfoTable[FLine[Run]] and CHAR_IS_SPACE<>0
        then FTokenID:=tkError    // %     
        else FTokenID:=IntNumber; //   %     
    end;
end;

procedure TSynNeoS0MSyn.DotProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      if CharInfoTable[FLine[Run+1]] and CHAR_IS_SPACE<>0 then
        begin
          FTokenID:=tkParamEnd;
          Inc(Run);
        end
        //    FP-  
        else FTokenID:=WhichNumber;
    end;//else
end;

procedure TSynNeoS0MSyn.AsteriskProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      Inc(Run);
      if CharInfoTable[FLine[Run]] and CHAR_IS_SPACE<>0
        then FTokenID:=tkError    // *     
        else
          begin //   %       
            if WhichNumber=tkError
              then FTokenID:=tkError
              else FTokenID:=tkFloatNumber;
          end;
    end;
end;

procedure TSynNeoS0MSyn.ATProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      Inc(Run);
      if CharInfoTable[FLine[Run]] and CHAR_IS_SPACE<>0
        then FTokenID:=tkError // @     
        else FTokenID:=VarRef; //   @      
    end;
end;

procedure TSynNeoS0MSyn.AmpersandProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
      //  ,   - , ,   ,     
      if FTokenID=tkDirective then
      begin
      //    
      while CharInfoTable[FLine[Run]] and CHAR_IS_END_OF_LINE=0 do
        begin
          if FLine[Run]=';' then
            begin
              if CharInfoTable[FLine[Run-1]] and CHAR_IS_SPACE<>0
                then Break;
            end;
          Inc(Run);
        end;//while
      end//if FTokenID=tkDirective
    end
  else
    begin
      //       
      FTokenID:=tkError;
      while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0
        do Inc(Run);
    end;//else
end;

procedure TSynNeoS0MSyn.UnderlineProc;
begin
  if FirstWord then
    begin
      CommandNameProc;
    end
  else
    begin
      if CharInfoTable[FLine[Run+1]] and CHAR_IS_SPACE<>0 then
        // _     
        begin
          FTokenID:=tkError;
          Inc(Run);
        end
        else FTokenID:=VarRef; //   _     
    end;
end;

function TSynNeoS0MSyn.AnyQuoteProc(c:WideChar):Boolean;
var HasOpenQuote,HasCloseQuote,HasUnpairedQuotes:Boolean;
begin
  HasOpenQuote:=False;
  HasCloseQuote:=False;
  HasUnpairedQuotes:=False;
  while (CharInfoTable[FLine[Run]] and CHAR_IS_END_OF_LINE=0) and (not HasCloseQuote) do
    begin
    if FLine[Run]=c then //  
      begin
        if HasOpenQuote then //    ?
          begin
            if CharInfoTable[FLine[Run+1]] and CHAR_IS_SPACE<>0 //    ?
              then HasCloseQuote:=True
              else
                begin
                  if FLine[Run+1]=c //    ?
                  then Inc(Run)
                  else HasUnpairedQuotes:=True;
                end;
          end//if HasOpenQuote
          else HasOpenQuote:=True; //   - 
      end;//FLine[Run]=c
      Inc(Run);
    end;//while
  //   -  ,    
  result:=HasOpenQuote and HasCloseQuote and (not HasUnpairedQuotes);
end;

procedure TSynNeoS0MSyn.QuoteProc;
begin
  if FirstWord then CommandNameProc
  else if AnyQuoteProc('''')
        then fTokenID:=tkAnsiString
        else fTokenID:=tkError;
end;

procedure TSynNeoS0MSyn.DoubleQuoteProc;
begin
  if FirstWord then CommandNameProc
  else if AnyQuoteProc('"')
        then fTokenID:=tkUnicodeString
        else fTokenID:=tkError;
end;

procedure TSynNeoS0MSyn.WOWProc;
var HasClose:Boolean;
    HexDigitCount:Cardinal;
    CharInfo:Byte;
begin
  HasClose:=False;
  HexDigitCount:=0;
  if FirstWord then CommandNameProc
  else
    begin
      Inc(Run);
      while (CharInfoTable[FLine[Run]] and CHAR_IS_END_OF_LINE=0) and (not HasClose) do
        begin
          CharInfo:=CharInfoTable[FLine[Run]];
          if FLine[Run]='!'
            then HasClose:=CharInfoTable[FLine[Run+1]] and CHAR_IS_SPACE<>0
            else if CharInfo and CHAR_IS_HEX_DIGIT<>0
                  then Inc(HexDigitCount);
          Inc(Run);
        end;//while
    end;
  if HasClose and (not Odd(HexDigitCount))
    then fTokenID:=tkHexDump
    else fTokenID:=tkError;
end;

procedure TSynNeoS0MSyn.HashProc;
begin
  if FirstWord
    then CommandNameProc
    else begin
            Inc(Run);
            fTokenID:=DateProc;
         end;
end;

procedure TSynNeoS0MSyn.UnknownProc;
var WasFirst:Boolean;
begin
  if Run=0
    then FirstWord:=True;
  WasFirst:=FirstWord;
  CommandNameProc;
  if WasFirst then //    
    begin
{      //    , ,    
      if fTokenID=tkNone
        then fTokenID:=tkError; }
    end
  else //      -      
    begin
      if fTokenID=tkCommandName
        then fTokenID:=tkError;
    end;
{  inc(Run);
  fTokenID := tkNone;}
end;

procedure TSynNeoS0MSyn.Next;
var c:WideChar;
begin
  fTokenPos := Run;
  c:=fLine[Run];
  if c>#255
    then UnknownProc
    else fProcTable[c];
  inherited;
end;

function TSynNeoS0MSyn.GetDefaultAttribute(Index: Integer): TSynHighLighterAttributes;
begin
  Result:=Attributes[tkNone];
end;

function TSynNeoS0MSyn.GetEol: Boolean;
begin
  Result:=Run=fLineLen + 1;
end;

function TSynNeoS0MSyn.GetKeyWords(TokenKind: Integer): UnicodeString;
begin
  Result :=
    'Hello,SynEdit,World';
end;

function TSynNeoS0MSyn.GetTokenID: TtkTokenKind;
begin
  Result := fTokenId;
end;

function TSynNeoS0MSyn.VarRef:TtkTokenKind;
var HasSign,HasDigits,HasDigitsAfterSign,HexNumber:Boolean;
begin
  Result:=tkGlobalVariable;
  HasSign:=False;
  HasDigits:=False;
  HasDigitsAfterSign:=false;
  HexNumber:=False;
  while CharInfoTable[FLine[Run]] and CHAR_IS_SPACE=0 do //      
    begin
      if Result<>tkError then
      begin
      case FLine[Run] of
        '_': begin
             //      
             if HexNumber or HasDigits or HasSign
              then Result:=tkError
              else Result:=tkLocalVariable;
             end;
        '$': begin
              //         
               if HexNumber or HasDigits
                then Result:=tkError
                else HexNumber:=True;
             end;
        '0'..'9':
             begin
              HasDigits:=True;
              if HasSign then HasDigitsAfterSign:=True;
             end;
        'a'..'f','A'..'F':
             begin
               //    16- 
               if HexNumber then
                begin
                  HasDigits:=True;
                  if HasSign then HasDigitsAfterSign:=True;
                end
               else Result:=tkError;
             end;
        '+':
             begin
              //  +     ,    
              if HexNumber or HasDigits or HasSign then Result:=tkError;
              HasSign:=True;
             end;
        //      
        else Result:=tkError;
      end;//case
      end;// if Result<>tkError
      Inc(Run);
    end;
  //  ,     
  if HasSign then
    begin
      if not HasDigitsAfterSign then Result:=tkError;
    end;
end;

function TSynNeoS0MSyn.GetTokenAttribute: TSynHighLighterAttributes;
begin
  Result:=Attributes[fTokenID];
end;

function TSynNeoS0MSyn.GetTokenKind: Integer;
begin
  Result := Ord(fTokenId);
end;

function TSynNeoS0MSyn.IsIdentChar(AChar: WideChar): Boolean;
begin
  Result:=CharInfoTable[AChar] and CHAR_IS_IDENT<>0;
end;

function TSynNeoS0MSyn.IsWordBreakChar(AChar: WideChar): Boolean;
begin
  Result:=CharInfoTable[AChar] and CHAR_IS_WORD_BREAKING<>0;
end;

function TSynNeoS0MSyn.GetSampleSource: UnicodeString;
begin
  Result :=
    '; Neo Sign 0f Misery highlighting sample'#13#10#13#10+
    'GALLOC 10 .'#13#10+
    'SECTION DATA section_name'#13#10 +
    '&DB 01 02 03'#13#10 +
    'EXTERNAL_CALL "test_dll*test_func" $ 1 $2 ''3'' ~4 ~~5 *6 %7 !08! #10.10.2010:11:11:11.111';
end;

function TSynNeoS0MSyn.IsFilterStored: Boolean;
begin
  Result := fDefaultFilter <> SYNS_FilterTest;
end;

class function TSynNeoS0MSyn.GetFriendlyLanguageName: UnicodeString;
begin
  Result := SYNS_FriendlyLangTest;
end;

class function TSynNeoS0MSyn.GetLanguageName: string;
begin
  Result := SYNS_LangTest;
end;

procedure TSynNeoS0MSyn.ResetRange;
begin
  fRange := rsAnyText;
end;

procedure TSynNeoS0MSyn.SetRange(Value: Pointer);
begin
  fRange := TRangeState(Value);
end;

procedure TSynNeoS0MSyn.SetStyleByToken(tk:TtkTokenKind; st: TSyntaxHighlightAttributes);
begin
  with Attributes[tk] do
    begin
      Background:=st.Background;
      Foreground:=st.Foreground;
      Style:=st.Style;
    end;
end;

function TSynNeoS0MSyn.GetRange: Pointer;
begin
  Result := Pointer(fRange);
end;

initialization
  //      WiseByte
  asm
    push ecx
    push edx
    xor eax,eax
    dec al // EAX=255
   @@10:
    mov edx,eax
    mov cl,8
   @@11:
    shr edx,1
    jnc @@1
    xor edx,0EDB88320h
   @@1:
    dec cl
    jnz @@11
    mov [OFFSET CRC32Table+eax*4],edx
    dec eax
    jns @@10
    pop edx
    pop ecx
  end;//asm
end.

