unit NC_CommandEditor;

interface

uses
  Windows, Forms, Messages, Variants, SysUtils, Classes, Graphics, Controls,
  Dialogs, ExtCtrls, StdCtrls, Buttons, NC_ParseDescription, PNC_Param;

const _BTN_SIZE=28;

type TFrameInfo=record
        Frame:TfrmParam;
        MayBeDefault,MayBeEnd,MayNotExist:Boolean;
        PrevState:Integer; //  : 0 - , 1 -  ., 2 - 
     end;

type
  TCommandEditorForm = class(TForm)
    Panel1: TPanel;
    Label1: TLabel;
    edtComment: TEdit;
    pbCommandIcon: TPaintBox;
    LabelCommandName: TLabel;
    LabelCommand: TLabel;
    Label2: TLabel;
    edtSpacing: TEdit;
    sbSpacingUp: TSpeedButton;
    sbSpacingDown: TSpeedButton;
    Panel2: TPanel;
    btnOk: TBitBtn;
    ScrollBox1: TScrollBox;
    sbAddParameter: TSpeedButton;
    sbRemoveParameter: TSpeedButton;
    btnCancel: TBitBtn;
    procedure FormCreate(Sender: TObject);
    procedure pbCommandIconPaint(Sender: TObject);
    procedure Panel2Resize(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure sbAddParameterClick(Sender: TObject);
    procedure ScrollBox1Resize(Sender: TObject);
    procedure sbSpacingUpClick(Sender: TObject);
    procedure UpdateFrameControls(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure sbRemoveParameterClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    LimParamCount:Integer; //    
    MLimParamCount:Integer; //    
    CommandPosInList:Integer; //     
    Parameters:array of TFrameInfo;
    CommandParamsAsText:UnicodeString;

    function AddParamAllowed:Integer;
    function RemoveParamAllowed:Integer;
    function AddCommand(cmd_index: Cardinal): UnicodeString;
    procedure ShowCommandInfo(cmd_pos_in_list:Cardinal);
    function CreateParameters(cmd_pos_in_list:Cardinal):UnicodeString;
    procedure CleanAllFrames;

    function AddParamToList(const InputMethod: String; const _Caption:UnicodeString; const DataControls:String; const _Value,DefaultValue:String; Abilities:Integer): UnicodeString;
    function RemoveParamFromList:Integer;
    procedure UpdateWidths;
    procedure UpdateAllFramesVisibility;
  end;

var
  CommandEditorForm: TCommandEditorForm=nil;

implementation

uses NC_AllVars, NC_DescrList, NeoS0M_UNIFIED, NeoS0M_CmdList, NeoS0M_Math, NC_Main, NC_FrameCreator,
     NC_ParamConversion, PNC_ParamAny;

{$R *.dfm}

{ TCommandEditorForm }

function TCommandEditorForm.AddCommand(cmd_index: Cardinal): UnicodeString;
var cmd_line:UnicodeString;
begin
  //      
  CleanAllFrames;
  Caption:=_Adding_command;
  edtComment.Text:='';
  CommandPosInList:=Form_Main.GetCmdIndexByCode(cmd_index);
  ShowCommandInfo(CommandPosInList);
  cmd_line:=CreateParameters(CommandPosInList);
  if cmd_line='' then
    begin
      if ShowModal=mrOk then
        begin
          cmd_line:=Form_Main.CmdInfoList[CommandPosInList].CmdName+CommandParamsAsText;
          Result:=cmd_line;
        end
      else Result:='';
    end
  else //      
    begin
      Application.MessageBox(PWideChar(cmd_line),PWideChar(_Error_in_parameters),MB_ICONERROR);
      //   
      CleanAllFrames;
      Result:='';    
    end;
end;

function TCommandEditorForm.AddParamAllowed:Integer;
var i:Integer;
begin
  if Length(Parameters)>=LimParamCount
    then Result:=1
    else
    begin
      Result:=0;
      for i:=Low(Parameters) to High(Parameters) do
        begin
          //   -    "  "
          if Parameters[i].Frame.btnParamEnd.Down then
            begin
              Result:=2;
              Break;
            end;
        end;//for
    end;
end;

function FindParamClass(const s:String):TComponentClass;
var ofc:TOneFrameCreator;
begin
  Result:=TfrmParamAny;
  for ofc in FrameCreators
    do if ofc.InputMethod=s then
        begin
          Result:=ofc.FrameClass;
          Exit;
        end;
end;

function TCommandEditorForm.AddParamToList(const InputMethod: String; const _Caption:UnicodeString; const DataControls:String; const _Value,DefaultValue:String; Abilities:Integer): String;
// Abilities -  :
// 0  -    " "
// 1  -    
// 2  -   
var tmp:TfrmParam;
    t:Integer;
    param_idx:Integer;
begin
  Result:='';
  param_idx:=Length(Parameters);
  //  
  tmp:=FindParamClass(InputMethod).Create(Self) as TfrmParam;

  //    
  with tmp do
  begin
    Parent:=ScrollBox1;
    Visible:=False; //      ,   
    Name:='';
    Tag:=param_idx;
  end;

  //        Y- . 
  if param_idx=0
    then t:=0
    else t:=Parameters[param_idx-1].Frame.Top+Parameters[param_idx-1].Frame.Height;

  //    
  SetLength(Parameters,param_idx+1);
  with Parameters[param_idx] do
    begin
      Frame:=tmp;
      //     
      Frame.Top:=t;
      //    ""    (  )
      PrevState:=0; // .  - 
      MayBeDefault:=(Abilities and 1)<>0;
      MayBeEnd:=(Abilities and 2)<>0;
      MayNotExist:=(Abilities and 4)<>0;
    end;

  //         
  tmp.Prepare(_Caption,DataControls);
  //      
  UpdateWidths;
  //    
  tmp.Go(_Value,DefaultValue,Abilities);

  //       
  tmp.Visible:=True;
end;

function TCommandEditorForm.RemoveParamAllowed: Integer;
//      
//  ,   
var i:Integer;
begin
  Result:=Length(Parameters)-1;
  for i:=0 to Result do
  begin
    if Parameters[i].Frame.btnParamEnd.Down then
      begin
        Result:=i;
        Break;
      end;
  end;//for

  //       -  
  if Result<MLimParamCount then
    begin
      if Result<0 //     ,    
        then Result:=-1
        else Result:=-2; //      
    end;//if
end;

function TCommandEditorForm.RemoveParamFromList: Integer;
var i,new_l:Integer;
begin
  Result:=-1;
  new_l:=RemoveParamAllowed;
  case new_l of
    -1: Application.MessageBox(PWideChar(_ParameterListIsEmpty),PWideChar(_Error),MB_ICONERROR);
    -2: Application.MessageBox(PWideChar(_CantDeleteParameter),PWideChar(_Error),MB_ICONERROR);
    else
      begin
        for i:=Length(Parameters)-1 downto new_l
          do Parameters[i].Frame.Free;
        SetLength(Parameters,new_l);
        Result:=0;
      end;
  end;
end;

procedure TCommandEditorForm.btnOkClick(Sender: TObject);
var spacing,i,ec:Integer;
    var s,err,err_full:UnicodeString;
begin
  CommandParamsAsText:='';
  err_full:='';
  //   
  for i:=0 to Length(Parameters)-1 do
    begin
      err:='';
      s:=Parameters[i].Frame.GetData(err);
      CommandParamsAsText:=CommandParamsAsText+' '+s;
      if err<>''
        then err_full:=err_full+'['+IntToStr(i+1)+']:'+err+_CRLF;
      //          
      if s='.'
        then Break;
    end;
  //  ,  
  if edtComment.Text<>'' then
    begin
      Val(edtSpacing.Text,spacing,ec);
      if ec<>0 then
        begin
          if edtSpacing.Text<>_Auto
            then err_full:=err_full+_Invalid_Spacing+_CRLF
            else spacing:=1; //   ()   
        end
      else //   
        begin
          if spacing<1
            then err_full:=err_full+_Invalid_Spacing+_CRLF
        end;
    end
  else spacing:=0;
  //      ,       
  if err_full=''
    then
      begin
        { DONE :       }
        CommandParamsAsText:=CommandParamsAsText;
        if spacing>0
          then CommandParamsAsText:=CommandParamsAsText+StringOfChar(' ',spacing)+'; '+edtComment.Text;
        ModalResult:=mrOk
      end
    else begin
          Application.MessageBox(PWideChar(err_full),PWideChar(_Error_in_parameters),MB_ICONERROR);
          ModalResult:=mrNone;
         end;
end;

procedure TCommandEditorForm.CleanAllFrames;
var i:Integer;
begin
  for i:=High(Parameters) downto 0
    do FreeAndNil(Parameters[i].Frame);
  SetLength(Parameters,0);
end;

function TCommandEditorForm.CreateParameters(cmd_pos_in_list: Cardinal):UnicodeString;
var i,e:Integer;
    DataControls:String;
begin
Result:='';
with Form_Main.CmdInfoList[cmd_pos_in_list] do
 begin
    if MaxParamCount=UNKNOWN_PARAM_COUNT then
      begin
        e:=MinParamCount;
        MLimParamCount:=MinParamCount;
        LimParamCount:=MaxInt;
      end
    else
      begin
        e:=MaxParamCount;
        MLimParamCount:=MaxParamCount;
        LimParamCount:=MaxParamCount;
      end;
   //     
   { DONE :         }
   for i:=0 to e-1 do
    begin
      with Form_Main.CmdInfoList[cmd_pos_in_list].ParamInfo[i] do
        begin
          if DataControlsIndex=-1
            then DataControls:=''
            else DataControls:=TextDescriptions[DataControlsIndex];
          Result:=Result+AddParamToList(InputMethod,Form_Main.GetParamDescription(cmd_pos_in_list,i),
                  DataControls,DefaultValue,DefaultValue,Form_Main.GetParamAbilities(cmd_pos_in_list,i));
        end;
    end;
 end;//with
end;

procedure TCommandEditorForm.UpdateAllFramesVisibility;
// ,    
//     
var i:Integer;
    v:Boolean;
begin
  v:=True;
  for i:=Low(Parameters) to High(Parameters) do
  begin
    if Parameters[i].Frame.Visible<>v
      then Parameters[i].Frame.Visible:=v;
    if v
      then v:=not (Parameters[i].Frame.btnParamEnd.Down);
  end;//for
end;

procedure TCommandEditorForm.UpdateFrameControls(Sender: TObject);
begin
  //        
  UpdateAllFramesVisibility;
end;

procedure TCommandEditorForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  CleanAllFrames;
end;

procedure TCommandEditorForm.FormCreate(Sender: TObject);
begin
  LabelCommandName.Font.Size:=CMD_NAME_FONT_SIZE;
  LabelCommand.Font.Size:=CMD_NAME_FONT_SIZE;
  pbCommandIcon.ClientHeight:=Form_Main.pilCommandGroups48.Height;
  pbCommandIcon.ClientWidth:=Form_Main.pilCommandGroups48.Width;
  LJoin2Controls(sbSpacingUp,edtSpacing);
  LJoin2RectangularControl(edtSpacing,sbSpacingDown);
  edtSpacing.Text:=_Auto;
end;

procedure TCommandEditorForm.pbCommandIconPaint(Sender: TObject);
begin
 Form_Main.pilCommandGroups48.Draw(pbCommandIcon.Canvas,0,0,pbCommandIcon.Tag);
end;

procedure TCommandEditorForm.Panel2Resize(Sender: TObject);
var bw:Integer;
begin
  bw:=Panel2.ClientWidth div 2;
  btnOk.Width:=bw;
  btnCancel.Left:=bw;
  btnCancel.Width:=bw;
end;

procedure TCommandEditorForm.sbAddParameterClick(Sender: TObject);
var err_text:UnicodeString;
begin
  case AddParamAllowed of
    0: begin
//        with
          begin
{ DONE :         }
            //           ,
            // ,     ..     
            err_text:=AddParamToList('','','','','',0);
          end;

        if err_text<>'' then
          begin

          end;
        UpdateAllFramesVisibility;
       end;
    1: Application.MessageBox(PWideChar(_TooManyParameters),PWideChar(_Error),MB_ICONERROR);
    2: Application.MessageBox(PWideChar(_NoParamsAfterDot),PWideChar(_Error),MB_ICONERROR);
  end;//case
end;

procedure TCommandEditorForm.sbRemoveParameterClick(Sender: TObject);
begin
  RemoveParamFromList;
end;

procedure TCommandEditorForm.sbSpacingUpClick(Sender: TObject);
var v,err:Integer;
begin
 Val(edtSpacing.Text,v,err);
 if err>0
  then
    begin
      if (edtSpacing.Text=_Auto) and ((Sender as TComponent).Tag=1)
        then edtSpacing.Text:='1'
        else edtSpacing.Text:=_Auto
    end
  else
    begin
      if (Sender as TComponent).Tag=1
        then Inc(v)
        else Dec(v);
      if v<1
        then edtSpacing.Text:=_Auto
        else edtSpacing.Text:=IntToStr(v);
    end;
 
end;

procedure TCommandEditorForm.ScrollBox1Resize(Sender: TObject);
begin
  UpdateWidths;
end;

procedure TCommandEditorForm.ShowCommandInfo(cmd_pos_in_list: Cardinal);
begin
  with Form_Main.CmdInfoList[cmd_pos_in_list] do
    begin
      LabelCommandName.Caption:=TextDescriptions[DescrIdx];
      LabelCommand.Caption:=CmdName;
      pbCommandIcon.Tag:=CmdGroupCode;
      pbCommandIcon.Repaint;
    end;
end;

procedure TCommandEditorForm.UpdateWidths;
var i,w:Integer;
begin
  w:=ScrollBox1.ClientWidth;
  for i:=Low(Parameters) to High(Parameters) do
    begin
      if Assigned(Parameters[i].Frame) then
        begin
          with Parameters[i].Frame
            do if Width<>w
                then Width:=w;
        end;//if Assigned...
    end;//for
end;

end.
