unit NC_Options;

interface

uses
  Windows, Messages, Forms, Controls, StdCtrls, Classes, Buttons, Dialogs,
  ComCtrls, ExtCtrls, Graphics, PngBitBtn, NC_ColorSettings,
  PngImageList, Vcl.ImgList;

type TNeoS0MFiles=array [1..7] of String;

type
  TForm_Options = class(TForm)
    btnOk: TBitBtn;
    OpenDialog1: TOpenDialog;
    pgcOptions: TPageControl;
    tsPaths: TTabSheet;
    lblCompilerDLL: TLabel;
    lblKernelDLL: TLabel;
    lblKernelEXE: TLabel;
    lblCmdDescr: TLabel;
    lblCmdHelp: TLabel;
    tsInterface: TTabSheet;
    edtMRUCount: TEdit;
    lblMRUCount: TLabel;
    lblFiles: TLabel;
    lblTextDescr: TLabel;
    lblGroups: TLabel;
    chkUseAutoParameters: TCheckBox;
    chkVisualEditor: TCheckBox;
    tsSyntaxHighlighting: TTabSheet;
    lbSyntaxElements: TListBox;
    btnChooseFont: TPngBitBtn;
    cbColorSchema: TComboBox;
    lblColorSchema: TLabel;
    cbBold: TCheckBox;
    cbItalic: TCheckBox;
    cbStrikeout: TCheckBox;
    cbUnderline: TCheckBox;
    btnTextColor: TButton;
    btnBgColor: TButton;
    btnLineNumbersBg: TButton;
    btnLineNumbers: TButton;
    btnGradientStart: TButton;
    btnGradientEnd: TButton;
    cbGradient: TCheckBox;
    btnDelimiter: TButton;
    cdColorSelector: TColorDialog;
    pbDelimiter: TPaintBox;
    pbLineNumbers: TPaintBox;
    pbLineNumbersBg: TPaintBox;
    pbGradientStart: TPaintBox;
    pbGradientEnd: TPaintBox;
    pbTextColor: TPaintBox;
    pbBgColor: TPaintBox;
    GroupBox1: TGroupBox;
    imgSample: TImage;
    fdFontSelector: TFontDialog;
    btnSaveSchema: TPngBitBtn;
    btnLoadSchema: TPngBitBtn;
    SaveDialog1: TSaveDialog;
    btnFindInAppFolder: TPngBitBtn;
    btnFindInAnyFolder: TPngBitBtn;
    gbCommandPanel: TGroupBox;
    CmdPosRadioGroup: TRadioGroup;
    rgCmdSortOrderByName: TRadioGroup;
    cbBigIcons: TCheckBox;
    cbBigButtons: TCheckBox;
    lblPathToIWS: TLabel;
    gbFilterSettigs: TGroupBox;
    cbIgnoreGroupsInFilteredList: TCheckBox;
    cbImmediateFilter: TCheckBox;
    cbFilterByCmdName: TCheckBox;
    cbCmdFilterByFirstChars: TCheckBox;
    cbFilterByDescription: TCheckBox;
    cbDescrFilterByFirstChars: TCheckBox;
    cbGroupIconsInFilteredList: TCheckBox;
    cbQuickInfoOnAutoComplete: TCheckBox;
    edtCompilerDLL: TButtonedEdit;
    edtKernelDLL: TButtonedEdit;
    edtKernelEXE: TButtonedEdit;
    edtCmdHelp: TButtonedEdit;
    edtCmdDescr: TButtonedEdit;
    edtTextDescr: TButtonedEdit;
    edtGroups: TButtonedEdit;
    edtPathToIWS: TButtonedEdit;
    procedure FormShow(Sender: TObject);
    procedure btnOkClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure SelectFile(Sender: TObject);
    procedure chkVisualEditorClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure pbDelimiterPaint(Sender: TObject);
    procedure pbLineNumbersPaint(Sender: TObject);
    procedure pbGradientStartPaint(Sender: TObject);
    procedure pbGradientEndPaint(Sender: TObject);
    procedure cbGradientClick(Sender: TObject);
    procedure pbTextColorPaint(Sender: TObject);
    procedure lbSyntaxElementsClick(Sender: TObject);
    procedure pbBgColorPaint(Sender: TObject);
    procedure pbLineNumbersBgPaint(Sender: TObject);
    procedure cbColorSchemaChange(Sender: TObject);
    procedure btnChooseFontClick(Sender: TObject);
    procedure btnDelimiterClick(Sender: TObject);
    procedure btnLineNumbersClick(Sender: TObject);
    procedure btnLineNumbersBgClick(Sender: TObject);
    procedure btnGradientStartClick(Sender: TObject);
    procedure btnGradientEndClick(Sender: TObject);
    procedure ChooseSyntaxItemColor(Sender: TObject);
    procedure ChangeFontStyle(Sender: TObject);
    procedure btnSaveSchemaClick(Sender: TObject);
    procedure btnLoadSchemaClick(Sender: TObject);
    procedure btnFindInAnyFolderClick(Sender: TObject);
    procedure btnFindInAppFolderClick(Sender: TObject);
    procedure cbFilterByCmdNameClick(Sender: TObject);
    procedure cbFilterByDescriptionClick(Sender: TObject);
  private
    { Private declarations }
    ColorSettings:TColorSettings;
    EditorFontName:TFontName;
    EditorFontSize:Integer;
    InUpdateFlag:Boolean;

  public
    { Public declarations }
    procedure ChooseColor(var cl:TColor);
    procedure JoinColorBox(b:TButton; pb:TPaintBox);

    procedure ShowColorSchema;
    procedure ShowSelectedColorFromSchema;
    procedure UserColorSchemaInCombo;
    procedure NeoS0MFilesSearch(const Folder:String);
  end;

var
  Form_Options: TForm_Options;
  _DLLFilter:UnicodeString=' |*.dll| |*.*';
  _EXEFilter:UnicodeString=' |*.exe| |*.*';
  _CHMFilter:UnicodeString=' HTML-|*.chm| |*.*';
  _CMDFilter:UnicodeString='  |*.cmd| |*.*';
  _TXTFilter:UnicodeString=' |*.txt| |*.*';
  _NSHFilter:UnicodeString='    (*.nsh)|*.nsh';
  _InvalidMaxMRU:UnicodeString='    ';

implementation

uses SysUtils, FileCtrl, Registry, SynEditMiscProcs,
     NC_Main, NC_SimpleRegAccess, NC_AllVars, Highlighting;

const NeoS0MFiles:TNeoS0MFiles=
  ('NEOS0M_COMPILER.DLL','NEOS0M.DLL','NEOS0MEX.EXE','NEOS0M.CMD','NEOS0M.CHM','NEOS0M_DESCR.TXT','GROUPS.TXT');

{$R *.dfm}

procedure PaintColor(pb:TPaintBox; cl:TColor);
begin
  with pb.Canvas do
    begin
      Brush.Style:=bsSolid;
      Brush.Color:=cl;
      Pen.Style:=psSolid;
      Pen.Color:=clBlack;
      Rectangle(Rect(0,0,pb.Width,pb.Height));
    end;
end;

procedure TForm_Options.FormShow(Sender: TObject);
begin
with Form_Main do
begin
 edtCompilerDLL.Text:=CompilerDLL;
 edtKernelDLL.Text:=KernelDLL;
 edtKernelEXE.Text:=KernelEXE;
 edtCmdDescr.Text:=CmdDescriptions;
 edtCmdHelp.Text:=CmdHelp;
 edtTextDescr.Text:=CmdTextDescr;
 edtGroups.Text:=GroupFile;
 edtMRUCount.Text:=IntToStr(MaxMRU);
 edtPathToIWS.Text:=PathToIWS;

 chkUseAutoParameters.Checked:=UseAutoParameters;
 chkVisualEditor.Checked:=VisualCommandEditor;
 CmdPosRadioGroup.ItemIndex:=CmdPanelPosition;
 cbBigIcons.Checked:=BigIcons;
 cbBigButtons.Checked:=BigButtons;

 //   
 EditorFontName:=sedMain.Font.Name;
 EditorFontSize:=sedMain.Font.Size;
 //   
 if ColorSchema=-1 then
    begin
      UserColorSchemaInCombo;
      cbColorSchema.ItemIndex:=cbColorSchema.Items.Count-1;
      ColorSettings.ReadFromRegistry;
    end
  else
    begin
      cbColorSchema.ItemIndex:=ColorSchema;
      ColorSettings.ColorSchemaByIndex(ColorSchema);
    end;
  InUpdateFlag:=True;
  ShowColorSchema;

  if CmdSortOrderByName
    then rgCmdSortOrderByName.ItemIndex:=1
    else rgCmdSortOrderByName.ItemIndex:=0;
  cbIgnoreGroupsInFilteredList.Checked:=IgnoreGroupsInFilteredList;
  cbImmediateFilter.Checked:=ImmediateFilter;
  cbFilterByCmdName.Checked:=FilterByCmdName;
  cbCmdFilterByFirstChars.Checked:=CmdFilterByFirstChars;
  cbFilterByDescription.Checked:=FilterByDescription;
  cbDescrFilterByFirstChars.Checked:=DescrFilterByFirstChars;
  cbGroupIconsInFilteredList.Checked:=GroupIconsInFilteredList;
  cbQuickInfoOnAutoComplete.Checked:=QuickInfoOnAutoComplete;
  //    ,    ,    
  cbFilterByCmdNameClick(cbFilterByCmdName);
  cbFilterByDescriptionClick(cbFilterByDescription);

  InUpdateFlag:=False;
end;//with
end;

procedure TForm_Options.JoinColorBox(b: TButton; pb: TPaintBox);
begin
  with b do
    begin
      pb.Top:=Top+2;
      pb.Left:=Left+Width+4;
      pb.Height:=Height-4;
      pb.Width:=Height-4;
    end;
end;

procedure TForm_Options.lbSyntaxElementsClick(Sender: TObject);
begin
  ShowSelectedColorFromSchema;
end;

procedure _NeoS0MFilesSearch(const Folder:String; var FoundFiles:TNeoS0MFiles);
var i:Integer;
    FName:String;
    SearchRec : TSearchRec;
begin
  if FindFirst(Folder+ '*.*',faAnyFile,SearchRec)=0 then
    repeat
      if (SearchRec.Name = '.') or (SearchRec.Name = '..')
        then Continue;
      FName:=UpperCase(ExtractFileName(SearchRec.Name));
      //    -    
      if (SearchRec.Attr and faDirectory) <> 0
        then _NeoS0MFilesSearch(Folder+SearchRec.Name+'\',FoundFiles)
        else
          begin
            //  ,     -   NeoS0M
            for i:=Low(NeoS0MFiles) to High(NeoS0MFiles)
              do if FName=NeoS0MFiles[i]
                  //       NeoS0M -  
                  then FoundFiles[i]:=Folder+SearchRec.Name;
          end;
    until FindNext(SearchRec)<>0;
  FindClose(SearchRec);
end;

procedure TForm_Options.NeoS0MFilesSearch(const Folder: String);
var i:Integer;
    FoundFiles: TNeoS0MFiles;
begin
  //    
  for i:=Low(NeoS0MFiles) to High(NeoS0MFiles)
    do FoundFiles[i]:='';
  //  
  _NeoS0MFilesSearch(IncludeTrailingBackslash(Folder),FoundFiles);
  //     
  edtCompilerDLL.Text:=FoundFiles[1];
  edtKernelDLL.Text:=FoundFiles[2];
  edtKernelEXE.Text:=FoundFiles[3];
  edtCmdDescr.Text:=FoundFiles[4];
  edtCmdHelp.Text:=FoundFiles[5];
  edtTextDescr.Text:=FoundFiles[6];
  edtGroups.Text:=FoundFiles[7];
end;

procedure TForm_Options.pbBgColorPaint(Sender: TObject);
var i:Integer;
begin
  i:=lbSyntaxElements.ItemIndex;
  if i=-1
    then PaintColor(Sender as TPaintBox,clBtnFace)
    else PaintColor(Sender as TPaintBox,ColorSettings.Schema.Colors[TokenDescriptions[i].Token].Background);
end;

procedure TForm_Options.pbDelimiterPaint(Sender: TObject);
begin
  PaintColor(Sender as TPaintBox,ColorSettings.Schema.LNBorderColor);
end;

procedure TForm_Options.pbGradientEndPaint(Sender: TObject);
begin
  PaintColor(Sender as TPaintBox,ColorSettings.Schema.LNGradientEnd);
end;

procedure TForm_Options.pbGradientStartPaint(Sender: TObject);
begin
  PaintColor(Sender as TPaintBox,ColorSettings.Schema.LNGradientStart);
end;

procedure TForm_Options.pbLineNumbersBgPaint(Sender: TObject);
begin
  PaintColor(Sender as TPaintBox,ColorSettings.Schema.LNBackground);
end;

procedure TForm_Options.pbLineNumbersPaint(Sender: TObject);
begin
  PaintColor(Sender as TPaintBox,ColorSettings.Schema.LNColor);
end;

procedure TForm_Options.pbTextColorPaint(Sender: TObject);
var i:Integer;
begin
  i:=lbSyntaxElements.ItemIndex;
  if i=-1
    then PaintColor(Sender as TPaintBox,clBtnFace)
    else PaintColor(Sender as TPaintBox,ColorSettings.Schema.Colors[TokenDescriptions[i].Token].Foreground);
end;

procedure TForm_Options.btnChooseFontClick(Sender: TObject);
begin
  fdFontSelector.Font.Name:=EditorFontName;
  fdFontSelector.Font.Size:=EditorFontSize;
  if fdFontSelector.Execute then
    begin
      EditorFontName:=fdFontSelector.Font.Name;
      EditorFontSize:=fdFontSelector.Font.Size;
      //     
      if (EditorFontSize<5) or (EditorFontSize>96)
        then EditorFontSize:=8;
      ShowSelectedColorFromSchema;
    end;
end;

procedure TForm_Options.btnDelimiterClick(Sender: TObject);
begin
  ChooseColor(ColorSettings.Schema.LNBorderColor);
  pbDelimiter.Repaint;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.btnFindInAnyFolderClick(Sender: TObject);
var f:String;
begin
  f:=ExtractFilePath(GetModuleName(0));
  if FileCtrl.SelectDirectory('  ','',f)
    then NeoS0MFilesSearch(f);
end;

procedure TForm_Options.btnFindInAppFolderClick(Sender: TObject);
begin
  NeoS0MFilesSearch(ExtractFilePath(GetModuleName(0)));
end;

procedure TForm_Options.btnGradientEndClick(Sender: TObject);
begin
  ChooseColor(ColorSettings.Schema.LNGradientEnd);
  pbGradientEnd.Repaint;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.btnGradientStartClick(Sender: TObject);
begin
  ChooseColor(ColorSettings.Schema.LNGradientStart);
  pbGradientStart.Repaint;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.btnLineNumbersBgClick(Sender: TObject);
begin
  ChooseColor(ColorSettings.Schema.LNBackground);
  pbLineNumbersBg.Repaint;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.btnLineNumbersClick(Sender: TObject);
begin
  ChooseColor(ColorSettings.Schema.LNColor);
  pbLineNumbers.Repaint;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.btnLoadSchemaClick(Sender: TObject);
begin
  OpenDialog1.Filter:=_NSHFilter;
  if OpenDialog1.Execute then
    begin
      ColorSettings.ReadFromFile(OpenDialog1.FileName);
      UserColorSchemaInCombo;
      ShowColorSchema;
    end;
end;

procedure TForm_Options.btnOkClick(Sender: TObject);
var m_res:TModalResult;
    CmdPanelSettings:Cardinal;
begin
m_res:=mrOk;
with Form_Main do
begin
 CompilerDLL:=edtCompilerDLL.Text;
 KernelDLL:=edtKernelDLL.Text;
 KernelEXE:=edtKernelEXE.Text;
 CmdDescriptions:=edtCmdDescr.Text;
 CmdTextDescr:=edtTextDescr.Text;
 CmdHelp:=edtCmdHelp.Text;
 GroupFile:=edtGroups.Text;
 PathToIWS:=edtPathToIWS.Text;
 UseAutoParameters:=chkUseAutoParameters.Checked;
 VisualCommandEditor:=chkVisualEditor.Checked;

 if SetMaxMRU(edtMRUCount.Text)
  then
    begin
      Application.MessageBox(PWideChar(_InvalidMaxMRU),PWideChar(_Error),MB_ICONERROR+MB_APPLMODAL);
      m_res:=mrNone;
    end
  else NC_StoreInteger(NeoS0M_RegKey,'MaxMRU',MaxMRU);

 NC_StoreString(NeoS0M_RegKey,'CompilerDLL',CompilerDLL);
 NC_StoreString(NeoS0M_RegKey,'KernelDLL',KernelDLL);
 NC_StoreString(NeoS0M_RegKey,'KernelEXE',KernelEXE);
 NC_StoreString(NeoS0M_RegKey,'CmdDescriptions',CmdDescriptions);
 NC_StoreString(NeoS0M_RegKey,'CmdHelp',CmdHelp);
 NC_StoreString(NeoS0M_RegKey,'CmdTextDescr',CmdTextDescr);
 NC_StoreString(NeoS0M_RegKey,'GroupFile',GroupFile);
 NC_StoreString(NeoS0M_RegKey,'PathToIWS',PathToIWS);
 actWinScanner.Enabled:=FileExists(PathToIWS);
 btnWinScanner.Enabled:=actWinScanner.Enabled;

 NC_StoreBoolean(NeoS0M_RegKey,'UseAutoParameters',UseAutoParameters);
 NC_StoreBoolean(NeoS0M_RegKey,'VisualCommandEditor',VisualCommandEditor);
 CmdPanelPosition:=CmdPosRadioGroup.ItemIndex;
 NC_StoreInteger(NeoS0M_RegKey,'CmdPanelPosition',CmdPanelPosition);
 Form_Main.ChangeCmdPanelPosition;
 if BigIcons<>cbBigIcons.Checked then
  begin
    BigIcons:=cbBigIcons.Checked;
    NC_StoreBoolean(NeoS0M_RegKey,'BigIcons',BigIcons);
    ReMeasureItems;
    lstCommandList.Invalidate;
  end;
  BigButtons:=cbBigButtons.Checked;
  SetButtonSize;
  NC_StoreBoolean(NeoS0M_RegKey,'BigButtons',BigButtons);
  //   
  //      
  //    
  ColorSchema:=cbColorSchema.ItemIndex;
  if ColorSchema<>cbColorSchema.Items.Count-1
    //    
    then ColorSettings.ColorSchemaByIndex(ColorSchema)
    else //     ("") -     
      begin
        //     -1
        ColorSchema:=-1;
        //    
        ColorSettings.WriteToRegistry;
      end;
  //    
  ColorSettings.ApplyToEditor(sedMain);
  //       
  NC_StoreInteger(NeoS0M_RegKey,'ColorSchema',ColorSchema);
  //    ,       
  if EditorFontName<>sedMain.Font.Name then
    begin
      sedMain.Font.Name:=EditorFontName;
      sedMain.Gutter.Font.Name:=EditorFontName;
      NC_StoreString(NeoS0M_RegKey,'Font',EditorFontName)
    end;
  if EditorFontSize<>sedMain.Font.Size then
    begin
      sedMain.Font.Size:=EditorFontSize;
      sedMain.Gutter.Font.Size:=EditorFontSize;
      NC_StoreInteger(NeoS0M_RegKey,'FontSize',EditorFontSize)
    end;

  //     
  //  
  CmdPanelSettings:=rgCmdSortOrderByName.ItemIndex;
  CmdSortOrderByName:=rgCmdSortOrderByName.ItemIndex=0;
  //      ?
  IgnoreGroupsInFilteredList:=cbIgnoreGroupsInFilteredList.Checked;
  if cbIgnoreGroupsInFilteredList.Checked
    then Inc(CmdPanelSettings,IGNORE_GROUPS_IN_FILTERED_LIST);
  //  ?
  ImmediateFilter:=cbImmediateFilter.Checked;
  if ImmediateFilter
    then Inc(CmdPanelSettings,IMMEDIATE_FILTER);
  //    ?
  FilterByCmdName:=cbFilterByCmdName.Checked;
  if FilterByCmdName
    then Inc(CmdPanelSettings,FILTER_BY_NAME);
  //       ?
  CmdFilterByFirstChars:=cbCmdFilterByFirstChars.Checked;
  if CmdFilterByFirstChars
    then Inc(CmdPanelSettings,FILTER_BY_NAME_FIRST_CHARS);
  //    ?
  FilterByDescription:=cbFilterByDescription.Checked;
  if FilterByDescription
    then Inc(CmdPanelSettings,FILTER_BY_DESCRIPTION);
  //       ?
  DescrFilterByFirstChars:=cbDescrFilterByFirstChars.Checked;
  if DescrFilterByFirstChars
    then Inc(CmdPanelSettings,FILTER_BY_DESCRIPTION_FIRST_CHARS);
  //      ?
  GroupIconsInFilteredList:=cbGroupIconsInFilteredList.Checked;
  if GroupIconsInFilteredList
    then Inc(CmdPanelSettings,GROUP_ICONS_IN_FILTERED_LIST);

  //     
  NC_StoreInteger(NeoS0M_RegKey,'CmdPanelSettings',CmdPanelSettings);
  //     
  QuickInfoOnAutoComplete:=cbQuickInfoOnAutoComplete.Checked;
  NC_StoreBoolean(NeoS0M_RegKey,'QuickInfoOnAutoComplete',QuickInfoOnAutoComplete);
end;//with
 ModalResult:=m_res;
end;

procedure TForm_Options.btnSaveSchemaClick(Sender: TObject);
begin
  SaveDialog1.FileName:='';
  SaveDialog1.DefaultExt:='nsh';
  SaveDialog1.Filter:=_NSHFilter;
  if SaveDialog1.Execute then
    begin
      ColorSettings.WriteToFile(SaveDialog1.FileName);
    end;
end;

procedure TForm_Options.ChooseSyntaxItemColor(Sender: TObject);
begin
  with ColorSettings.Schema.Colors[TokenDescriptions[lbSyntaxElements.ItemIndex].Token] do
    if (Sender as TButton).Tag=0 then
      begin
        ChooseColor(Foreground);
        pbBgColor.Repaint;
      end
    else
      begin
        ChooseColor(Background);
        pbTextColor.Repaint;
      end;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.ChangeFontStyle(Sender: TObject);
const _sts:array[0..3] of TFontStyle=(fsBold,fsItalic,fsUnderline,fsStrikeout);
begin
  with ColorSettings.Schema.Colors[TokenDescriptions[lbSyntaxElements.ItemIndex].Token], Sender as TCheckBox
    do if Checked
        then Include(Style,_sts[Tag])
        else Exclude(Style,_sts[Tag]);
  if not InUpdateFlag
    then UserColorSchemaInCombo;
  ShowSelectedColorFromSchema;
end;

procedure TForm_Options.cbColorSchemaChange(Sender: TObject);
var cs:Integer;
begin
  cs:=cbColorSchema.ItemIndex;
  if cs<>cbColorSchema.Items.Count-1
    //    
    then ColorSettings.ColorSchemaByIndex(cs);
//  UserColorSchemaInCombo;
  ShowColorSchema;
end;

procedure TForm_Options.cbFilterByCmdNameClick(Sender: TObject);
begin
  cbCmdFilterByFirstChars.Visible:=cbFilterByCmdName.Checked;
end;

procedure TForm_Options.cbFilterByDescriptionClick(Sender: TObject);
begin
  cbDescrFilterByFirstChars.Visible:=cbFilterByDescription.Checked;
end;

procedure TForm_Options.cbGradientClick(Sender: TObject);
begin
  btnGradientStart.Enabled:=cbGradient.Checked;
  btnGradientEnd.Enabled:=cbGradient.Checked;
  ColorSettings.Schema.LNGradientPresent:=cbGradient.Checked;
  if not InUpdateFlag
    then UserColorSchemaInCombo;
end;

procedure TForm_Options.chkVisualEditorClick(Sender: TObject);
begin
  chkUseAutoParameters.Visible:=not chkVisualEditor.Checked;
end;

procedure TForm_Options.ChooseColor(var cl: TColor);
begin
  cdColorSelector.Color:=cl;
  if cdColorSelector.Execute then
    begin
      cl:=cdColorSelector.Color;
      cbColorSchema.ItemIndex:=cbColorSchema.Items.Count-1;
    end;
end;

procedure TForm_Options.FormCreate(Sender: TObject);
var i:Integer;
begin
 with Form_Main.pilBigButtons do
  begin
    btnLoadSchema.PngImage:=PngImages[1].PngImage;
    btnSaveSchema.PngImage:=PngImages[2].PngImage;
    btnChooseFont.PngImage:=PngImages[6].PngImage;
    btnFindInAppFolder.PngImage:=PngImages[7].PngImage;
    btnFindInAnyFolder.PngImage:=PngImages[8].PngImage;
  end;

 JoinColorBox(btnDelimiter,pbDelimiter);
 JoinColorBox(btnLineNumbers,pbLineNumbers);
 JoinColorBox(btnLineNumbersBg,pbLineNumbersBg);
 JoinColorBox(btnGradientStart,pbGradientStart);
 JoinColorBox(btnGradientEnd,pbGradientEnd);

 JoinColorBox(btnTextColor,pbTextColor);
 JoinColorBox(btnBgColor,pbBgColor);

 pgcOptions.ActivePageIndex:=0;
 //    
 lbSyntaxElements.Items.BeginUpdate;
 for i:=Low(TokenDescriptions) to High(TokenDescriptions)
  do lbSyntaxElements.Items.Add(TokenDescriptions[i].Description);
 lbSyntaxElements.Items.EndUpdate;

 //  " " (  )
 ColorSettings:=TColorSettings.Create;
 //     
 for i:=Low(ColorSchemas) to High(ColorSchemas)
  do cbColorSchema.Items.Add(ColorSchemas[i].Name);
 cbColorSchema.Items.Add(_UserColorSchema);
end;

procedure TForm_Options.FormDestroy(Sender: TObject);
begin
   FreeAndNil(ColorSettings);
end;

procedure TForm_Options.SelectFile(Sender: TObject);
begin
with OpenDialog1 do
 begin
  case (Sender as TControl).Tag of
   0: begin
        Filter:=_DLLFilter;
        FileName:='NeoS0M_Compiler.dll';
      end;
   1: begin
        Filter:=_DLLFilter;
        FileName:='NeoS0M.dll';
      end;
   2: begin
        Filter:=_EXEFilter;
        FileName:='NeoS0MEx.exe';
      end;
   3: begin
        Filter:=_CMDFilter;
        FileName:='NeoS0M.cmd';
      end;
   4: begin
        Filter:=_CHMFilter;
        FileName:='NeoS0M.chm';
      end;
   5: begin
        Filter:=_TXTFilter;
        FileName:='NeoS0M_descr.txt';
      end;
   6: begin
        Filter:=_TXTFilter;
        FileName:='Groups.txt';
      end;
   7: begin
        Filter:=_EXEFilter;
        FileName:='WinScanner.exe';
      end;
  end;//case
  if OpenDialog1.Execute then
   begin
    case (Sender as TControl).Tag of
     0: edtCompilerDLL.Text:=FileName;
     1: edtKernelDLL.Text:=FileName;
     2: edtKernelEXE.Text:=FileName;
     3: edtCmdDescr.Text:=FileName;
     4: edtCmdHelp.Text:=FileName;
     5: edtTextDescr.Text:=FileName;
     6: edtGroups.Text:=FileName;
     7: edtPathToIWS.Text:=FileName;
    end;//case
   end;//OpenDialog1.Execute
 end;//with
end;

procedure TForm_Options.ShowColorSchema;
begin
  InUpdateFlag:=True;
  cbGradient.Checked:=ColorSettings.Schema.LNGradientPresent;
  cbGradientClick(cbGradient);
  pbDelimiter.Repaint;
  pbLineNumbers.Repaint;
  pbLineNumbersBg.Repaint;
  pbGradientStart.Repaint;
  pbGradientEnd.Repaint;
  ShowSelectedColorFromSchema;
  InUpdateFlag:=False;
end;

procedure TForm_Options.ShowSelectedColorFromSchema;
const PSEUDO_LINE_NUMBER='1234';
var b,OldInUpdate:Boolean;
    i:Integer;
    LNWidth,isH:Integer;
    OldTranspMode:Integer;
    token:TtkTokenKind;
begin
  i:=lbSyntaxElements.ItemIndex;
  b:=i>=0;
  btnTextColor.Visible:=b;
  btnBgColor.Visible:=b;
  pbTextColor.Visible:=b;
  pbBgColor.Visible:=b;
  cbBold.Visible:=b;
  cbItalic.Visible:=b;
  cbUnderline.Visible:=b;
  cbStrikeout.Visible:=b;
  if b then
    begin
      token:=TokenDescriptions[i].Token;
      pbTextColor.Repaint;
      pbBgColor.Repaint;
      with ColorSettings.Schema.Colors[token] do
        begin
          OldInUpdate:=InUpdateFlag;
          InUpdateFlag:=True;
          cbBold.Checked:=fsBold in Style;
          cbItalic.Checked:=fsItalic in Style;
          cbUnderline.Checked:=fsUnderline in Style;
          cbStrikeout.Checked:=fsStrikeOut in Style;
          InUpdateFlag:=OldInUpdate;

          //   
          isH:=imgSample.Height;
          with imgSample.Canvas do
          begin
            Font.Name:=EditorFontName;
            Font.Size:=EditorFontSize;
            Font.Style:=[];
            //      
            LNWidth:=TextWidth(PSEUDO_LINE_NUMBER);
            //   -  
            if ColorSettings.Schema.LNGradientPresent then
              begin
                SynDrawGradient(imgSample.Canvas,
                                ColorSettings.Schema.LNGradientStart,ColorSettings.Schema.LNGradientEnd,
                                48,Rect(0,0,LNWidth+4,isH),True);
              end
            else
              begin
                Brush.Color:=ColorSettings.Schema.LNBackground;
                Brush.Style:=bsSolid;
                FillRect(Rect(0,0,LNWidth+4,isH));
              end;
            Pen.Style:=psSolid;
            //    
            Pen.Color:=ColorSettings.Schema.LNBackground;
            MoveTo(LNWidth+4,0);
            LineTo(LNWidth+4,isH);
            //  
            Pen.Color:=ColorSettings.Schema.LNBorderColor;
            MoveTo(LNWidth+5,0);
            LineTo(LNWidth+5,isH);
            Brush.Color:=ColorSettings.Schema.Colors[token].Background;
            //   
            OldTranspMode:=SetBkMode(Handle,TRANSPARENT);
            Font.Color:=ColorSettings.Schema.LNColor;
            TextOut(2,0,PSEUDO_LINE_NUMBER);
            //  " "
            Brush.Style:=bsSolid;
            Brush.Color:=ColorSettings.Schema.Colors[tkNone].Background;
            FillRect(Rect(LNWidth+6,0,Width,Height));
            //     
            SetBkMode(Handle,OldTranspMode);
            Font.Color:=Foreground;
            Brush.Color:=Background;
            Font.Style:=Style;
            TextOut(LNWidth+8,0,TokenDescriptions[i].Sample);
          end;//with
        end;//with
    end

end;

procedure TForm_Options.UserColorSchemaInCombo;
begin
  with cbColorSchema
    do ItemIndex:=Items.Count-1;
end;

end.
