unit NC_ParamConversion;

interface
uses SysUtils,StrUtils;

//      
procedure Trim(var s:UnicodeString);
function IsHex(const s:UnicodeString):Boolean;
// ,    ( !    )  16- 
function IsHexDump(const s:UnicodeString):Boolean;

//       
function AsAnyIntNumber(const s:UnicodeString; var err:Boolean):Int64;

function FromBin(const s:UnicodeString; var err:Boolean):Int64;
function FromOct(const s:UnicodeString; var err:Boolean):Int64;
function FromDec(const s:UnicodeString; var err:Boolean):Int64;
function FromHex(const s:UnicodeString; var err:Boolean):Int64;
function FromDate(const s:UnicodeString; var err:Boolean):Int64;

function FromANSIBinBlock(const s:UnicodeString;var err:Boolean):UnicodeString;
function FromUNICODEBinBlock(const s:UnicodeString;var err:Boolean):UnicodeString;

function ToBin(n:Int64):UnicodeString;
function ToOct(n:Int64):UnicodeString;
function ToDec(n:Int64):UnicodeString;
function ToHex(n:Int64):UnicodeString;
function ToDate(n:Int64):UnicodeString;

function ToANSIBinBlock(const s:UnicodeString):UnicodeString;
function ToUNICODEBinBlock(const s:UnicodeString):UnicodeString;

implementation

uses Windows, NC_Chars;

procedure Trim(var s:UnicodeString);
var i,t:Integer;
begin
  t:=-1;
  for I := 1 to Length(s)
    do if s[i]<=' '
        then t:=i
        else Break;
  if t>0 then
    begin
      Delete(s,1,t);
      t:=-1;
    end;//if t>0
  for I:=Length(s) downto 1
    do if s[i]<=' '
      then t:=i
      else Break;
  if t>0
    then Delete(s,t,Length(s)-t+1);
end;

function IsHex(const s:UnicodeString):Boolean;
var i:Integer;
begin
  if s<>'' then
    begin
      if s[1]<>'$'
        then Result:=False
        else
          begin
            if s='$'
              then Result:=False
              else
                begin
                  Result:=True;
                  for i:=2 to Length(s) do
                    //     -   ,    
                    if CharInfoTable[s[i]] and CHAR_IS_HEX_DIGIT=0 then
                      begin
                        Result:=False;
                        Break;
                      end;
                end;//else
          end;//else
    end//if s<>''
  else Result:=False;
end;

function FromBin(const s:UnicodeString; var err:Boolean):Int64;
var i:Integer;
    HasNumber,HasSign:Boolean;
begin
  Result:=0;
  err:=False;
  HasSign:=False;
  HasNumber:=False;
  for i:=1 to Length(s) do
    begin
      case s[i] of
        '0':  begin
                HasNumber:=True;
                Result:=Result shl 1;
              end;
        '1':  begin
                HasNumber:=True;
                Result:=(Result shl 1) or 1;
              end;
        '-':  begin
                if HasSign or HasNumber
                  then begin err:=True; Break end
                  else HasSign:=True;
              end;
        else begin err:=True; Break; end;
      end;//case
    end;//for
    if HasNumber then
      begin
        if (not err) and (HasSign) then Result:=-Result;
      end
    else err:=True; //      
end;

function FromOct(const s:UnicodeString; var err:Boolean):Int64;
var i:Integer;
    HasNumber,HasSign:Boolean;
begin
  Result:=0;err:=False;HasSign:=False;HasNumber:=False;
  for i:=1 to Length(s) do
    begin
      case s[i] of
        '0'..'7':
              begin
                Result:=(Result shl 3) or (Ord(s[i])-Ord('0'));
                HasNumber:=True;
              end;
        '-':  begin
                if HasSign or HasNumber
                  then begin err:=True; Break end
                  else HasSign:=True;
              end;
        else begin err:=True; Break; end;
      end;//case
    end;//for
    if HasNumber then
      begin
        if (not err) and (HasSign) then Result:=-Result;
      end
    else err:=True; //      
end;

function FromDec(const s:UnicodeString; var err:Boolean):Int64;
var i:Integer;
    HasNumber,HasSign:Boolean;
begin
  Result:=0;
  err:=False;
  HasSign:=False;
  HasNumber:=False;
  for i:=1 to Length(s) do
    begin
      case s[i] of
        '0'..'9':
              begin
                Result:=Result*10+(Ord(s[i])-Ord('0'));
                HasNumber:=True;
              end;
        '-':  begin
                if HasSign or HasNumber
                  then begin err:=True; Break end
                  else HasSign:=True;
              end;
        else begin err:=True; Break; end;
      end;//case
    end;//for
    if HasNumber then
      begin
        if (not err) and (HasSign) then Result:=-Result;
      end
    else err:=True; //      
end;

function FromHex(const s:UnicodeString; var err:Boolean):Int64;
var i:Integer;
    HasNumber,HasSign:Boolean;
begin
  Result:=0;err:=False;HasSign:=False;HasNumber:=False;
  for i:=1 to Length(s) do
    begin
      case s[i] of
        '0'..'9':
              begin
                Result:=(Result shl 4) or (Ord(s[i])-Ord('0'));
                HasNumber:=True;
              end;
        'a'..'f':
              begin
                Result:=(Result shl 4) or (Ord(s[i])-Ord('a')+10);
                HasNumber:=True;
              end;
        'A'..'F':
              begin
                Result:=(Result shl 4) or (Ord(s[i])-Ord('A')+10);
                HasNumber:=True;
              end;
        '-':  begin
                if HasSign or HasNumber
                  then begin err:=True; Break end
                  else HasSign:=True;
              end;
        else begin err:=True; Break; end;
      end;//case
    end;//for
    if HasNumber then
      begin
        if (not err) and (HasSign) then Result:=-Result;
      end
    else err:=True; //      
end;

function ToBin(n:Int64):UnicodeString;
begin
  Result:='';
  repeat
    if (n and 1)=1
      then Result:='1'+Result
      else Result:='0'+Result;
    n:=n shr 1;
  until n=0;
end;

function ToOct(n:Int64):UnicodeString;
begin
  Result:='';
  repeat
    Result:=WideChar(Ord('0')+(n and 7))+Result;
    n:=n shr 3;
  until n=0;
end;

function ToDec(n:Int64):UnicodeString;
begin
  Result:=IntToStr(n);
end;

function ToHex(n:Int64):UnicodeString;
begin
  Result:=IntToHex(n,0);
end;

function FromDate(const s:UnicodeString; var err:Boolean):Int64;
  label DoIt;
  var pos_in_str:Integer;
      IsLast:Boolean;
      tmp_int:Integer;
      st:SYSTEMTIME;

  function Str2Int(const s:UnicodeString; var value:Integer):Boolean;
  //   True -    
  var err_code:Integer;
  begin
    Val(s,value,err_code);
    Result:=err_code=0;
  end;

  function ExtractSubstrFromStr(const s,delimiter:UnicodeString; var StartPos:Integer; var IsLast:Boolean):UnicodeString;
  var delim_found:Integer;
  begin
   delim_found:=PosEx(delimiter,s,StartPos);
   if delim_found=0 then //   
     begin
      IsLast:=True;
      Result:=Copy(s,StartPos,Length(s)-StartPos+1);
     end
    else
     begin //  
      Result:=Copy(s,StartPos,delim_found-StartPos);
      IsLast:=delim_found=Length(s);
      StartPos:=delim_found+1;
     end;
  end;

begin
  err:=True;
  with st do //    - 0:00:00.000
  begin
    wHour:=0;
    wMinute:=0;
    wSecond:=0;
    wMilliseconds:=0;
    //    -  
    pos_in_str:=1;

    //   
    if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
      then wDay:=tmp_int
      else Exit;
    //     
    if IsLast
      then Exit;

    //  
    if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
      then wMonth:=tmp_int
      else Exit;
    //     
    if IsLast
      then Exit;

    //  
    if Str2Int(ExtractSubstrFromStr(s,',',pos_in_str,IsLast),tmp_int)
      then wYear:=tmp_int
      else Exit;
    //      -    
    if IsLast
      then goto DoIt;

    //  
    if Str2Int(ExtractSubstrFromStr(s,':',pos_in_str,IsLast),tmp_int)
      then wHour:=tmp_int
      else Exit;
    if IsLast
      then goto DoIt;

    //  
    if Str2Int(ExtractSubstrFromStr(s,':',pos_in_str,IsLast),tmp_int)
      then wMinute:=tmp_int
      else Exit;
    if IsLast
      then goto DoIt;

    //  
    if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
      then wSecond:=tmp_int
      else Exit;
    if IsLast
      then goto DoIt;

    //  
    if Str2Int(ExtractSubstrFromStr(s,'.',pos_in_str,IsLast),tmp_int)
      then wMilliseconds:=tmp_int
      else Exit;
    //      ,     -  (..   )
    if not IsLast
      then Exit;
  end;//with

DoIt:
  err:=not SystemTimeToFileTime(st,FILETIME(Result));
end;

function NumToStr(n:Integer;digits:Integer):UnicodeString;
begin
  Str(n,Result);
  if Length(Result)<digits
    then Result:=StringOfChar('0',digits-Length(Result))+Result;
end;

function ToDate(n:Int64):UnicodeString;
label _exit;
var st:SYSTEMTIME;
begin
  FileTimeToSystemTime(FILETIME(n),st);
  with st do
    begin
      Result:=NumToStr(wDay,2)+'.'+NumToStr(wMonth,2)+'.'+NumToStr(wYear,0);
      if (wHour=0) and (wMinute=0) and (wSecond=0) and (wMilliseconds=0) then
        goto _exit
        else Result:=Result+','+NumToStr(wHour,2);
      if (wMinute=0) and (wSecond=0) and (wMilliseconds=0) then
        goto _exit
        else Result:=Result+':'+NumToStr(wMinute,2);
      if (wSecond=0) and (wMilliseconds=0) then
        goto _exit
        else Result:=Result+':'+NumToStr(wSecond,2);
      if wMilliseconds<>0
        then Result:=Result+'.'+NumToStr(wMilliseconds,0);
    end;//with
  _exit:
end;

function IsHexDump(const s:UnicodeString):Boolean;
var i,c:Integer;
begin
  c:=0;
  for i:=1 to Length(s)
    do if CharInfoTable[s[i]] and CHAR_IS_HEX_DIGIT<>0
        then Inc(c);
  Result:=not Odd(c);
end;


function FromXBinBlock(const s:UnicodeString;quote:WideChar;var err:Boolean):UnicodeString;
var i:Integer;
    OddQuote:Boolean;
begin
  OddQuote:=False;
  err:=False;
  Result:='';
  for i:=1 to Length(s) do
    begin
      if s[i]=quote then //   
        begin
          if OddQuote
            //      -   
            then OddQuote:=False
            //          -    
            else
              begin
                OddQuote:=True;
                Result:=Result+s[i];
              end
        end
      else //   
        begin
          if OddQuote then //        -  
            begin
              err:=True;
              Break
            end
            else Result:=Result+s[i];
        end;
    end;//for
  //     -  
  if OddQuote then err:=True;
end;

function FromANSIBinBlock(const s:UnicodeString;var err:Boolean):UnicodeString;
begin
  Result:=FromXBinBlock(s,'''',err);
end;

function ToANSIBinBlock(const s:UnicodeString):UnicodeString;
var i:Integer;
begin
  Result:='';
  for i:=1 to Length(s)
    do if s[i]=''''
        then Result:=Result+''''''
        else Result:=Result+s[i];
end;

function FromUNICODEBinBlock(const s:UnicodeString;var err:Boolean):UnicodeString;
begin
  Result:=FromXBinBlock(s,'"',err);
end;

function ToUNICODEBinBlock(const s:UnicodeString):UnicodeString;
var i:Integer;
begin
  for i:=1 to Length(s)
    do if s[i]='"'
        then Result:=Result+'""'
        else Result:=Result+s[i];
end;

function AsAnyIntNumber(const s:UnicodeString; var err:Boolean):Int64;
var v:UnicodeString;
    e:Extended;
    _err:Integer;
begin
  v:=s;
  if v<>'' then
  begin
    if (v[1]='*') or (v[1]='%') then Delete(v,1,1);
  end;
  Val(v,Result,_err);
  if _err=0
    then err:=False
  else //      -     
    begin
      Val(v,e,_err);
      if _err=0 then
        begin
          err:=False;
          Result:=Trunc(e);
        end
        else err:=True;
    end;
  
end;

end.
