unit NC_ParseDescription;

interface

uses Classes, SysUtils, StrUtils;

type TParamDialogTypes=(
        pdtAnyTextParam=0,  //   
        pdtAnyDWORD=1,      //    DWORD
        pdtAnyFLOAT=2,      //    FLOAT
        pdtAnyINTEGER=3,    //    INTEGER
        pdtAnyBinBlock=4   //    BIN BLOCK
//        pdtDWORDCheckBox=10 // DWORD-
    );

type TOnePTypeDescription=record
        Name:UnicodeString;
        DialogType:TParamDialogTypes;
        Delim:UnicodeString;
        DefaultValue:UnicodeString;
        Values:UnicodeString;
     end;

     TPTypeDescriptions=class(TObject)
       private
        Descriptions:array of TOnePTypeDescription;
        function AddLine(const s:UnicodeString):Boolean;
        function GetDescription(const name:UnicodeString):Integer;
        function GetDefaultValue(const name:UnicodeString):UnicodeString;
        function GetParamType(idx:Integer):TParamDialogTypes;
       public
        constructor Create(const fname:UnicodeString);
        destructor Done;

        property DescriptionByName[const name:UnicodeString]:Integer read GetDescription;
        property DefaultValueByName[const name:UnicodeString]:UnicodeString read GetDefaultValue;
        property ParamType[idx:Integer]:TParamDialogTypes read GetParamType;
     end;

implementation

{ TCmdDescriptions }

function TPTypeDescriptions.AddLine(const s: UnicodeString): Boolean;
var _Name,_Delim,_DefaultValue,_Value,tmp:UnicodeString;
    _DialogType:TParamDialogTypes;
    comma_pos,prev_comma_pos,counter:Integer;
begin
  inherited Create;
  _Name:='';
  _DialogType:=pdtAnyTextParam;
  _Delim:='|';
  _DefaultValue:='';
  _Value:='';

  prev_comma_pos:=0;
  counter:=0;
  Result:=False;
  repeat
    comma_pos:=PosEx(',',s,prev_comma_pos+1);
    if (comma_pos<>0) or (counter=4) then // - 
      begin
        if (comma_pos=0) and (counter=4)
          then  tmp:=Copy(s,prev_comma_pos+1,Length(s)-prev_comma_pos-1)
          else  tmp:=Copy(s,prev_comma_pos+1,comma_pos-prev_comma_pos-1);
        if tmp<>'' then
          begin
            case counter of
              0:  _Name:=tmp; //   command_descr
              1:  _DialogType:=TParamDialogTypes(StrToInt(tmp));
              2:  _Delim:=tmp;
              3:  _DefaultValue:=tmp;
              4:  begin _Value:=tmp; Result:=True; break; end;
            end;//case
          end;
        Inc(counter);
        prev_comma_pos:=comma_pos;
      end;
  until comma_pos=0;
  Result:=Result and (_Name<>'') and (_DialogType<>pdtAnyTextParam) and (_Value<>'');
  // !
  if Result then
    begin
      counter:=Length(Descriptions);
      SetLength(Descriptions,counter+1);
      with Descriptions[counter] do
        begin
          Name:=_Name;
          DialogType:=_DialogType;
          Delim:=_Delim;
          DefaultValue:=_DefaultValue;
          Values:=_Value;
        end;
    end;
end;

constructor TPTypeDescriptions.Create(const fname: UnicodeString);
var sl:TStringList;
    i:Integer;
begin
  sl:=TStringList.Create;
  try
    sl.LoadFromFile(fname);
    for i:= 0 to sl.Count-1 do
      begin
        if not AddLine(sl.Strings[i])
          then raise EStringListError.Create('      '+IntToStr(i)+'  '+fname);
      end;
  finally
    sl.Free;
  end;
end;

destructor TPTypeDescriptions.Done;
begin
  Descriptions:=nil;
  inherited;
end;

function TPTypeDescriptions.GetDefaultValue(const name: UnicodeString): UnicodeString;
var i:Integer;
begin
  i:=GetDescription(name);
  if (i<Low(Descriptions)) or (i>High(Descriptions))
    then Result:=''
    else Result:=Descriptions[i].DefaultValue;
end;

function TPTypeDescriptions.GetDescription(const name: UnicodeString): Integer;
var i:Integer;
begin
  Result:=-1;
  for i:=Low(Descriptions) to High(Descriptions) do
    begin
      if Descriptions[i].name=name then begin Result:=i; break; end;
    end;
end;

function TPTypeDescriptions.GetParamType(idx: Integer): TParamDialogTypes;
begin
  if (idx<Low(Descriptions)) or (idx>High(Descriptions))
    then Result:=pdtAnyTextParam
    else Result:=Descriptions[idx].DialogType;
end;

end.
