unit PNC_ParamINTEGER;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, PNC_Param, StdCtrls, ExtCtrls, Buttons, PngSpeedButton, ComCtrls,
  PngBitBtn;

type
  TfrmParamINTEGER = class(TfrmParam)
    gbAsNumber: TGroupBox;
    lblValue: TLabel;
    cbForcePrefix: TCheckBox;
    edtValue: TEdit;
    cbHexValue: TCheckBox;
    gbAsTimeInterval: TGroupBox;
    gbAsDateTime: TGroupBox;
    : TLabel;
    edtDY: TEdit;
    Label1: TLabel;
    edtHR: TEdit;
    Label2: TLabel;
    edtMN: TEdit;
    Label3: TLabel;
    edtMS: TEdit;
    Label4: TLabel;
    edtSC: TEdit;
    dtpDate: TDateTimePicker;
    cbNegativeValue: TCheckBox;
    btnAsNumber: TPngSpeedButton;
    btnAsDateTime: TPngSpeedButton;
    btnAsTimeInterval: TPngSpeedButton;
    Label5: TLabel;
    btnCurrent: TPngBitBtn;
    edtHours: TEdit;
    Label6: TLabel;
    edtMinutes: TEdit;
    Label7: TLabel;
    edtSeconds: TEdit;
    c: TLabel;
    edtMSecs: TEdit;
    Label8: TLabel;
    procedure btnCurrentClick(Sender: TObject);
    procedure SelectInputMethod(Sender: TObject);
    procedure AnyButtonClicked(Sender: TObject);
  private
    { Private declarations }
    CurrentPanel:Integer;

    function GetAsInt:Int64;
    function GetAsTimeInterval:Int64;
    function GetAsDateTime:Int64;

    function GetIntValue:Int64;
    procedure SetIntValue(i:Int64);
    procedure SetCurrentTime;

    procedure ShowTimeInterval(i:Int64);
    procedure ShowDateTime(i:Int64);
    procedure ShowInteger(i:Int64);
  protected
    //    ,     
    procedure SetData(const s:String);override;
    //   ,      ,  
    function IsDataValid(const s:String):String;override;
    //  ,   
    function GetValue:String;override;
  public
    { Public declarations }
    procedure Prepare(const Description,DataControls:String);override;
    procedure ShowPanel;
    //  :             
    procedure Go(const s, DefaultValue:String; const Abilities:Integer);override;
  end;

var
  frmParamINTEGER: TfrmParamINTEGER;

implementation

uses NC_Main, Math;

{$R *.dfm}

{ TfrmParamINTEGER }

procedure TfrmParamINTEGER.AnyButtonClicked(Sender: TObject);
begin
  inherited;
  ShowPanel;
end;

procedure TfrmParamINTEGER.btnCurrentClick(Sender: TObject);
begin
  SetCurrentTime;
end;

function TfrmParamINTEGER.GetAsDateTime: Int64;
var st:SYSTEMTIME;
begin
  DecodeDate(dtpDate.Date,st.wYear,st.wMonth,st.wDay);
  st.wHour:=StrToIntDef(edtHours.Text,0);
  st.wMinute:=StrToIntDef(edtMinutes.Text,0);
  st.wSecond:=StrToIntDef(edtSeconds.Text,0);
  st.wMilliseconds:=StrToIntDef(edtMSecs.Text,0);
  if not SystemTimeToFileTime(st,FILETIME(Result))
    then Result:=0;
end;

function TfrmParamINTEGER.GetAsInt: Int64;
var s:String;
    ec:Integer;
begin
  s:=edtValue.Text;
  //    -  $   
  if cbHexValue.Checked
    then InsertBucksSign(s);
  //     
  Val(s,Result,ec);
  //  ,  0
  if ec<>0
    then Result:=0;
end;

function TfrmParamINTEGER.GetAsTimeInterval: Int64;
var i64:Int64;
    ec:Integer;
begin
      Result:=0;
      if edtDY.Text<>'' then
        begin
          Val(edtDY.Text,i64,ec);
          Inc(Result,i64*864000000000);
        end;
      if edtHR.Text<>'' then
        begin
          Val(edtHR.Text,i64,ec);
          Inc(Result,i64*36000000000);
        end;
      if edtMN.Text<>'' then
        begin
          Val(edtMN.Text,i64,ec);
          Inc(Result,i64*600000000);
        end;
      if edtSC.Text<>'' then
        begin
          Val(edtSC.Text,i64,ec);
          Inc(Result,i64*10000000);
        end;
      if edtMS.Text<>'' then
        begin
          Val(edtMS.Text,i64,ec);
          Inc(Result,i64*10000);
        end;
      if cbNegativeValue.Checked
        then Result:=-Result;
end;

function TfrmParamINTEGER.GetIntValue: Int64;
begin
  if btnAsNumber.Down
    then Result:=GetAsInt
    else if btnAsDateTime.Down
          then Result:=GetAsDateTime
          else Result:=GetAsTimeInterval;
end;

function TfrmParamINTEGER.GetValue: String;
var y,m,d:Word;
begin
  if btnAsNumber.Down then
    begin
      Result:=edtValue.Text;
      if cbHexValue.Checked
        then InsertBucksSign(Result);
      if cbForcePrefix.Checked
        then Result:='%'+Result;
    end;
  if btnAsDateTime.Down then
    begin
      DecodeDate(dtpDate.Date,y,m,d);
      Result:=MakeDateTime(IntToStr(y),IntToStr(m),IntToStr(d),edtHours.Text,edtMinutes.Text,edtSeconds.Text,edtMsecs.Text);
    end;
  if btnAsTimeInterval.Down
    then Result:=IntToStr(GetIntValue);
end;

procedure TfrmParamINTEGER.Go(const s, DefaultValue: String;
  const Abilities: Integer);
begin
  inherited;
  ShowPanel;
end;

function TfrmParamINTEGER.IsDataValid(const s: String): String;
var i:Int64;
begin
  if Length(s)>0 then
    begin
      if s[1]='#'
        then Result:=IsValidDateTime(s,i)
        else Result:=IsValidINTEGER(s,i);
    end;
end;

procedure TfrmParamINTEGER.Prepare(const Description, DataControls: String);
var h:Integer;
begin
  inherited;
  CurrentPanel:=0;
  btnCurrent.PngImage:=Form_Main.pilBigButtons.PngImages[11].PngImage;
  with Form_Main.pilSmallButtons do
    begin
      btnAsNumber.PngImage:=PngImages[14].PngImage;
      btnAsDateTime.PngImage:=PngImages[15].PngImage;
      btnAsTimeInterval.PngImage:=PngImages[16].PngImage;
    end;
  gbAsNumber.Top:=btnAsNumber.Top+btnAsNumber.Height+2;
  gbAsDateTime.Top:=gbAsNumber.Top;
  gbAsTimeInterval.Top:=gbAsNumber.Top;
  h:=Max(gbAsNumber.Height,gbAsDateTime.Height);
  h:=Max(h,gbAsTimeInterval.Height);
  ClientHeight:=gbAsNumber.Top+h+bvlDelimiter.Height+2;
  btnAsNumber.Down:=True;
end;

procedure TfrmParamINTEGER.SelectInputMethod(Sender: TObject);
var i64:Int64;
begin
  case CurrentPanel of
    0: i64:=GetAsInt;
    1: i64:=GetAsDateTime;
    2: i64:=GetAsTimeInterval;
  end;//case
  SetIntValue(i64);
  ShowPanel;
  if btnAsNumber.Down
    then CurrentPanel:=0;
  if btnAsDateTime.Down
    then CurrentPanel:=1;
  if btnAsTimeInterval.Down
    then CurrentPanel:=2;
end;

procedure TfrmParamINTEGER.SetCurrentTime;
var st:SYSTEMTIME;
begin
  GetLocalTime(st);
  dtpDate.Date:=EncodeDate(st.wYear,st.wMonth,st.wDay);
  edtHours.Text:=Format('%2.2d',[st.wHour]);
  edtMinutes.Text:=Format('%2.2d',[st.wMinute]);
  edtSeconds.Text:=Format('%2.2d',[st.wSecond]);
  edtMsecs.Text:=Format('%3.3d',[st.wMilliseconds]);
end;

procedure TfrmParamINTEGER.SetData(const s: String);
var e:Integer;
    i64:Int64;
begin
  if Length(s)>=1 then
    case s[1] of
      '%':  begin
              cbForcePrefix.Checked:=True;
              Val(Copy(s,2,Length(s)-1),i64,e);
              //     
              cbHexValue.Checked:=Pos('$',s)<>0;
            end;
      '#':  begin
              cbForcePrefix.Checked:=False;
              if IsValidDateTime(s,i64)=''
                then e:=0
                else e:=1;
            end;
      else  begin
              cbForcePrefix.Checked:=False;
              Val(s,i64,e);
              //     
              cbHexValue.Checked:=Pos('$',s)<>0;
            end;
    end;
  //    
  if e=0
    then SetIntValue(i64)
    else SetIntValue(0);
end;

procedure TfrmParamINTEGER.SetIntValue(i: Int64);
begin
  //  /
  ShowDateTime(i);
  //   
  ShowTimeInterval(i);
  //  
  ShowInteger(i);
end;

procedure TfrmParamINTEGER.ShowDateTime(i: Int64);
var st:SYSTEMTIME;
begin
  //  /
  FileTimeToSystemTime(FILETIME(i),st);
  dtpDate.Date:=EncodeDate(st.wYear,st.wMonth,st.wDay);
  edtHours.Text:=Format('%2.2d',[st.wHour]);
  edtMinutes.Text:=Format('%2.2d',[st.wMinute]);
  edtSeconds.Text:=Format('%2.2d',[st.wSecond]);
  edtMsecs.Text:=Format('%3.3d',[st.wMilliseconds]);
end;

procedure TfrmParamINTEGER.ShowInteger(i: Int64);
begin
  if cbHexValue.Checked
    then edtValue.Text:=IntToHex(i,0)
    else edtValue.Text:=IntToStr(i);
end;

procedure TfrmParamINTEGER.ShowPanel;
var v:Boolean;
begin
  v:=btnValue.Down;
  gbAsNumber.Visible:=btnAsNumber.Down and v;
  gbAsTimeInterval.Visible:=btnAsTimeInterval.Down and v;
  gbAsDateTime.Visible:=btnAsDateTime.Down and v;
end;

procedure TfrmParamINTEGER.ShowTimeInterval(i: Int64);
begin
  cbNegativeValue.Checked:=i<0;
  if i<0
    then i:=-i;
  edtDY.Text:=IntToStr(i div 864000000000);
  i:=i mod 864000000000;
  edtHR.Text:=Format('%2.2d',[i div 36000000000]);
  i:=i mod 36000000000;
  edtMN.Text:=Format('%2.2d',[i div 600000000]);
  i:=i mod 600000000;
  edtSC.Text:=Format('%2.2d',[i div 10000000]);
  i:=i mod 10000000;
  edtMS.Text:=Format('%2.2d',[i div 10000]);
end;

end.
