unit NeoS0M_Executor;

interface

uses NeoS0M_Types, NeoS0M_Variables, NeoS0M_AutoRestoreTime, Windows;

var KernelName:UnicodeString;
    MyModuleHandle:Cardinal=0;

var BaseTimeToRestore:FILETIME;      //   (    
                                     //  /)
    BaseTickCount:Cardinal;          //  TickCount
                                     //    /
    DateTimeChanged:Boolean=False;   //    /

function LoadParameter_DWORD(var context:TScriptContext; var State:TParamState; def_value:Cardinal):Cardinal;
function LoadParameter_INTEGER(var context:TScriptContext; var State:TParamState; const def_value:Int64):Int64;
function LoadParameter_FLOAT(var context:TScriptContext; var State:TParamState; const def_value:Extended):Extended;
function LoadParameter_BinBlock(var context:TScriptContext; var State:TParamState):PUA_BinaryBlock;
function LoadParameter_UnicodeString(var context:TScriptContext; var State:TParamState):PUA_UnicodeString;
function LoadParameter_xArray(var context:TScriptContext; var State:TParamState; NeededType:TParamType):Pointer;

procedure PushRetAddr(var context:TScriptContext; RetAddr:Pointer);
procedure makeCALL(var context:TScriptContext; NewIP:Pointer);
function GetParamCount(start:PByte; var NextCommand:Pointer):Cardinal;

procedure InitContext(var context:TScriptContext; _CodeStart:Pointer;_CodeLen:cardinal);
procedure DestroyContext(var context:TScriptContext);
function CompileScript(var context:TScriptContext):integer;
function ExecuteByteCode(var context:TScriptContext; Section:PAnsiChar):Integer;

exports InitContext;
exports CompileScript;
exports ExecuteByteCode;

procedure RestoreDateTime;
//    ,    

implementation

uses NeoS0M_Constants, NeoS0M_Commands, NeoS0M_VarManager,NeoS0M_ServiceFns, NeoS0M_Common, NeoS0M_UStr;

procedure BuildSectionList(var context:TScriptContext);
var p:Pointer;
    i,l,sum_ofs:Cardinal;
begin
with context do
begin
//         
 p:=Pointer(Cardinal(ScriptStart)+8);
 SectionCount:=PCardinal(p)^;
 GetMem(SectionIndex,SectionCount*SizeOf(TSection));
//        
 sum_ofs:=0;
 Inc(NativeUInt(p),4);
 for i:=1 to SectionCount do
  begin
   with SectionIndex^[i] do
   begin
    Name:=p;
    l:=lstrlenA(p);
    Inc(NativeUInt(p),l+1);
    SectionFlag:=PCardinal(p)^;
    //  4  -   
    Inc(NativeUInt(p),SizeOf(Cardinal));
    //        -  
    SectionPtr:=Pointer(sum_ofs);
    SectionLength:=PCardinal(p)^;
    // __=__+_
    Inc(sum_ofs,PCardinal(p)^);
    //   4  -  
    Inc(NativeUInt(p),SizeOf(Cardinal));
   end;//with
  end;//for
 //   p      
 //   ,      
 for i:=1 to SectionCount do
  begin
   with SectionIndex^[i]
    do SectionPtr:=Pointer(Cardinal(SectionPtr)+Cardinal(p));
  end;//for

end;//with
end;

procedure InitContext(var context:TScriptContext; _CodeStart:Pointer;_CodeLen:cardinal);
var s_flag,s_len:Cardinal;
begin
with context do
begin
  MyOwnModule:=MyModuleHandle;
  ScriptStart:=_CodeStart;
  ScriptLength:=_CodeLen;
  StackFrames:=nil;
  CallStack:=nil;

  SecurityAttributes:=nil;
  BuildSectionList(context);
  LongDataSection:=GetSectionPtrByName(context,LDS_SECTION_NAME,s_flag, s_len);

  ActionAfterError:=aaeTerminateWithMessageBox;
  KeyPressTime:=50;
  SendMsgTimeout:=500;
  IgnoreSymlinks:=0; //  "      " 
  FileDialogBufferSize:=$8000; //   - 32K 

  ExitFlag:=False;
end;//with
end;

procedure DestroyContext(var context:TScriptContext);
begin
with context do
begin
 DestroyAllVariables(context);
 FreeMem(SectionIndex);
end;
end;

function BuildCommandList(var CmdCount:Cardinal):PExecutors;
//      ;    
begin
 CmdCount:=High(Executors)+1;
 result:=@Executors;      { TODO :       .   . }
end;

function ValidParameterCount(var codez:Pointer; required_param_count:integer):Boolean;
//    ;     - 
//  codez   
var param_c:integer;
    pp:Pointer;
    _exit:Boolean;
begin
 result:=True;
 NativeUInt(pp):=NativeUInt(codez)+4; //     
 if required_param_count>0 then
  begin
   param_c:=0;
   _exit:=False;
   repeat
    case PByte(pp)^ of
     Byte(ptDefault)..Byte(ptVariable),Byte(ptByte),Byte(ptWord),Byte(ptUnicodeString),Byte(ptUnicodeStringList):
     //    ,     
       begin
        Inc(param_c);
        Inc(NativeUInt(pp),ParamTypeSize[TParamType(PByte(pp)^)]+1);
       end;
     Byte(ptEndOfParam): //  "  " -  1 
       begin
        Inc(NativeUInt(pp));
        _exit:=true;
       end;
     else //   
       begin
        result:=False;
        _exit:=True;
       end;
    end; //case
    _exit:=_exit or (required_param_count=param_c);
   until _exit;
  end;//else
  if Result then codez:=pp;
end;

function CompileScript(var context:TScriptContext):integer;
var cmd_count,section:Cardinal;
    CmdList:PExecutors;

    function CompileByteCode(codez:Pointer; code_len:Cardinal):Cardinal;
    var idx,code_end:Cardinal;
    begin
      result:=CBC_OK;
      code_end:=Cardinal(codez)+code_len-1;
      //    
      while Cardinal(codez)<=code_end do
      begin
        idx:=PCardinal(codez)^;
        if idx>=cmd_count //    
          then begin context.Error_At:=codez; result:=CBC_INVALID_COMMAND; Break; end
        else //   
          begin
            with CmdList^[idx] do
            begin
              PCommandExecutor(codez)^:=Executor;
                if not ValidParameterCount(codez,ParamCount) then
                  begin
                    context.Error_At:=codez;
                    result:=CBC_INVALID_PARAMETER_COUNT;
                    Break;
                  end;//if not ValidParameterCount
            end;//with
          end;//else
      end;//while
    end;//function

    procedure ApplyJumpTable(jts_index:Cardinal);
    //   (  )     
    var item_ptr,addr_to_correct:PCardinal;
        i:Cardinal;
    begin
     with context,SectionIndex^[jts_index] do
      begin
       item_ptr:=SectionPtr;
       // _=_JTS/4
       for i:=1 to (SectionLength shr 2) do
        begin
         addr_to_correct:=PCardinal(Cardinal(ScriptStart)+item_ptr^);
         Inc(addr_to_correct^,Cardinal(ScriptStart));
         Inc(item_ptr);
        end;//for
      end;//with
    end;

begin
Result:=CBC_OK;
CmdList:=BuildCommandList(Cmd_count);
//   ,   
with context do
begin
 for section:=1 to SectionCount do
 begin
 with SectionIndex^[section] do
  begin
  if (SectionFlag and CODE_SECTION)=CODE_SECTION
    then Result:=CompileByteCode(SectionPtr,SectionLength) //     -  
    else
    //    
     begin
      // ,     JTS
      if lstrcmpA(Name,JUMP_SECTION_NAME)=0 then ApplyJumpTable(section);
     end;//else
  end;//with
 end//for
end;//with
end;

type TErrorPlaceholder=array[1..64] of WideChar;

{ TODO :     64-  }
{$IFDEF Win64}
procedure PrintErrorMessage(const c:TErrorPlaceholder; const text:PChar; const ea,sec:Cardinal);
asm
  call wsprintf
end;
{$ENDIF}

procedure ShowErrorInfo(const context:TScriptContext);
var Buffer:TErrorPlaceholder;
    ea,sec:Cardinal;
begin
  with context do
  begin
    ea:=Cardinal(Error_At)-Cardinal(ScriptStart);
    sec:=ScriptErrorCode
  end;//with

  {$IFDEF Win32}
  asm
    push ea
    push sec
    push _ERROR_IN_SCRIPT
    lea eax,Buffer
    push eax
    call wsprintfW
    add esp,16
  end;//asm
  {$ENDIF}

  {$IFDEF Win64}
  PrintErrorMessage(Buffer,_ERROR_IN_SCRIPT,ea,sec);
  {$ENDIF}

  MessageBox(0,@Buffer,PWideChar(KernelName),MB_APPLMODAL or MB_ICONERROR);
end;


procedure PushRetAddr(var context:TScriptContext; RetAddr:Pointer);
var l:Integer;
begin
with context do
begin
   l:=Length(CallStack);
   SetLength(CallStack,l+1);
   CallStack[l]:=RetAddr;
end;//with
end;

procedure makeCALL(var context:TScriptContext; NewIP:Pointer);
begin
with context do
begin
  PushRetAddr(context,IP);
  IP:=NewIP;
end;//with
end;

function ExecuteOneCommand(var context:TScriptContext):Integer;
var err_addr:Pointer;
    cmd:TCommandExecutor;
begin
with context do
begin
 err_addr:=IP;
 cmd:=PCommandExecutor(IP)^;
 Inc(NativeUInt(IP),4);
 //     IP    ,   ,
 //    
 result:=TCommandExecutor(cmd)(context);
 ScriptErrorCode:=Result;
 if Result<>CMD_OK then //      
  begin
   Error_At:=err_addr; //   error_at  
   case ActionAfterError of
    aaeTerminate:               ExitFlag:=True;
    aaeTerminateWithMessageBox: begin
                                  ExitFlag:=True;
                                  ShowErrorInfo(context);
                                end;
    aaeHook:                    begin
                                 if OnErrorGoto<>nil //   
                                  then MakeCALL(context,OnErrorGoto);
                                end;
   end;//case
  end;//Result<>CMD_OK
end;//with
end;

function ExecuteByteCode(var context:TScriptContext; Section:PAnsiChar):Integer;
var f,len:Cardinal;
begin
with context do
 begin
  IP:=GetSectionPtrByName(Context,Section,f,len);
  if IP=nil then begin Result:=CMD_NO_SECTION; Exit; end;
  if (f and CODE_SECTION)<>CODE_SECTION
   then Result:=CBC_NON_EXECUTABLE_SECTION
   else
    begin
     //     
     result:=C_LFRAME_ADD(context);
     try
       while not ExitFlag do
       begin
        result:=ExecuteOneCommand(context);
       end;//while
     except
       Result:=CMD_EXCEPTION;
     end//try-except
    end;//else
 end;//with
end;

function LoadParameter_DWORD(var context:TScriptContext; var State:TParamState; def_value:Cardinal):Cardinal;
var param_ptr:Pointer;
begin
 result:=0;
 state:=PARAM_OK;
with context do
begin
 param_ptr:=Pointer(Cardinal(IP)+1);
 case PByte(IP)^ of
  Byte(ptDefault):  begin
                     State:=PARAM_DEFAULT;
                     Result:=def_value;
                    end;
  Byte(ptEndOfParam): begin
                      State:=PARAM_END;
                      Result:=def_value;
                      end;
  Byte(ptDWORD):    result:=PCardinal(param_ptr)^;
  Byte(ptInteger):  result:=PInt64(param_ptr)^;
  Byte(ptByte):     result:=PShortInt(param_ptr)^;
  Byte(ptWord):     result:=PSmallInt(param_ptr)^;
  Byte(ptVariable): begin
{ DONE :     DWORD'   }
                     result:=PCardinal(param_ptr)^;

                     if Variable_qLoadDWORD(context,Result,Result)<>CMD_OK
                      then State:=PARAM_INVALID;
                    end;
  else State:=PARAM_INVALID;
 end;//case
 //  IP    
 Inc(NativeUInt(IP),ParamTypeSize[TParamType(PByte(IP)^)]+1);
end;//with
end;

function LoadParameter_INTEGER(var context:TScriptContext; var State:TParamState; const def_value:Int64):Int64;
var param_ptr:Pointer;
    tmp_dword:Cardinal;
begin
 result:=0;
 State:=PARAM_OK;
with context do
begin
 param_ptr:=Pointer(Cardinal(IP)+1);
 case PByte(IP)^ of
  Byte(ptDefault):  begin
                     State:=PARAM_DEFAULT;
                     Result:=def_value;
                    end;
  Byte(ptEndOfParam): begin
                      State:=PARAM_END;
                      Result:=def_value;
                      end;
  Byte(ptDWORD):    result:=PCardinal(param_ptr)^;
  Byte(ptInteger):  result:=PInt64(param_ptr)^;
  Byte(ptByte):     result:=PShortInt(param_ptr)^;
  Byte(ptWord):     result:=PSmallInt(param_ptr)^;
  Byte(ptVariable): begin
{ DONE :     DWORD'   }
                     tmp_dword:=PCardinal(param_ptr)^;

                     if Variable_qLoadINTEGER(context,tmp_dword,Result)<>CMD_OK
                      then State:=PARAM_INVALID;
                    end;
  else State:=PARAM_INVALID;
 end;//case
 //  IP    
 Inc(NativeUInt(IP),ParamTypeSize[TParamType(PByte(IP)^)]+1);
end;//with
end;

function LoadParameter_FLOAT(var context:TScriptContext; var State:TParamState; const def_value:Extended):Extended;
var param_ptr:Pointer;
    tmp_dword:Cardinal;
begin
 result:=0;
 State:=PARAM_OK;
with context do
begin
 param_ptr:=Pointer(Cardinal(IP)+1);
 case PByte(IP)^ of
  Byte(ptDefault):  begin
                     State:=PARAM_DEFAULT;
                     Result:=def_value;
                    end;
  Byte(ptEndOfParam): begin
                      State:=PARAM_END;
                      Result:=def_value;
                      end;
  Byte(ptExtended): result:=PExtended(param_ptr)^;
  Byte(ptDWORD):    result:=PCardinal(param_ptr)^;
  Byte(ptInteger):  result:=PInt64(param_ptr)^;
  Byte(ptByte):     result:=PShortInt(param_ptr)^;
  Byte(ptWord):     result:=PSmallInt(param_ptr)^;
  Byte(ptVariable): begin
                     tmp_dword:=PCardinal(param_ptr)^;

                     if Variable_qLoadFLOAT(context,tmp_dword,Result)<>CMD_OK
                      then State:=PARAM_INVALID;
                    end;
  else State:=PARAM_INVALID;
 end;//case
 //  IP    
 Inc(NativeUInt(IP),ParamTypeSize[TParamType(PByte(IP)^)]+1);
end;//with
end;


function LoadParameter_xArray(var context:TScriptContext; var State:TParamState; NeededType:TParamType):Pointer;
var var_idx:Cardinal;
begin
 result:=nil;
 State:=PARAM_OK;
with context do
begin
 case PByte(IP)^ of
  Byte(ptDefault):  begin
                     State:=PARAM_DEFAULT;
                    end;
  Byte(ptEndOfParam): begin
                      State:=PARAM_END;
                      end;
  Byte(ptVariable): begin
                      var_idx:=PCardinal(Cardinal(IP)+1)^;
                      if Variable_qLoadxArray(context,var_idx,CompatibleVarType[NeededType],Result)<>CMD_OK
                        then State:=PARAM_INVALID;
                    end;
  else              begin
                      if PByte(IP)^=Byte(NeededType) //          
                        then result:=Pointer(Cardinal(LongDataSection)+PCardinal(Cardinal(IP)+1)^)
                        else State:=PARAM_INVALID;
                    end;
 end;//case
 //  IP    
 Inc(NativeUInt(IP),ParamTypeSize[TParamType(PByte(IP)^)]+1);
end;//with
end;

function LoadParameter_BinBlock(var context:TScriptContext; var State:TParamState):PUA_BinaryBlock;
begin
  Result:=LoadParameter_xArray(context,State,ptBinBlock);
end;

function LoadParameter_UnicodeString(var context:TScriptContext; var State:TParamState):PUA_UnicodeString;
var v_a:PNeoS0MVariable;
begin
 Result:=nil;
 State:=PARAM_OK;
  with context do
  begin
    case PByte(IP)^ of
      Byte(ptDefault):        State:=PARAM_DEFAULT;
      Byte(ptEndOfParam):     State:=PARAM_END;
      Byte(ptBinBlock):       begin
                                //       Unicode-
                                Result:=BB_To_UStr(PUA_BinaryBlock(Cardinal(LongDataSection)+PCardinal(Cardinal(IP)+1)^));
                              end;
      Byte(ptUnicodeString):  begin
                                //   Unicode-
                                Result:=PUA_UnicodeString(Cardinal(LongDataSection)+PCardinal(Cardinal(IP)+1)^);
                              end;
      Byte(ptVariable):       begin
                                //   
                                v_a:=Variable_GetAddress(context,PCardinal(Cardinal(IP)+1)^);
                                if v_a<>nil then //  ?
                                begin
                                case v_a^.VarType of
                                                    //   Unicode-    
                                  vtBinBlock:       Result:=BB_To_UStr(PUA_BinaryBlock(v_a^.UA));
                                                    //  Unicode-;      Unicode-
                                  vtUnicodeString:  Result:=UStr_Duplicate(PUA_UnicodeString(v_a^.UA));
                                  //      ,  Unicode-
                                  else State:=PARAM_INVALID
                                end;//case
                                end//if v_a<>nil
                                else State:=PARAM_INVALID; //   
                              end;
  else State:=PARAM_INVALID;
 end;//case
 //  IP    
 Inc(NativeUInt(IP),ParamTypeSize[TParamType(PByte(IP)^)]+1);
end;//with


// Result:=LoadParameter_xArray(context,State,ptUnicodeString);
end;

function GetParamCount(start:PByte; var NextCommand:Pointer):Cardinal;
var pt:TParamType;
begin
 Result:=0;
 while True do
 begin
  pt:=TParamType(start^);
  //   .   . 
  Inc(start,ParamTypeSize[pt]+1);
  if pt=ptEndOfParam then //       -     
    begin
     NextCommand:=start;
     Break;
    end; //if
  //     
  Inc(Result);
 end;
end;

procedure RestoreDateTime;
var st:SYSTEMTIME;
    d2:Int64;
begin
 if DateTimeChanged then
  begin
   d2:=Cardinal(GetTickCount-BaseTickCount);
   Inc(Int64(BaseTimeToRestore),d2*10000); //     10000  SYSTEMTIME
   FileTimeToSystemTime(BaseTimeToRestore,st);
   SetSystemTime(st);
   DateTimeChanged:=False;
  end;
end;

initialization
begin
  KernelName:=NeoS0M_Name+' v. '+WideChar(MajorKernelVersion+Ord('0')) +'.'+WideChar(MinorKernelVersion+Ord('0'));
  if AdditionalMinorKernelVersion<>0
    then KernelName:=KernelName+'.'+WideChar(AdditionalMinorKernelVersion+Ord('0'));
end;

finalization
begin
  //   
  UnloadAllLibraries;
  if AutoRestoreTime<>0
    then RestoreDateTime;
end;

end.
