unit NeoS0M_UnifiedArray;

interface

uses NeoS0M_Types;

function UA_Create(init_item_count:cardinal; _ItemSize:cardinal):PArrayHeader;
//        
function UA_CreateR(init_item_count,_ItemSize:Cardinal):PArrayHeader;
//      c   init_item_count
//         REALLOC_GRANULARITY
procedure UA_Reallocate(var arr:PArrayHeader; _ItemCount,_ItemSize:Cardinal);
//    .   !     !
procedure UA_Resize(var arr:PArrayHeader; NewSize:cardinal);
//   ,   -  .   !
procedure UA_Destroy(var arr:PArrayHeader);
//     .      0

implementation

uses NeoS0M_Math;

const REALLOC_GRANULARITY=32; //   ,        

function UA_Create(init_item_count:cardinal; _ItemSize:cardinal):PArrayHeader;
begin
  result:=nil;
  UA_Reallocate(result,init_item_count,_ItemSize);
  if result<>nil
    then result^.ElemCount:=init_item_count;
end;

function UA_CreateR(init_item_count,_ItemSize:Cardinal):PArrayHeader;
begin
  Result:=UA_Create(RoundUp(init_item_count,REALLOC_GRANULARITY),_ItemSize);
  if Result<>nil
    then Result^.ElemCount:=init_item_count;
end;

procedure UA_Reallocate(var arr:PArrayHeader; _ItemCount,_ItemSize:Cardinal);
var OldSize,NewSize,OldItemCount:cardinal;
    NewArray:PArrayHeader;
begin
 NewSize:=_ItemSize*_ItemCount+SizeOf(TArrayHeader);
 GetMem(NewArray,NewSize);
 if NewArray<>nil then
  begin
  if (arr<>nil) then
  //     -     
   begin
    with arr^ do
     begin
      OldItemCount:=ElemCount;
      OldSize:=ElemSize*ElemCount+SizeOf(TArrayHeader);
      Move(arr^,NewArray^,Min(NewSize,OldSize));
      //    -    
      if IsDynamic
        then FreeMem(arr);
     end;
   end
   else OldItemCount:=0;
   //      
   with NewArray^ do
    begin
     ElemSize:=_ItemSize;
     TrueElemCount:=_ItemCount;
     ElemCount:=Min(_ItemCount,OldItemCount);
     IsDynamic:=True;
    end;
 end;//if NewArray<>nil
 //     
 arr:=NewArray;
end;

procedure UA_Resize(var arr:PArrayHeader; NewSize:Cardinal);
var AllocSize:Cardinal;
begin
  AllocSize:=RoundUp(NewSize,REALLOC_GRANULARITY);
  if arr<>nil then
    begin
      if arr^.IsDynamic then
      begin
        with arr^ do
          begin
            //      ,   
            if (NewSize>TrueElemCount) or (AllocSize<TrueElemCount)
              then UA_Reallocate(arr,AllocSize,ElemSize);
          end;//with
      end//if IsDynamic
      else UA_Reallocate(arr,AllocSize,arr^.ElemSize); //   -      
    end // if arr<>nil
  else UA_Reallocate(arr,AllocSize,arr^.ElemSize); //      -   
  //       
  if arr<>nil
    then arr^.ElemCount:=NewSize;
end;

procedure UA_Destroy(var arr:PArrayHeader);
begin

  if arr<>nil then
  begin
    if arr^.IsDynamic
      then FreeMem(arr);
    arr:=nil;
  end;//if arr<>nil
end;

end.
